/* PLIX ESP8285 MODULE FIRMWARE */

#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266httpUpdate.h>
#include <ArduinoJson.h>
#include <LittleFS.h>
#include <time.h>
#include <OneWire.h>
#include <DallasTemperature.h>

const String FIRMWARE_VERSION = "1.5.3";
const int MAX_NETWORKS_CNT = 10;
const int JSON_DATA_SIZE = 2048;
const int SERVER_PORT = 80;
const int MAX_TEMPERATURE_SENSORS_CNT = 3;
const int TEMPERATURE_ADDRESS_SIZE = 8;

int NETWORKS_CNT = 0;
struct NetworkItem {
  String ssid;
  String password;
};
struct NetworkItem NETWORKS[MAX_NETWORKS_CNT];

int SCANNED_NETWORKS_CNT = 0;
struct ScannedNetworkItem {
  String ssid;
  int rssi;
  String bssid;
  int channel;
  int encryptionType;
};
struct ScannedNetworkItem SCANNED_NETWORKS[MAX_NETWORKS_CNT];

int TEMPERATURE_SENSORS_CNT = 0;
DeviceAddress TEMPERATURE_SENSOR_ADDRESSES[MAX_TEMPERATURE_SENSORS_CNT];

DynamicJsonDocument JSON_DATA(JSON_DATA_SIZE);
ESP8266WebServer SERVER(SERVER_PORT);
WiFiClientSecure CLIENT;
DallasTemperature TEMPERATURE_SENSORS;

class Debug {
  private:
    static const int LEVEL = 3;
    static const int BAUD_RATE = 115200;
  public:
    static void init();
    static void print(String message);
    static void printLine(String message);
    static void printStart(String message);
    static void printSeparationLine(String message);
    static void printWarning(String message);
    static void printError(String message);
    static void printNull();
    static void printBoolean(bool value);
    static void printSetupInfo();
    static void printJsonData(String message);
    static void printNetworksData(String message);
    static void printScannedNetworksData(String message);
};

class Data {
  private:
    static const String DEFAULT_FILE;
    static const String VARIABLES_FILE;
    static const String CONFIG_FILE;
    static const String SETTINGS_FILE;
    static const String NETWORKS_FILE;
    static const String DEVICE_NAME_PREFIX;
    static bool loadJsonFromFile(String fileName);
    static bool saveJsonToFile(String fileName);
  public:
    static String DEVICE_ID;
    static String DEVICE_NAME;
    static String DEVICE_TOKEN;
    static bool init();
    static bool clearJson();
    static bool loadDefault();
    static bool loadVariables();
    static bool loadConfig();
    static bool loadSettings();
    static bool loadNetworks();
    static bool saveVariables();
    static bool saveConfig();
    static bool saveSettings();
    static bool saveNetworks();
    static bool loadBooleanFromJson(bool &variable, String key);
    static bool loadIntegerFromJson(int &variable, String key);
    static bool loadStringFromJson(String &variable, String key);
    static bool loadIsTrueFromJson(String key);
    static bool loadVariablesFromJson();
    static bool loadConfigFromJson();
    static bool loadSettingsFromJson();
    static void saveVariablesToJson();
    static void saveConfigToJson();
    static void saveSettingsToJson();
    static int loadNetworksFromJson(String key);
    static void saveNetworksToJson(String key);
    static void scanAndSaveNetworksToJson(String key);
    static bool loadTempSensorAddressesFromJson(String key);
    static void saveTempSensorAddressesToJson(String key);
};
const String Data::DEFAULT_FILE = "/def.json";
const String Data::VARIABLES_FILE = "/var.json";
const String Data::CONFIG_FILE = "/cfg.json";
const String Data::SETTINGS_FILE = "/stg.json";
const String Data::NETWORKS_FILE = "/net.json";
const String Data::DEVICE_NAME_PREFIX = "PLIX-SM-";
String Data::DEVICE_ID = "";
String Data::DEVICE_NAME = "";
String Data::DEVICE_TOKEN = "";

class Pin {
  private:
    static const int LED_BLINK_ON_TIME = 200;
    static const int LED_BLINK_DELAY = 400;
    static const int BUTTON_MIN_DELAY = 100;
    static const int BUTTON_LONG_PRESS_DELAY = 3000;
  public:
    static const int NONE = -1;
    static int BUTTON;
    static int RELAY;
    static int LED;
    static int EXT_BUTTON;
    static int EXT_LED;
    static int TEMP_SENSORS;
    static int LED_1;
    static int LED_2;
    static int LED_3;
    static bool RELAY_STATE;
    static bool RELAY_BUTTON_MODE;
    static bool RELAY_DEFAULT_STATE;
    static int RELAY_MAX_STATE_DELAY;
    static int LED_1_VALUE;
    static int LED_2_VALUE;
    static int LED_3_VALUE;
    static int DEFAULT_LED_1_VALUE;
    static int DEFAULT_LED_2_VALUE;
    static int DEFAULT_LED_3_VALUE;
    static unsigned long LAST_BUTTON_FALL_TIME;
    static unsigned long LAST_BUTTON_RISE_TIME;
    static void init();
    static void changeRelayState(bool state);
    static void pressRelay();
    static void changeLedValues(int led1, int led2, int led3);
    static void protectLongRelayStateChange();
    static void setLedOn();
    static void setLedOff();
    static void ledBlink(int cnt);
    static void ICACHE_RAM_ATTR buttonInterrupt();
    static void ICACHE_RAM_ATTR extButtonInterrupt();
    static void buttonRiseInterrupt();
};
int Pin::BUTTON = Pin::NONE;
int Pin::RELAY = Pin::NONE;
int Pin::LED = Pin::NONE;
int Pin::EXT_BUTTON = Pin::NONE;
int Pin::EXT_LED = Pin::NONE;
int Pin::LED_1 = Pin::NONE;
int Pin::LED_2 = Pin::NONE;
int Pin::LED_3 = Pin::NONE;
int Pin::TEMP_SENSORS = Pin::NONE;
bool Pin::RELAY_STATE = false;
bool Pin::RELAY_BUTTON_MODE = false;
bool Pin::RELAY_DEFAULT_STATE = false;
int Pin::RELAY_MAX_STATE_DELAY = 60000;
int Pin::LED_1_VALUE = 0;
int Pin::LED_2_VALUE = 0;
int Pin::LED_3_VALUE = 0;
int Pin::DEFAULT_LED_1_VALUE = 63;
int Pin::DEFAULT_LED_2_VALUE = 63;
int Pin::DEFAULT_LED_3_VALUE = 63;
unsigned long Pin::LAST_BUTTON_FALL_TIME = 0;
unsigned long Pin::LAST_BUTTON_RISE_TIME = 0;

class Network {
  private:
    static const int MIN_TIMESTAMP = 1577836800;
    static const int MAX_CONNECTION_DELAY = 60000;
    static const int MAX_REQUEST_REPEAT_CNT = 5;
    static const int REQUEST_REPEAT_DELAY = 8000;
    static const int ENCRYPTION_TYPE_ERROR = -1;
    static const int ENCRYPTION_TYPE_NONE = 0;
    static const int ENCRYPTION_TYPE_WEP = 1;
    static const int ENCRYPTION_TYPE_WPA = 2;
    static const int ENCRYPTION_TYPE_WPA2 = 3;
    static const int ENCRYPTION_TYPE_AUTO = 4;
    static const int PORT = 443;
    static const String HOST;
    static const String SERVER_MODE_CONFIG_URL;
    static IPAddress LOCAL_IP;
    static IPAddress GATEWAY;
    static IPAddress SUBNET;
    static bool connectToServer();
    static bool httpsRequestPost();
    static void sort();
    static int getNetworkPos(String ssid);
  public:
    static const String REST_API_BASE_URL;
    static int CONNECT_STATUS_OK;
    static int CONNECT_STATUS_AP_ERROR;
    static int CONNECT_STATUS_SERVER_ERROR;
    static unsigned long LAST_SUCCESS_REQUEST_TIME;
    static String SERVER_PASSWORD;
    static String REST_API_URL;
    static String UPDATE_URL;
    static void init();
    static bool isConnected();
    static bool disconnect();
    static int connect();
    static int connectTo(String ssid, String pass);
    static bool synchronizeTime();
    static bool httpsRequest();
    static bool httpsUpdate();
    static bool scan();
    static int getEncryptionType(int type);
};
const String Network::HOST = "sm.plix.cz";
const String Network::SERVER_MODE_CONFIG_URL = "/plix-sm-config";
IPAddress Network::LOCAL_IP(192, 168, 0, 1);
IPAddress Network::GATEWAY(192, 168, 0, 1);
IPAddress Network::SUBNET(255, 255, 255, 0);
const String Network::REST_API_BASE_URL = "/api/v1/";
int Network::CONNECT_STATUS_OK = 0;
int Network::CONNECT_STATUS_AP_ERROR = 1;
int Network::CONNECT_STATUS_SERVER_ERROR = 2;
unsigned long Network::LAST_SUCCESS_REQUEST_TIME = 0;
String Network::SERVER_PASSWORD = "";
String Network::REST_API_URL = "";
String Network::UPDATE_URL = "";

class Main {
  private:
    static const int SETUP_DELAY = 3000;
    static const int MAX_CLIENT_LOOP_DELAY = 60000;
    static const int MAX_SERVER_LOOP_DELAY = 300000;
    static void protectLongServerLoop();
    static void processAll();
    static bool checkNetworkConnection();
    static void prepareRegularData();
    static void processRelayState();
    static void processLedState();
    static void processUpdate();
    static void processRestart();
    static void processTmpDelay();
    static void processConnect();
    static void loopDelay();
  public:
    static const int MIN_DELAY = 100;
    static const int SLEEP_DELAY = 1000;
    static bool SERVER_MODE;
    static bool BREAK_SLEEP;
    static int LOOP_DELAY;
    static int TMP_DELAY;
    static unsigned long LAST_LOOP_TIME;
    static void setup();
    static void restart();
    static void serverNetworkSaveRequest();
    static void serverLoop();
    static void clientLoop();
    static void changeModuleMode(bool serverMode);
    static void sleep(int duration);
};
bool Main::SERVER_MODE = false;
bool Main::BREAK_SLEEP = false;
int Main::LOOP_DELAY = 10000;
int Main::TMP_DELAY = 0;
unsigned long Main::LAST_LOOP_TIME = 0;

void setup() {
  Main::setup();
}

void loop() {
  Debug::printSeparationLine("----------------- LOOP -----------------");
  Debug::printLine("Time: " + String(millis()) + " ms");
  Debug::printLine("Free heap: " + String(ESP.getFreeHeap()) + " B");
  if (Main::SERVER_MODE) {
    Main::serverLoop();
  } else {
    Main::clientLoop();
  }
}
