/* NETWORK CLASS */

void Network::init() {
  if (Main::SERVER_MODE) {
    Network::disconnect();
    Debug::printStart("Configuring server network");
    if (Network::SERVER_PASSWORD.length() > 0) {
      if (!WiFi.softAP(Data::DEVICE_NAME, Network::SERVER_PASSWORD)) {
        Main::changeModuleMode(false);
        return;
      }
    } else {
      if (!WiFi.softAP(Data::DEVICE_NAME)) {
        Main::changeModuleMode(false);
        return;
      }
    }
    WiFi.softAPConfig(Network::LOCAL_IP, Network::GATEWAY, Network::SUBNET);
    delay(Main::MIN_DELAY);
    SERVER.on(SERVER_MODE_CONFIG_URL, Main::serverNetworkSaveRequest);
    SERVER.begin();
    delay(Main::MIN_DELAY);
    Debug::printLine("OK");
  } else {
    Debug::printStart("Configuring client network");
    WiFi.mode(WIFI_STA);
    WiFi.hostname(Data::DEVICE_NAME);
    Debug::printLine("OK");
    Network::disconnect();
  }
}

bool Network::isConnected() {
  Debug::printStart("Checking network connection");
  if (WiFi.status() == WL_CONNECTED) {
    Debug::printLine("OK");
    return true;
  }
  Debug::printWarning("NETWORK CHECK: Network is not connected.");
  return false;
}

bool Network::disconnect() {
  Debug::printStart("Disconnecting from network");
  WiFi.disconnect();
  delay(Main::MIN_DELAY);
  Debug::printLine("OK");
  return false;
}

int Network::connect() {
  int lastNetworkStatus = Network::CONNECT_STATUS_AP_ERROR;
  Network::scan();
  Debug::printLine("NETWORK: Connecting to best available network: " + String(SCANNED_NETWORKS_CNT));
  for (int i = 0; i < SCANNED_NETWORKS_CNT; i++) {
    String ssid = SCANNED_NETWORKS[i].ssid;
    int pos = Network::getNetworkPos(ssid);
    if (pos >= 0) {
      String password = NETWORKS[pos].password;
      lastNetworkStatus = Network::connectTo(ssid, password);
      if (lastNetworkStatus == Network::CONNECT_STATUS_OK) {
        return Network::CONNECT_STATUS_OK;
      }
    }
  }
  Debug::printWarning("NETWORK: Cannot connect to best available network.");
  Debug::printLine("NETWORK: Connecting to saved network: " + String(NETWORKS_CNT));
  for (int i = 0; i < NETWORKS_CNT; i++) {
    String ssid = NETWORKS[i].ssid;
    String password = NETWORKS[i].password;
    lastNetworkStatus = Network::connectTo(ssid, password);
    if (lastNetworkStatus == Network::CONNECT_STATUS_OK) {
      return Network::CONNECT_STATUS_OK;
    }
  }
  Debug::printWarning("NETWORK: Cannot connect to saved network.");
  return lastNetworkStatus;
}

int Network::connectTo(String ssid, String pass) {
  Network::disconnect();
  Debug::printStart("Connecting to '" + ssid + "'");
  WiFi.begin(ssid.c_str(), pass.c_str());
  for (int i = 0; i < Network::MAX_CONNECTION_DELAY / Main::SLEEP_DELAY; i++) {
    byte status = WiFi.status();
    if (status == WL_CONNECTED) {
      Debug::printLine("OK");
      if (Network::synchronizeTime()) {
        if (Network::connectToServer()) {
          return Network::CONNECT_STATUS_OK;
        }
      }
      return Network::CONNECT_STATUS_SERVER_ERROR;
    } else if (status == WL_NO_SSID_AVAIL) {
      Debug::printWarning("NETWOTK CONNECT: Network '" + ssid + "' is not available.");
      return Network::CONNECT_STATUS_AP_ERROR;
    } else if (status == WL_CONNECT_FAILED) {
      Debug::printWarning("NETWOTK CONNECT: Password of network '" + ssid + "' is incorrect.");
      return Network::CONNECT_STATUS_AP_ERROR;
    }
    Debug::print(".");
    delay(Main::SLEEP_DELAY);
  }
  Debug::printError("NETWOTK CONNECT: Cannost connect to network '" + ssid + "'.");
  return Network::CONNECT_STATUS_AP_ERROR;
}

bool Network::synchronizeTime() {
  Debug::printStart("Syncing time");
  configTime(0, 0, "pool.ntp.org", "time.nist.gov");
  time_t now = time(nullptr);
  for (int i = 0; i < Network::MAX_CONNECTION_DELAY / Main::SLEEP_DELAY; i++) {
    if (!Network::isConnected()) {
      Debug::printError("TIME SYNC: Network is no longer available.");
      return false;
    }
    if (now < Network::MIN_TIMESTAMP) {
      Debug::print(".");
      delay(Main::SLEEP_DELAY);
      now = time(nullptr);
    } else {
      struct tm timeinfo;
      gmtime_r(&now, &timeinfo);
      Debug::printStart("OK");
      Debug::print(String(asctime(&timeinfo)));
      delay(Main::SLEEP_DELAY);
      return true;
    }
  }
  Debug::printError("TIME SYNC: Cannot synchronize time.");
  return false;
}

bool Network::connectToServer() {
  if (!Network::isConnected()) {
    Debug::printError("CONNECT TO SERVER: Network is no longer available.");
    return false;
  }

  Debug::printStart("Connecting to server 'sm.plix.cz' with certificate #1");
  CLIENT.setTrustAnchors(&CERT_1);
  CLIENT.connect(Network::HOST, Network::PORT);
  if (CLIENT.connected()) {
    Debug::printLine("OK");
    JSON_DATA["connCert"] = 1;
    CLIENT.stop();
    return true;
  }
  Debug::printError("CONNECT TO SERVER: Cannot connect to server 'sm.plix.cz' with certificate #1.");

  Debug::printStart("Connecting to server 'sm.plix.cz' with certificate #2");
  CLIENT.setTrustAnchors(&CERT_2);
  CLIENT.connect(Network::HOST, Network::PORT);
  if (CLIENT.connected()) {
    Debug::printLine("OK");
    JSON_DATA["connCert"] = 2;
    CLIENT.stop();
    return true;
  }
  Debug::printError("CONNECT TO SERVER: Cannot connect to server 'sm.plix.cz' with certificate #2.");

  Debug::printStart("Connecting to server 'sm.plix.cz' with certificate #3");
  CLIENT.setTrustAnchors(&CERT_3);
  CLIENT.connect(Network::HOST, Network::PORT);
  if (CLIENT.connected()) {
    Debug::printLine("OK");
    JSON_DATA["connCert"] = 3;
    CLIENT.stop();
    return true;
  }
  Debug::printError("CONNECT TO SERVER: Cannot connect to server 'sm.plix.cz' with certificate #3.");

  Debug::printStart("Connecting to server 'sm.plix.cz' with certificate #4");
  CLIENT.setTrustAnchors(&CERT_4);
  CLIENT.connect(Network::HOST, Network::PORT);
  if (CLIENT.connected()) {
    Debug::printLine("OK");
    JSON_DATA["connCert"] = 4;
    CLIENT.stop();
    return true;
  }
  Debug::printError("CONNECT TO SERVER: Cannot connect to server 'sm.plix.cz' with certificate #4.");

  /*Debug::printStart("Backup option: Connecting to server 'sm.plix.cz' without any certificate");
    CLIENT.setInsecure();
    CLIENT.connect(Network::HOST, Network::PORT);
    if (CLIENT.connected()) {
    Debug::printLine("OK");
    JSON_DATA["connCert"] = -1;
    CLIENT.stop();
    return true;
    }
    Debug::printError("CONNECT TO SERVER: Cannot connect to server 'sm.plix.cz' without any certificate.");*/

  Debug::printError("CONNECT TO SERVER: Cannot connect to server.");
  return false;
}

bool Network::httpsRequest() {
  for (int i = 0; i < Network::MAX_REQUEST_REPEAT_CNT; i++) {
    if (!Network::isConnected()) {
      Debug::printError("HTTPS REQUEST: Network is no longer available.");
      return false;
    }
    if (Network::httpsRequestPost()) {
      return true;
    }
    Main::sleep(REQUEST_REPEAT_DELAY);
  }
  return false;
}

bool Network::httpsRequestPost() {
  Debug::printJsonData("Posting JSON data");
  Debug::printStart("Requesting JSON data at '" + Network::REST_API_URL + "'");

  CLIENT.connect(Network::HOST, Network::PORT);
  if (!CLIENT.connected()) {
    Debug::printError("HTTPS REQUEST: Server is no longer available.");
    return false;
  }

  String jsonData = "";
  serializeJson(JSON_DATA, jsonData);
  CLIENT.print("POST ");
  CLIENT.print(Network::REST_API_URL);
  CLIENT.print(" HTTP/1.0");
  CLIENT.print("\r\nHost: ");
  CLIENT.print(Network::HOST);
  CLIENT.print("\r\nConnection: keep-alive");
  CLIENT.print("\r\nContent-Type: application/json");
  CLIENT.print("\r\nContent-Length: ");
  CLIENT.print(measureJson(JSON_DATA));
  CLIENT.print("\r\n\r\n");
  CLIENT.print(jsonData);
  delay(Main::MIN_DELAY);

  String line = CLIENT.readStringUntil('\r');
  if (line != "HTTP/1.0 200 OK") {
    Debug::printError("HTTPS REQUEST: Unexpected response status.");
    Debug::printLine(line);
    while (CLIENT.available()) {
      line = CLIENT.readStringUntil('\r');
      Debug::printError(line);
    }
    CLIENT.stop();
    return false;
  }
  while (CLIENT.available()) {
    line = CLIENT.readStringUntil('\r');
    if (line == "\n") {
      Debug::printLine("OK");
      Debug::printStart("Deserializing downloaded JSON data");
      DeserializationError error = deserializeJson(JSON_DATA, CLIENT);
      if (error) {
        Debug::printError("HTTPS REQUEST: Cannot deserialize JSON data.");
        Debug::printError(error.f_str());
        CLIENT.stop();
        return false;
      }
      Debug::printLine("OK");
      Debug::printJsonData("Downloaded JSON data");
      CLIENT.stop();
      return true;
    }
  }
  Debug::printError("HTTPS REQUEST: Server not available.");
  CLIENT.stop();
  return false;
}

bool Network::httpsUpdate() {
  Debug::printStart("Updating firmware");
  if (Network::UPDATE_URL.length() > 0) {
    ESPhttpUpdate.update(CLIENT, String("https://") + Network::HOST + Network::UPDATE_URL);
    Debug::printError("FIRMWARE UPDATE: Firmware update failed.");
    Network::UPDATE_URL = "";
  }
  return false;
}

bool Network::scan() {
  Network::disconnect();
  Debug::printStart("Scanning networks");
  SCANNED_NETWORKS_CNT = 0;
  int netCnt = WiFi.scanNetworks(false, true);
  if (netCnt < 0) {
    Debug::printError("NETWORK SCAN: Cannot scan networks.");
    return false;
  } else {
    if (netCnt == 0) {
      Debug::printWarning("NETWORK SCAN: No networks available.");
      return false;
    }
    if (netCnt > MAX_NETWORKS_CNT) {
      Debug::printWarning("NETWORK SCAN: More than maximum networks available.");
      netCnt = MAX_NETWORKS_CNT;
    }
    Debug::printLine("OK");
    for (int i = 0; i < netCnt; i++) {
      String ssid = WiFi.SSID(i);
      int rssi = WiFi.RSSI(i);
      String bssid = WiFi.BSSIDstr(i);
      int channel = WiFi.channel(i);
      int encryptionType = getEncryptionType(WiFi.encryptionType(i));
      SCANNED_NETWORKS[i].ssid = ssid;
      SCANNED_NETWORKS[i].rssi = rssi;
      SCANNED_NETWORKS[i].bssid = bssid;
      SCANNED_NETWORKS[i].channel = channel;
      SCANNED_NETWORKS[i].encryptionType = encryptionType;
    }
    SCANNED_NETWORKS_CNT = netCnt;
    Debug::printScannedNetworksData("Available networks");
    Network::sort();
    return true;
  }
}

void Network::sort() {
  Debug::printStart("Sorting networks");
  ScannedNetworkItem temp;
  for (int i = 0; i < SCANNED_NETWORKS_CNT; i++) {
    for (int j = i + 1; j < SCANNED_NETWORKS_CNT; j++) {
      if (SCANNED_NETWORKS[j].rssi > SCANNED_NETWORKS[i].rssi) {
        temp = SCANNED_NETWORKS[i];
        SCANNED_NETWORKS[i] = SCANNED_NETWORKS[j];
        SCANNED_NETWORKS[j] = temp;
        Debug::print(String(i) + " <-> " + String(j) + ", ");
      }
    }
  }
  Debug::printLine("OK");
  Debug::printScannedNetworksData("Available networks (sorted)");
};

int Network::getEncryptionType(int type) {
  switch (type) {
    case ENC_TYPE_WEP:
      return Network::ENCRYPTION_TYPE_WEP;
    case ENC_TYPE_TKIP:
      return Network::ENCRYPTION_TYPE_WPA;
    case ENC_TYPE_CCMP:
      return Network::ENCRYPTION_TYPE_WPA2;
    case ENC_TYPE_NONE:
      return Network::ENCRYPTION_TYPE_NONE;
    case ENC_TYPE_AUTO:
      return Network::ENCRYPTION_TYPE_AUTO;
  }
  return Network::ENCRYPTION_TYPE_ERROR;
}

int Network::getNetworkPos(String ssid) {
  for (int i = 0; i < NETWORKS_CNT; i++) {
    if (NETWORKS[i].ssid == ssid) {
      return i;
    }
  }
  return -1;
}
