import {alertController, loadingController, toastController} from '@ionic/vue';
import {ref} from "vue";

export interface NetworkInterface {
  ssid: string;
  password: string;
}

export default abstract class Base {
  private static readonly loadingEnabled = ref<boolean>(false);
  private static readonly network = ref<NetworkInterface>({ssid: '', password: ''});

  public static openLoading(message: string) {
    loadingController.create({message}).then((loading) => {
      loading.present().then(() => {
        this.loadingEnabled.value = true;
      });
    });
  }

  public static closeLoading(it?: number) {
    it = it ? it + 1 : 1;
    if (this.loadingEnabled.value || it >= 50) {
      this.loadingEnabled.value = false;
      loadingController.dismiss().then();
    } else {
      setTimeout(() => {
        this.closeLoading(it);
      }, 100);
    }
  }

  public static openAlert(header: string, message: string) {
    alertController.create({header, message, buttons: [{text: 'OK'}]}).then((alert) => {
      alert.present().then();
    });
  }

  public static openSuccessAlert(header: string, message: string) {
    alertController.create({cssClass: 'color-success', header, message, buttons: [{text: 'OK'}]}).then((alert) => {
      alert.present().then();
    });
  }

  public static openDangerAlert(header: string, message: string) {
    alertController.create({cssClass: 'color-danger', header, message, buttons: [{text: 'OK'}]}).then((alert) => {
      alert.present().then();
    });
  }

  public static openToast(message: string) {
    toastController.create({message, duration: 2500}).then((toast) => {
      toast.present().then();
    });
  }

  public static getNetwork(): NetworkInterface {
    return this.network.value;
  }

  public static setNetwork(ssid: string, password: string) {
    this.network.value = {ssid, password};
  }

  public static printDate(datetime: string) {
    const date = new Date(datetime);
    return date.getDate() + '. ' + (date.getMonth() + 1) + '. ' + date.getFullYear();
  }

  public static printTime(datetime: string) {
    const date = new Date(datetime);
    return this.appendZero(date.getHours()) + ':' + this.appendZero(date.getMinutes()) + ':' + this.appendZero(date.getSeconds());
  }

  public static printShortTime(datetime: string) {
    const date = new Date(datetime);
    return this.appendZero(date.getHours()) + ':' + this.appendZero(date.getMinutes());
  }

  public static printDatetime(datetime: string) {
    return this.printDate(datetime) + ' ' + this.printTime(datetime);
  }

  public static printDateFromTimestamp(timestamp: number) {
    const date = new Date(timestamp * 1000);
    return date.getDate() + '. ' + (date.getMonth() + 1) + '. ' + date.getFullYear();
  }

  public static printTimeFromTimestamp(timestamp: number) {
    const date = new Date(timestamp * 1000);
    return this.appendZero(date.getHours()) + ':' + this.appendZero(date.getMinutes()) + ':' + this.appendZero(date.getSeconds());
  }

  public static printDatetimeFromTimestamp(timestamp: number) {
    return this.printDateFromTimestamp(timestamp) + ' ' + this.printTimeFromTimestamp(timestamp);
  }

  public static printDatetimeInterval(startDatetime: string, endDatetime: string, unitsCnt: number) {
    const startDate = startDatetime === 'NOW' ? Date.now() : Date.parse(startDatetime);
    const endDate = endDatetime === 'NOW' ? Date.now() : Date.parse(endDatetime);
    const millis = endDate - startDate;
    return this.printDatetimeIntervalFromTimestamp(Math.floor(millis / 1000), unitsCnt);
  }

  public static printDatetimeIntervalFromTimestamp(timestamp: number, unitsCnt: number) {
    let seconds = timestamp;
    if (seconds < 10) {
      return '< 10 s';
    }
    let result = '';
    const d = Math.floor(seconds / 86400);
    result += d ? d + ' d ' : '';
    if (unitsCnt == 1 && d) {
      return result.trim();
    }
    seconds %= 86400;
    const h = Math.floor(seconds / 3600);
    result += d || h ? h + ' h ' : '';
    if ((unitsCnt == 1 && h) || (unitsCnt == 2 && d)) {
      return result.trim();
    }
    seconds %= 3600;
    const m = Math.floor(seconds / 60);
    result += d || h || m ? m + ' m ' : '';
    if ((unitsCnt == 1 && m) || (unitsCnt == 2 && h) || (unitsCnt == 3 && d)) {
      return result.trim();
    }
    seconds %= 60;
    const s = Math.floor(seconds);
    result += s + ' s';
    return result;
  }

  public static roundNumber(value: number, step: number, fixed: number): string {
    const inv = 1 / step;
    return (Math.round(value * inv) / inv).toFixed(fixed);
  }

  public static formatNumber(value: number | string) {
    return value.toString().replace(".", ",")
  }

  public static printEncryptionType(value: number) {
    switch (value) {
      case 0:
        return 'volná';

      case 1:
        return 'WEP';

      case 2:
        return 'WPA';

      case 3:
        return 'WPA2';

      case 4:
        return 'auto';
    }
    return '';
  }

  private static appendZero(value: number) {
    return value < 10 ? '0' + value : value;
  }
}
