import {ref} from 'vue';
import {Plugins} from '@capacitor/core';
import {AvailableResult} from 'capacitor-native-biometric';
import DeviceInfo from '@/services/DeviceInfo';

const {NativeBiometric, Storage} = Plugins;
const STORAGE_VARIABLE = 'fingerPrintEnabled';

export default abstract class FingerPrint {
  private static readonly enabled = ref<boolean>(false);

  public static async initialize() {
    await this.load();
  }

  public static isEnabled(): boolean {
    return this.enabled.value;
  }

  public static async isAvailable(): Promise<boolean> {
    if (DeviceInfo.isPlatformAndroid()) {
      const result: AvailableResult = await NativeBiometric.isAvailable();
      return result.isAvailable;
    }
    return false;
  }

  public static async toggle(value: boolean): Promise<boolean> {
    if (value) {
      if (await this.isAvailable()) {
        try {
          await NativeBiometric.verifyIdentity({
            reason: 'Pro jednoduché přihlášení',
            title: 'Ověření',
            description: 'Přiložte prst na čtečku otisků.',
            negativeButtonText: 'Zrušit',
          });
          this.enabled.value = true;
          await this.save();
          return true;
        } catch (error) {
          this.enabled.value = false;
          await this.save();
          return false;
        }
      }
    }
    this.enabled.value = false;
    await this.save();
    return false;
  }

  public static async verify(): Promise<boolean> {
    if (await this.isAvailable()) {
      try {
        await NativeBiometric.verifyIdentity({
          reason: 'Pro jednoduché přihlášení',
          title: 'Ověření',
          description: 'Přiložte prst na čtečku otisků.',
          negativeButtonText: 'Zrušit',
        });
        return true;
      } catch (error) {
        return false;
      }
    }
    return false;
  }

  private static async load() {
    const ret = await Storage.get({key: STORAGE_VARIABLE});
    if (ret.value) {
      this.enabled.value = JSON.parse(ret.value);
    }
  }

  private static async save() {
    await Storage.set({key: STORAGE_VARIABLE, value: JSON.stringify(this.enabled.value)});
  }
}
