import {ref} from 'vue';
import {Plugins} from '@capacitor/core';

const {Storage} = Plugins;

export const MODULE_STATUS_ERROR = -2;
export const MODULE_STATUS_FAIL = -1;
export const MODULE_STATUS_LOADING = 0;
export const MODULE_STATUS_OK = 1;
export const MODULE_STATUS_WAITING = 2;
export const MODULE_STATUS_CONFIRM = 3;

export interface ModuleInterface {
  deviceId: string;
  controlToken: string;
  statusToken: string;
  name: string;
  status: number;
  isOnline: boolean;
  lastAccessTime: string;
  buttonMode: boolean;
  toggle: {
    enabled: boolean;
    value: boolean;
  };
  temperatures: number[];
  conditionsEnabled: boolean;
  scheduleEnabled: boolean;
  notificationsEnabled: boolean;
}

export default abstract class Module {
  private static readonly modules = ref<ModuleInterface[]>([]);
  private static readonly module = ref<ModuleInterface[]>([]);

  public static async initialize() {
    await this.load();
  }

  public static getSelected(): ModuleInterface[] {
    return this.module.value;
  }

  public static setSelected(index: number) {
    this.module.value.push(this.modules.value[index]);
  }

  public static getAll(): ModuleInterface[] {
    return this.modules.value;
  }

  public static countAll() {
    return this.modules.value.length;
  }

  public static get(index: number): ModuleInterface {
    return this.modules.value[index];
  }

  public static getStatusTokens(startIndex: number, length: number): string[] {
    const tokens = [];
    const endIndex = startIndex + length;
    for (let i = startIndex; i < endIndex; i++) {
      const token = this.modules.value[i].statusToken;
      if (token) {
        tokens.push(token);
      }
    }
    return tokens;
  }

  public static async add(deviceId: string, controlToken: string, statusToken: string) {
    if (this.getIndexByDeviceId(deviceId) === -1) {
      this.modules.value.push({
        deviceId: deviceId,
        controlToken: controlToken,
        statusToken: statusToken,
        name: 'Nový modul',
        status: MODULE_STATUS_LOADING,
        isOnline: false,
        lastAccessTime: '',
        buttonMode: false,
        toggle: {enabled: false, value: false},
        temperatures: [],
        conditionsEnabled: false,
        scheduleEnabled: false,
        notificationsEnabled: false,
      });
      await this.save();
      return true;
    }
    return false;
  }

  public static async rename(index: number, name: string) {
    this.modules.value[index].name = name;
    await this.save();
  }

  public static async delete(index: number) {
    this.modules.value.splice(index, 1);
    await this.save();
  }

  public static toggle(index: number, value: boolean) {
    if (this.modules.value[index].toggle.value !== value) {
      this.modules.value[index].toggle.value = value;
    }
  }

  public static changeStatusByDeviceId(deviceId: string, status: number, isOnline: boolean, toggleValue: boolean, buttonMode?: boolean, temperatures?: number[], conditionsEnabled?: boolean, scheduleEnabled?: boolean, notificationsEnabled?: boolean) {
    if (deviceId) {
      const index = this.getIndexByDeviceId(deviceId);
      const module = this.modules.value[index];
      if (module.deviceId === deviceId) {
        if (module.status !== status) {
          module.status = status;
        }
        if (module.isOnline !== isOnline) {
          module.isOnline = isOnline;
        }
        if (!module.buttonMode) {
          this.toggle(index, toggleValue);
          this.modules.value[index].toggle.enabled = true;
        }
        if (temperatures) {
          module.temperatures = temperatures;
        }
        if (buttonMode === true || buttonMode === false) {
          module.buttonMode = buttonMode;
        }
        if (conditionsEnabled === true || conditionsEnabled === false) {
          module.conditionsEnabled = conditionsEnabled;
        }
        if (scheduleEnabled === true || scheduleEnabled === false) {
          module.scheduleEnabled = scheduleEnabled;
        }
        if (notificationsEnabled === true || notificationsEnabled === false) {
          module.notificationsEnabled = notificationsEnabled;
        }
        return true;
      }
    }
    return false;
  }

  public static setLoadingStatus(index: number) {
    const module = this.modules.value[index];
    module.status = MODULE_STATUS_LOADING;
  }

  public static setLoadingStatusToAll() {
    for (const module of this.modules.value) {
      module.status = MODULE_STATUS_LOADING;
    }
  }

  public static isAllModuleStatusOk() {
    for (const module of this.modules.value) {
      if (module.status !== MODULE_STATUS_OK) {
        return false;
      }
    }
    return true;
  }

  public static async saveReordered(reorderedModules: ModuleInterface[]) {
    this.modules.value = reorderedModules;
    await this.save();
  }

  private static getIndexByDeviceId(deviceId: string): number {
    return this.modules.value.findIndex(x => x.deviceId === deviceId);
  }

  private static async load() {
    const ret = await Storage.get({key: 'modules'});
    this.modules.value = [];
    if (ret.value) {
      const raw = JSON.parse(ret.value);
      if (Array.isArray(raw)) {
        for (const module of raw) {
          this.modules.value.push({
            deviceId: module.deviceId,
            controlToken: module.controlToken,
            statusToken: module.statusToken,
            name: module.name,
            status: MODULE_STATUS_LOADING,
            isOnline: true,
            lastAccessTime: '',
            buttonMode: false,
            toggle: {enabled: false, value: false},
            temperatures: [],
            conditionsEnabled: false,
            scheduleEnabled: false,
            notificationsEnabled: false,
          });
        }
      }
    }
  }

  private static async save() {
    const raw: any = [];
    for (const module of this.modules.value) {
      raw.push({
        deviceId: module.deviceId,
        controlToken: module.controlToken,
        statusToken: module.statusToken,
        name: module.name,
      });
    }
    await Storage.set({key: 'modules', value: JSON.stringify(raw)});
  }
}
