import '@capacitor-community/http';
import {MODULE_STATUS_CONFIRM, MODULE_STATUS_FAIL, MODULE_STATUS_OK, MODULE_STATUS_WAITING} from '@/services/Module';
import DeviceInfo from '@/services/DeviceInfo';
import {Plugins} from '@capacitor/core';

const REST_API_BASE_URL = 'https://sm.plix.cz/api/v1';

const {Http} = Plugins;

export interface StatusesResponse {
  deviceId: string;
  isOnline: boolean;
  status: number;
  buttonMode: boolean;
  toggleValue: boolean;
  lastAccessTime: boolean;
  temperatures: number[];
  conditionsEnabled: boolean;
  scheduleEnabled: boolean;
  notificationsEnabled: boolean;
}

export interface ModuleResponse {
  deviceId: string;
  controlToken: string;
  statusToken: string;
  isOnline: boolean;
  status: number;
  buttonMode: boolean;
  toggleValue: boolean;
  lastAccessTime: boolean;
  relayState: boolean;
  newRelayState: boolean;
  ssid: string;
  rssi: number;
  rssiLevel: number;
  rssiPercentage: number;
  temperatures: number[];
  conditionsEnabled: boolean;
  scheduleEnabled: boolean;
  notificationsEnabled: boolean;
  accessLog: AccessLogResponse[];
  outputLog: OutputLogResponse[];
  inputLog: InputLogResponse[];
}

export interface StatusResponse {
  status: number;
}

export interface ProcessResponse {
  process: boolean;
}

export interface RelayStateResponse {
  deviceId: string;
  isOnline: boolean;
  status: number;
  toggleValue: boolean;
  lastAccessTime: boolean;
}

export interface NetworkResponse {
  ssid: string;
  password: string;
}

export interface NetworksResponse {
  status: number;
  networks: NetworkResponse[];
}

export interface AvailableNetworkResponse {
  ssid: string;
  rssi: number;
  rssiLevel: number;
  rssiPercentage: number;
  bssid: string;
  channel: number;
  encryptionType: number;
}

export interface AvailableNetworksResponse {
  status: number;
  networks: AvailableNetworkResponse[];
}

export interface SettingsResponse {
  status: number;
  buttonMode: boolean;
  defaultState: boolean;
  stateDelay: number;
  loopDelay: number;
  defaultLed1?: number;
  defaultLed2?: number;
  defaultLed3?: number;
}

export interface ConfigurationResponse {
  status: number;
  serverPassword: string;
  pinRelay?: boolean;
  pinButton?: boolean;
  pinLed?: boolean;
  pinExtButton?: boolean;
  pinExtLed?: boolean;
  pinLed1?: boolean;
  pinLed2?: boolean;
  pinLed3?: boolean;
  pinTemperature?: boolean;
  tempAddresses: string[];
}

export interface FirmwareUpdateResponse {
  status: number;
  currentVersion: string;
  availableVersion: string;
}

export interface AutoControlResponse {
  conditionsEnabled: boolean;
  conditionsThreshold: number;
  conditionsData: any;
}

export interface ScheduleControlResponse {
  scheduleEnabled: boolean;
  scheduleData: {
    days: string[];
    time: string;
    dur: number;
  }[];
}

export interface NotificationSettingsResponse {
  notificationsEnabled: boolean;
  notificationsMessage: number;
  notificationsData: any;
}

export interface AccessLogResponse {
  isOnline: boolean;
  startDatetime: string;
  endDatetime: string;
  ssid: string;
}

export interface OutputLogResponse {
  isOnline: boolean;
  startDatetime: string;
  endDatetime: string;
  state: number;
}

export interface InputLogResponse {
  temperatures: number[];
  datetime: string;
}

export default abstract class RestApi {
  public static async updateFcm(fcmToken: string): Promise<boolean> {
    try {
      const response = await Http.request({method: 'POST', url: REST_API_BASE_URL + '/fcm', headers: this.getPostHeaders(), data: {fcmToken}});
      return response.status === 200 && response.data && response.data.status === 'OK';
    } catch (e) {
      return false;
    }
  }

  public static async getStatuses(tokens: string[]): Promise<StatusesResponse[]> {
    const entities: StatusesResponse[] = [];
    if (tokens.length > 0) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules', headers: this.getHeaders(), params: {token: tokens.join(',')}});
        if (response.status === 200) {
          const dataList = response.data;
          for (const data of dataList) {
            entities.push({
              deviceId: data.deviceId,
              isOnline: data.isOnline,
              status: this.convertStatus(data.status),
              buttonMode: data.buttonMode,
              toggleValue: this.convertToggleValue(data.status, data.relayState, data.newRelayState),
              lastAccessTime: data.lastAccessTime,
              temperatures: data.temperatures,
              conditionsEnabled: data.conditionsEnabled,
              scheduleEnabled: data.scheduleEnabled,
              notificationsEnabled: data.notificationsEnabled,
            });
          }
        }
      } catch (e) {
        return entities;
      }
    }
    return entities;
  }

  public static async getModule(controlToken: string): Promise<ModuleResponse | null> {
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken, headers: this.getHeaders()});
        if (response.status === 200) {
          const data = response.data;
          return {
            deviceId: data.deviceId,
            controlToken: data.controlToken,
            statusToken: data.statusToken,
            isOnline: data.isOnline,
            status: this.convertStatus(data.status),
            buttonMode: data.buttonMode,
            toggleValue: this.convertToggleValue(data.status, data.relayState, data.newRelayState),
            lastAccessTime: data.lastAccessTime,
            relayState: data.relayState,
            newRelayState: data.newRelayState,
            ssid: data.ssid,
            rssi: data.rssi,
            rssiLevel: data.rssiLevel,
            rssiPercentage: data.rssiPercentage,
            temperatures: data.temperatures,
            conditionsEnabled: data.conditionsEnabled,
            scheduleEnabled: data.scheduleEnabled,
            notificationsEnabled: data.notificationsEnabled,
            accessLog: data.accessLog,
            outputLog: data.outputLog,
            inputLog: data.inputLog,
          };
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async setRelayState(controlToken: string, newValue: boolean): Promise<RelayStateResponse | null> {
    if (controlToken) {
      try {
        const response = await Http.request({method: 'POST', url: REST_API_BASE_URL + '/modules/' + controlToken + '/output', headers: this.getPostHeaders(), data: {newRelayState: newValue}});
        if (response.status === 200) {
          const data = response.data;
          return {
            deviceId: data.deviceId,
            isOnline: data.isOnline,
            status: this.convertStatus(data.status),
            toggleValue: this.convertToggleValue(data.status, data.relayState, data.newRelayState),
            lastAccessTime: data.lastAccessTime,
          }
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async getNetworks(controlToken: string): Promise<NetworksResponse | null> {
    return await this.getModuleRequest(controlToken, '/networks');
  }

  public static async setNetworks(controlToken: string, data: NetworksResponse): Promise<NetworksResponse | null> {
    return await this.postModuleRequest(controlToken, '/networks', data);
  }

  public static async getSettings(controlToken: string): Promise<SettingsResponse | null> {
    return await this.getModuleRequest(controlToken, '/settings');
  }

  public static async setSettings(controlToken: string, data: SettingsResponse): Promise<SettingsResponse | null> {
    return await this.postModuleRequest(controlToken, '/settings', data);
  }

  public static async getConfiguration(controlToken: string): Promise<ConfigurationResponse | null> {
    return await this.getModuleRequest(controlToken, '/configuration');
  }

  public static async setConfiguration(controlToken: string, data: ConfigurationResponse): Promise<ConfigurationResponse | null> {
    return await this.postModuleRequest(controlToken, '/configuration', data);
  }

  public static async getConnection(controlToken: string): Promise<StatusResponse | null> {
    return await this.getModuleRequest(controlToken, '/connection');
  }

  public static async setConnection(controlToken: string, data: NetworkResponse): Promise<StatusResponse | null> {
    return await this.postModuleRequest(controlToken, '/connection', data);
  }

  public static async getAvailableNetworks(controlToken: string): Promise<AvailableNetworksResponse | null> {
    return await this.getModuleRequest(controlToken, '/available-networks');
  }

  public static async setAvailableNetworks(controlToken: string, data: ProcessResponse): Promise<AvailableNetworksResponse | null> {
    return await this.postModuleRequest(controlToken, '/available-networks', data);
  }

  public static async getRestart(controlToken: string): Promise<StatusResponse | null> {
    return await this.getModuleRequest(controlToken, '/restart');
  }

  public static async setRestart(controlToken: string, data: ProcessResponse): Promise<StatusResponse | null> {
    return await this.postModuleRequest(controlToken, '/restart', data);
  }

  public static async getFirmwareUpdate(controlToken: string): Promise<FirmwareUpdateResponse | null> {
    return await this.getModuleRequest(controlToken, '/firmware-update');
  }

  public static async setFirmwareUpdate(controlToken: string, data: ProcessResponse): Promise<FirmwareUpdateResponse | null> {
    return await this.postModuleRequest(controlToken, '/firmware-update', data);
  }

  public static async getAutoControl(controlToken: string): Promise<AutoControlResponse | null> {
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + '/auto-control', headers: this.getHeaders()});
        if (response.status === 200) {
          const data = response.data;
          return {
            conditionsEnabled: data.conditions.enabled,
            conditionsThreshold: data.conditions.threshold,
            conditionsData: data.conditions.data,
          };
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async setAutoControl(controlToken: string, conditionsEnabled: boolean, conditionsThreshold: number, conditionsData: any): Promise<AutoControlResponse | null> {
    if (controlToken) {
      const data = {
        conditions: {
          enabled: conditionsEnabled,
          threshold: conditionsThreshold,
          data: conditionsData,
        },
      };
      try {
        const response = await Http.request({method: 'POST', url: REST_API_BASE_URL + '/modules/' + controlToken + '/auto-control', headers: this.getPostHeaders(), data: data});
        if (response.status === 200) {
          const data = response.data;
          return {
            conditionsEnabled: data.conditions.enabled,
            conditionsThreshold: data.conditions.threshold,
            conditionsData: data.conditions.data,
          };
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async getScheduleControl(controlToken: string): Promise<ScheduleControlResponse | null> {
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + '/schedule-control', headers: this.getHeaders()});
        if (response.status === 200) {
          const data = response.data;
          return {
            scheduleEnabled: data.schedule.enabled,
            scheduleData: data.schedule.data ? data.schedule.data : [],
          };
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async setScheduleControl(controlToken: string, scheduleEnabled: boolean, scheduleData: any): Promise<ScheduleControlResponse | null> {
    if (controlToken) {
      const data = {
        schedule: {
          enabled: scheduleEnabled,
          data: scheduleData,
        },
      };
      try {
        const response = await Http.request({method: 'POST', url: REST_API_BASE_URL + '/modules/' + controlToken + '/schedule-control', headers: this.getPostHeaders(), data: data});
        if (response.status === 200) {
          const data = response.data;
          return {
            scheduleEnabled: data.schedule.enabled,
            scheduleData: data.schedule.data ? data.schedule.data : [],
          };
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async getNotificationSettings(controlToken: string): Promise<NotificationSettingsResponse | null> {
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + '/notification-settings', headers: this.getHeaders()});
        if (response.status === 200) {
          const data = response.data;
          return {
            notificationsEnabled: data.notifications.enabled,
            notificationsMessage: data.notifications.message,
            notificationsData: data.notifications.data,
          };
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async setNotificationSettings(controlToken: string, notificationsEnabled: boolean, notificationsMessage: string, notificationsData: any): Promise<NotificationSettingsResponse | null> {
    if (controlToken) {
      const data = {
        notifications: {
          enabled: notificationsEnabled,
          message: notificationsMessage,
          data: notificationsData,
        },
      };
      try {
        const response = await Http.request({method: 'POST', url: REST_API_BASE_URL + '/modules/' + controlToken + '/notification-settings', headers: this.getPostHeaders(), data: data});
        if (response.status === 200) {
          const data = response.data;
          return {
            notificationsEnabled: data.notifications.enabled,
            notificationsMessage: data.notifications.message,
            notificationsData: data.notifications.data,
          };
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  public static async testNotifications(controlToken: string): Promise<boolean> {
    if (controlToken) {
      const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + '/notification-test', headers: this.getHeaders()});
      try {
        if (response.status === 200) {
          const data = response.data;
          return data.status === 'OK';
        }
      } catch (e) {
        return false;
      }
    }
    return false;
  }

  public static async getAccessLog(controlToken: string, offsetDatetime: string): Promise<AccessLogResponse[]> {
    const entities: AccessLogResponse[] = [];
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + '/access-log', headers: this.getHeaders(), params: {offsetDatetime}});
        if (response.status === 200) {
          const dataList = response.data;
          for (const data of dataList) {
            entities.push({
              isOnline: data.isOnline,
              startDatetime: data.startDatetime,
              endDatetime: data.endDatetime,
              ssid: data.ssid,
            });
          }
        }
      } catch (e) {
        return entities;
      }
    }
    return entities;
  }

  public static async getOutputLog(controlToken: string, offsetDatetime: string): Promise<OutputLogResponse[]> {
    const entities: OutputLogResponse[] = [];
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + '/output-log', headers: this.getHeaders(), params: {offsetDatetime}});
        if (response.status === 200) {
          const dataList = response.data;
          for (const data of dataList) {
            entities.push({
              isOnline: data.isOnline,
              startDatetime: data.startDatetime,
              endDatetime: data.endDatetime,
              state: data.state,
            });
          }
        }
      } catch (e) {
        return entities;
      }
    }
    return entities;
  }

  public static async getInputLog(controlToken: string, offsetDatetime: string): Promise<InputLogResponse[]> {
    const entities: InputLogResponse[] = [];
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + '/input-log', headers: this.getHeaders(), params: {offsetDatetime}});
        if (response.status === 200) {
          const dataList = response.data;
          for (const data of dataList) {
            entities.push({
              temperatures: data.temperatures,
              datetime: data.datetime,
            });
          }
        }
      } catch (e) {
        return entities;
      }
    }
    return entities;
  }

  public static async getServerStatus(): Promise<any> {
    try {
      const data = {'api_key': 'm786265921-c028a379f6608f6ed14582a6', 'all_time_uptime_ratio': '1', 'logs': '1'};
      const response = await Http.request({method: 'POST', url: 'https://api.uptimerobot.com/v2/getMonitors', headers: {'Content-Type': 'application/x-www-form-urlencoded'}, data});
      if (response.status === 200) {
        if (response.data && response.data.stat === 'ok' && response.data.monitors && response.data.monitors[0]) {
          return response.data.monitors[0];
        }
      }
    } catch (e) {
      return null;
    }
    return null;
  }

  public static async testOfflineMode(): Promise<boolean> {
    try {
      const params = {test: 'plix_smart_module_1234'};
      const response = await Http.request({method: 'GET', url: 'http://192.168.0.1/plix-sm-config', params});
      console.log(response);
      if (response.status === 200) {
        const data: any = response.data;
        console.log(data);
        return data && data.status === 'OK' && data.test === 'plix_smart_module_1234';
      }
    } catch (e) {
      console.log(e);
      return false;
    }
    return false;
  }

  public static async setNetworkOfflineMode(ssid: string, password: string): Promise<boolean> {
    try {
      const params = {ssid: ssid, pw: password};
      const response = await Http.request({method: 'GET', url: 'http://192.168.0.1/plix-sm-config', params});
      if (response.status === 200) {
        const data: any = response.data;
        console.log(data);
        return data && data.status === 'OK';
      }
    } catch (e) {
      return false;
    }
    return false;
  }

  private static async getModuleRequest(controlToken: string, path: string): Promise<any> {
    if (controlToken) {
      try {
        const response = await Http.request({method: 'GET', url: REST_API_BASE_URL + '/modules/' + controlToken + path, headers: this.getHeaders()});
        if (response.status === 200) {
          const data = response.data;
          data.status = this.convertStatus(response.data.status);
          console.log(data);
          return data;
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  private static async postModuleRequest(controlToken: string, path: string, data: any): Promise<any> {
    if (controlToken) {
      try {
        delete data.status;
        const response = await Http.request({method: 'POST', url: REST_API_BASE_URL + '/modules/' + controlToken + path, headers: this.getPostHeaders(), data});
        if (response.status === 200) {
          const data = response.data;
          data.status = this.convertStatus(response.data.status);
          return data;
        }
      } catch (e) {
        return null;
      }
    }
    return null;
  }

  private static getPostHeaders() {
    return {
      'Content-Type': 'application/json',
      'X-Platform': DeviceInfo.getPlatform(),
      'X-Version': DeviceInfo.getAppVersion(),
      'X-Device-Id': DeviceInfo.getUuid(),
    }
  }

  private static getHeaders() {
    return {
      'X-Platform': DeviceInfo.getPlatform(),
      'X-Version': DeviceInfo.getAppVersion(),
      'X-Device-Id': DeviceInfo.getUuid(),
    }
  }

  private static convertStatus(status: number): number {
    if (status === 0) {
      return MODULE_STATUS_OK;
    } else if (status === 1) {
      return MODULE_STATUS_WAITING;
    } else if (status === 2) {
      return MODULE_STATUS_CONFIRM;
    } else {
      return MODULE_STATUS_FAIL;
    }
  }

  private static convertToggleValue(status: number, relayState: boolean, newRelayState: boolean): boolean {
    if (status === 0) {
      return relayState;
    } else if (status === 1) {
      return newRelayState;
    } else if (status === 2) {
      return newRelayState;
    } else {
      return relayState;
    }
  }
}
