<?php

namespace Drupal\ban_ip\Service;

use Drupal\ban\BanIpManager;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DrupalDateTime;
use Exception;

/**
 * Class BanManager.
 */
class BanManager {

  /**
   * Database connection service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  public Connection $connection;

  /**
   * Config service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected ConfigFactoryInterface $config;

  /**
   * Constructor.
   */
  public function __construct(Connection $connection, ConfigFactoryInterface $config) {
    $this->connection = $connection;
    $this->config = $config;
  }

  /**
   * Bans passed IP address.
   */
  public function banIp($ip_address) {
    if (!$ip_address) {
      return;
    }

    $ban_manager = new BanIpManager($this->connection);
    $current_timestamp = (new DrupalDateTime('NOW'))->getTimestamp();
    $settings = $this->config->get('ban_ip.settings');

    try {
      // Add the IP address to the log.
      $fields = ['time' => $current_timestamp, 'ip' => $ip_address];
      $this->connection->insert('ban_ip_log')->fields($fields)->execute();
    }
    catch (Exception $e) {
      // Ban the IP address if database query failed.
      $ban_manager->banIp($ip_address);
    }

    // Ban the IP address if number of log entries exceeds the threshold value.
    $measurement_interval = $current_timestamp - $settings->get('measurement_interval');
    $query = $this->connection->select('ban_ip_log', 't');
    $query->condition('t.ip', $ip_address);
    $query->condition('time', $measurement_interval, '>=');
    $query->addExpression('COUNT(*)');
    $count = $query->execute()->fetchField();
    if ($count >= $settings->get('threshold')) {
      $ban_manager->banIp($ip_address);
    }
  }

  /**
   * Unbans all expired IP addresses.
   */
  public function unbanExpiredIps() {
    $ban_manager = new BanIpManager($this->connection);
    $current_timestamp = (new DrupalDateTime('NOW'))->getTimestamp();
    $settings = $this->config->get('ban_ip.settings');

    // Delete all expired IP addresses from the log.
    $duration_time = $current_timestamp - $settings->get('duration_time');
    $this->connection->delete('ban_ip_log')->condition('time', $duration_time, '<')->execute();

    // Get all IP addresses from the log.
    $ip_addresses = [];
    $query = $this->connection->select('ban_ip_log', 't');
    $query->addField('t', 'ip');
    $query->groupBy('t.ip');
    $results = $query->execute();
    foreach ($results as $result) {
      if ($ip_address = $result->ip) {
        $ip_addresses[] = $ip_address;
      }
    }

    // Unban all IP addresses which are not in the log.
    $results = $ban_manager->findAll();
    foreach ($results as $result) {
      if ($ip_address = $result->ip) {
        if (!in_array($ip_address, $ip_addresses)) {
          $ban_manager->unbanIp($ip_address);
        }
      }
    }
  }

}
