<?php

namespace Drupal\sm\Controller;

use Drupal\ban_ip\Service\BanManager;
use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\node\NodeInterface;
use Drupal\sm\Object\ModuleLastValues;
use Drupal\sm\Service\Data\AppData;
use Drupal\sm\Service\Data\ModuleNodeData;
use Drupal\sm\Service\Database;
use Drupal\sm\Service\PushNotification;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Class AppController.
 */
class AppController extends ControllerBase {

  /**
   * Database service.
   *
   * @var \Drupal\sm\Service\Database
   */
  protected Database $database;

  /**
   * Module node data service.
   *
   * @var \Drupal\sm\Service\Data\ModuleNodeData
   */
  protected ModuleNodeData $moduleNodeData;

  /**
   * Ban manager.
   *
   * @var \Drupal\ban_ip\Service\BanManager
   */
  protected BanManager $banManager;

  /**
   * Push notification service.
   *
   * @var \Drupal\sm\Service\PushNotification
   */
  protected PushNotification $pushNotification;

  /**
   * App data service.
   *
   * @var \Drupal\sm\Service\Data\AppData
   */
  protected AppData $appData;

  /**
   * Constructor.
   */
  public function __construct(Database $database, ModuleNodeData $module_node_data, BanManager $ban_manager, PushNotification $push_notification, AppData $app_data) {
    $this->database = $database;
    $this->moduleNodeData = $module_node_data;
    $this->banManager = $ban_manager;
    $this->pushNotification = $push_notification;
    $this->appData = $app_data;
  }

  /**
   * Static create function.
   */
  public static function create(ContainerInterface $container) {
    /** @var \Drupal\sm\Service\Database $database */
    $database = $container->get('sm.database');
    /** @var \Drupal\sm\Service\Data\ModuleNodeData $module_node_data */
    $module_node_data = $container->get('sm.module_node_data');
    /** @var \Drupal\ban_ip\Service\BanManager $ban_manager */
    $ban_manager = $container->get('ban_ip.ban_manager');
    /** @var \Drupal\sm\Service\PushNotification $push_notification */
    $push_notification = $container->get('sm.push_notification');
    /** @var \Drupal\sm\Service\Data\AppData $app_data */
    $app_data = $container->get('sm.app_data');
    return new static($database, $module_node_data, $ban_manager, $push_notification, $app_data);
  }

  /**
   * Updates firebase cloud messaging token and notifications data.
   */
  public function updateFcm(Request $request) {
    $platform_id = $this->appData->getPlatformId(Xss::filter($request->headers->get('X-Platform')));
    $app_version = Xss::filter($request->headers->get('X-Version'));
    $device_uuid = Xss::filter($request->headers->get('X-Device-Id'));
    $data = Json::decode($request->getContent());
    $fcm_token = Xss::filter($data['fcmToken']);
    if ($platform_id && $app_version && $device_uuid && $fcm_token) {
      $current_datetime = new DrupalDateTime('NOW');
      $query = $this->database->connection->select('sm_app', 't');
      $query->condition('t.uuid', $device_uuid);
      $query->addField('t', 'id');
      $result = $query->execute()->fetchObject();
      $fields = [];
      $fields['platform'] = $platform_id;
      $fields['version'] = $app_version;
      $fields['fcm_token'] = $fcm_token;
      $fields['access_time'] = $current_datetime->getTimestamp();
      if ($result) {
        $this->database->connection->update('sm_app')->fields($fields)->condition('uuid', $device_uuid)->execute();
      }
      else {
        $fields['uuid'] = $device_uuid;
        $this->database->connection->insert('sm_app')->fields($fields)->execute();
      }
      return new JsonResponse(['status' => 'OK']);
    }
    $this->banManager->banIp($request->getClientIp());
    return new JsonResponse([
      'errors' => [
        [
          'id' => 2000,
          'status' => 401,
          'code' => 'WRONG_PARAMETERS',
          'title' => 'Unauthorized',
          'detail' => 'Wrong request parameters.',
        ],
      ],
    ], 401);
  }

  /**
   * Returns statuses of modules.
   */
  public function getModuleStatuses(Request $request) {
    $tokens_raw = Xss::filter($request->query->get('token'));
    $tokens = explode(',', $tokens_raw);
    if (!$tokens) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2010,
            'status' => 401,
            'code' => 'STATUS_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter STATUS_TOKEN.',
          ],
        ],
      ], 401);
    }

    $current_datetime = new DrupalDateTime('NOW');
    $query = $this->database->nodeStorage->getQuery();
    $query->condition('type', 'module');
    $query->condition('field_status_token', $tokens, 'IN');
    $query->groupBy('nid');
    $module_ids = $query->execute();
    $module_nodes = $module_ids ? $this->database->nodeStorage->loadMultiple($module_ids) : NULL;

    $response = [];
    foreach ($module_nodes as $module_node) {
      $last_values = $this->moduleNodeData->getLastValues($module_node);
      $response[] = self::getModuleBaseData($module_node, $current_datetime, $last_values);;
    }

    $this->database->connection->update('sm_last_values')->fields(['ctrl_time' => $current_datetime->getTimestamp()])->condition('nid', $module_ids, 'IN')->execute();

    $device_uuid = Xss::filter($request->headers->get('X-Device-Id'));
    self::updateModuleAppCombinations($device_uuid, $module_ids);

    return new JsonResponse($response);
  }

  /**
   * Returns module.
   */
  public function getModule(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2020,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $current_datetime = new DrupalDateTime('NOW');
    $last_values = $this->moduleNodeData->getLastValues($module_node);
    $response = self::getModuleBaseData($module_node, $current_datetime, $last_values);
    $response['controlToken'] = $module_node->get('field_control_token')->value;
    $response['statusToken'] = $module_node->get('field_status_token')->value;

    $response['ssid'] = $last_values->ssid;
    $response['rssi'] = $last_values->rssi;
    $response['rssiLevel'] = $this->moduleNodeData->getRssiLevel($last_values->rssi);
    $response['rssiPercentage'] = $this->moduleNodeData->getRssiPercentage($last_values->rssi);

    $response['accessLog'] = [];
    foreach ($this->moduleNodeData->getAccessLog($module_node, $current_datetime, 5) as $item) {
      $response['accessLog'][] = [
        'isOnline' => $item->isOnline,
        'startDatetime' => $item->startDatetime ? $item->startDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'endDatetime' => $item->endDatetime ? $item->endDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'ssid' => $item->ssid,
      ];
    }

    $response['outputLog'] = [];
    foreach ($this->moduleNodeData->getRelayStateLog($module_node, $current_datetime, 10) as $item) {
      $response['outputLog'][] = [
        'isOnline' => $item->isOnline,
        'startDatetime' => $item->startDatetime ? $item->startDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'endDatetime' => $item->endDatetime ? $item->endDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'state' => $item->state,
      ];
    }

    $response['inputLog'] = [];
    foreach ($this->moduleNodeData->getTemperatureLog($module_node, $current_datetime, 12) as $item) {
      $response['inputLog'][] = [
        'temperatures' => $item->temperatures,
        'datetime' => $item->datetime ? $item->datetime->format(SM_DATETIME_API_FORMAT) : NULL,
      ];
    }

    $this->database->connection->update('sm_last_values')->fields(['ctrl_time' => $current_datetime->getTimestamp()])->condition('nid', $module_node->id())->execute();
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) module output values.
   */
  public function output(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2030,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }

    $content = $request->getContent();
    $params = $content ? Json::decode($content) : [];
    if (!is_array($params) || !isset($params['newRelayState'])) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2031,
            'status' => 401,
            'code' => 'WRONG_REQUEST_DATA',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request data.',
          ],
        ],
      ], 401);
    }
    $current_datetime = new DrupalDateTime('NOW');

    $response = [];
    $response['deviceId'] = $module_node->get('field_device_id')->value;
    $new_relay_state = !!Xss::filter($params['newRelayState']);
    $this->moduleNodeData->setNewValues($module_node, $new_relay_state);
    $response['newRelayState'] = $new_relay_state;

    $last_values = $this->moduleNodeData->getLastValues($module_node);
    $response['isOnline'] = $this->moduleNodeData->isOnline($last_values->accessDatetime, $current_datetime);
    $response['relayState'] = $last_values->relay;
    $response['lastAccessTime'] = $last_values->accessDatetime ? $last_values->accessDatetime->format(SM_DATETIME_API_FORMAT) : NULL;
    $response['status'] = $this->moduleNodeData->getStatuses($module_node)->pinChange;

    $this->database->connection->update('sm_last_values')->fields(['ctrl_time' => $current_datetime->getTimestamp()])->condition('nid', $module_node->id())->execute();

    Cache::invalidateTags($module_node->getCacheTags());
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) saved networks.
   */
  public function networks(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2040,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $status = $this->moduleNodeData->getStatuses($module_node)->netSave;
    $network_save_p_values = $module_node->get('field_network_save')->getValue();
    $data = Json::decode($request->getContent());
    if ($data && isset($data['networks'])) {
      $networks = $data['networks'];
      if (count($networks) > SM_MAX_NETWORK_CNT) {
        return new JsonResponse([
          'errors' => [
            [
              'id' => 2041,
              'status' => 400,
              'code' => 'NETWORKS_LENGTH_ERROR',
              'title' => 'Unauthorized',
              'detail' => 'Wrong networks length.',
            ],
          ],
        ], 400);
      }
      for ($i = 0; $i < SM_MAX_NETWORK_CNT; $i++) {
        if ($network_save_p = $this->database->paragraphStorage->load($network_save_p_values[$i]['target_id'])) {
          $data = isset($networks[$i]) ? $networks[$i] : [];
          self::setStringValueToEntityField($network_save_p, 'field_ssid', $data, 'ssid', TRUE);
          self::setStringValueToEntityField($network_save_p, 'field_password', $data, 'password', TRUE);
          try {
            $network_save_p->save();
            $status = SM_STATUS_PROCESSING;
          }
          catch (EntityStorageException $e) {
          }
        }
      }
      if ($status) {
        $this->database->connection->update('sm_status')->fields(['net_save' => $status])->condition('nid', $module_node->id())->execute();
      }
    }
    $response = [];
    $response['status'] = $status;
    $response['networks'] = [];
    foreach ($network_save_p_values as $i => $network_save_p_value) {
      if ($network_save_p = $this->database->paragraphStorage->load($network_save_p_value['target_id'])) {
        $ssid = $network_save_p->get('field_ssid')->value;
        if ($ssid) {
          $response['networks'][] = [
            'ssid' => $ssid,
            'password' => $network_save_p->get('field_password')->value,
          ];
        }
      }
    }
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) module settings.
   */
  public function settings(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2050,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $status = $this->moduleNodeData->getStatuses($module_node)->stgSave;
    $settings_p_values = $module_node->get('field_settings')->getValue();
    $data = Json::decode($request->getContent());
    if ($data) {
      if (isset($settings_p_values[0]['target_id'])) {
        if ($settings_p = $this->database->paragraphStorage->load($settings_p_values[0]['target_id'])) {
          $data['stateDelay'] = isset($data['stateDelay']) ? Xss::filter($data['stateDelay']) * 1000 : NULL;
          $data['loopDelay'] = isset($data['loopDelay']) ? Xss::filter($data['loopDelay']) * 1000 : NULL;
          self::setBooleanValueToEntityField($settings_p, 'field_button_mode', $data, 'buttonMode');
          self::setIntegerValueToEntityField($settings_p, 'field_state_delay', $data, 'stateDelay');
          self::setIntegerValueToEntityField($settings_p, 'field_loop_delay', $data, 'loopDelay');
          self::setBooleanValueToEntityField($settings_p, 'field_default_state', $data, 'defaultState');
          try {
            $settings_p->save();
            $status = SM_STATUS_PROCESSING;
          }
          catch (EntityStorageException $e) {
          }
        }
      }
      if ($status) {
        $this->database->connection->update('sm_status')->fields(['stg_save' => $status])->condition('nid', $module_node->id())->execute();
      }
    }
    $response = [];
    $response['status'] = $status;
    if (isset($settings_p_values[0]['target_id'])) {
      if ($settings_p = $this->database->paragraphStorage->load($settings_p_values[0]['target_id'])) {
        $response['buttonMode'] = !!($settings_p->get('field_button_mode')->value);
        $response['stateDelay'] = (float) ($settings_p->get('field_state_delay')->value / 1000);
        $response['loopDelay'] = (float) ($settings_p->get('field_loop_delay')->value / 1000);
        $response['defaultState'] = !!($settings_p->get('field_default_state')->value);
      }
    }
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) module configuration.
   */
  public function configuration(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2060,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $status = $this->moduleNodeData->getStatuses($module_node)->cfgSave;
    $configuration_p_values = $module_node->get('field_configuration')->getValue();
    $data = Json::decode($request->getContent());
    if ($data) {
      if (isset($configuration_p_values[0]['target_id'])) {
        if ($configuration_p = $this->database->paragraphStorage->load($configuration_p_values[0]['target_id'])) {
          self::setStringValueToEntityField($configuration_p, 'field_server_password', $data, 'serverPassword');
          self::setPinValueToEntityField($configuration_p, 'field_pin_button', $data, 'pinButton');
          self::setPinValueToEntityField($configuration_p, 'field_pin_relay', $data, 'pinRelay');
          self::setPinValueToEntityField($configuration_p, 'field_pin_led', $data, 'pinLed');
          self::setPinValueToEntityField($configuration_p, 'field_pin_ext_button', $data, 'pinExtButton');
          self::setPinValueToEntityField($configuration_p, 'field_pin_ext_led', $data, 'pinExtLed');
          self::setPinValueToEntityField($configuration_p, 'field_pin_led_1', $data, 'pinLed1');
          self::setPinValueToEntityField($configuration_p, 'field_pin_led_2', $data, 'pinLed2');
          self::setPinValueToEntityField($configuration_p, 'field_pin_led_3', $data, 'pinLed3');
          self::setPinValueToEntityField($configuration_p, 'field_pin_temp_sensors', $data, 'pinTemperature');
          try {
            $configuration_p->save();
            $status = SM_STATUS_PROCESSING;
          }
          catch (EntityStorageException $e) {
          }
        }
      }
      if ($status) {
        $this->database->connection->update('sm_status')->fields(['cfg_save' => $status])->condition('nid', $module_node->id())->execute();
      }
    }
    $response = [];
    $response['status'] = $status;
    if (isset($configuration_p_values[0]['target_id'])) {
      if ($configuration_p = $this->database->paragraphStorage->load($configuration_p_values[0]['target_id'])) {
        $response['serverPassword'] = $configuration_p->get('field_server_password')->value;
        $response['pinButton'] = self::getPinEnabled($configuration_p->get('field_pin_button')->value);
        $response['pinRelay'] = self::getPinEnabled($configuration_p->get('field_pin_relay')->value);
        $response['pinLed'] = self::getPinEnabled($configuration_p->get('field_pin_led')->value);
        $response['pinExtButton'] = self::getPinEnabled($configuration_p->get('field_pin_ext_button')->value);
        $response['pinExtLed'] = self::getPinEnabled($configuration_p->get('field_pin_ext_led')->value);
        $response['pinLed1'] = self::getPinEnabled($configuration_p->get('field_pin_led_1')->value);
        $response['pinLed2'] = self::getPinEnabled($configuration_p->get('field_pin_led_2')->value);
        $response['pinLed3'] = self::getPinEnabled($configuration_p->get('field_pin_led_3')->value);
        $response['pinTemperature'] = self::getPinEnabled($configuration_p->get('field_pin_temp_sensors')->value);
        $response['tempAddresses'] = [];
        $temp_sensor_values = $configuration_p->get('field_temp_sensors')->getValue();
        foreach ($temp_sensor_values as $temp_sensor_value) {
          if ($temp_sensor_value && $temp_sensor_value['target_id']) {
            $temp_sensor_node = $this->database->nodeStorage->load($temp_sensor_value['target_id']);
            $temp_sensor_address = $temp_sensor_node->get('field_device_address')->value;
            $response['tempAddresses'][] = $temp_sensor_address;
          }
        }
      }
    }
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) available networks.
   */
  public function availableNetworks(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2070,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $status = $this->moduleNodeData->getStatuses($module_node)->netScan;
    $data = Json::decode($request->getContent());
    if ($data && isset($data['process'])) {
      if (!!Xss::filter($data['process'])) {
        $status = SM_STATUS_PROCESSING;
      }
      if ($status) {
        $this->database->connection->update('sm_status')->fields(['net_scan' => $status])->condition('nid', $module_node->id())->execute();
      }
    }
    $response = [];
    $response['status'] = $status;
    $response['networks'] = [];
    $network_scan_p_values = $module_node->get('field_network_scan')->getValue();
    foreach ($network_scan_p_values as $i => $network_scan_p_value) {
      if ($network_scan_p = $this->database->paragraphStorage->load($network_scan_p_value['target_id'])) {
        $ssid = $network_scan_p->get('field_ssid')->value;
        $rssi = $network_scan_p->get('field_rssi')->value;
        $bssid = $network_scan_p->get('field_bssid')->value;
        if ($ssid || $bssid) {
          $response['networks'][] = [
            'ssid' => $ssid,
            'rssi' => (int) $rssi,
            'rssiLevel' => (int) $this->moduleNodeData->getRssiLevel($rssi),
            'rssiPercentage' => (int) $this->moduleNodeData->getRssiPercentage($rssi),
            'bssid' => $bssid,
            'channel' => (int) $network_scan_p->get('field_channel')->value,
            'encryptionType' => (int) $network_scan_p->get('field_encryption_type')->value,
          ];
        }
      }
    }
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) connection.
   */
  public function connection(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2080,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $status = $this->moduleNodeData->getStatuses($module_node)->connect;
    $data = Json::decode($request->getContent());
    if ($data) {
      self::setStringValueToEntityField($module_node, 'field_connect_ssid', $data, 'ssid', TRUE);
      self::setStringValueToEntityField($module_node, 'field_connect_password', $data, 'password', TRUE);
      try {
        $module_node->save();
        $status = SM_STATUS_PROCESSING;
      }
      catch (EntityStorageException $e) {
      }
      if ($status) {
        $this->database->connection->update('sm_status')->fields(['connect' => $status])->condition('nid', $module_node->id())->execute();
      }
    }
    $response = [];
    $response['status'] = $status;
    return new JsonResponse($response);
  }

  /**
   * Restarts module.
   */
  public function restart(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2090,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $status = $this->moduleNodeData->getStatuses($module_node)->restart;
    $data = Json::decode($request->getContent());
    if ($data && isset($data['process'])) {
      if (!!Xss::filter($data['process'])) {
        $status = SM_STATUS_PROCESSING;
      }
      if ($status) {
        $this->database->connection->update('sm_status')->fields(['restart' => $status])->condition('nid', $module_node->id())->execute();
      }
    }
    $response = [];
    $response['status'] = $status;
    return new JsonResponse($response);
  }

  /**
   * Updates module firmware.
   */
  public function firmwareUpdate(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2100,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $status = $this->moduleNodeData->getStatuses($module_node)->fwUpdate;
    $current_version = $this->moduleNodeData->getLastValues($module_node)->firmwareVersion;

    $query = $this->database->nodeStorage->getQuery();
    $query->condition('type', 'module_firmware');
    $query->sort('created', 'DESC');
    $result = $query->execute();
    $module_firmware_node_id = reset($result);
    $module_firmware_node = $module_firmware_node_id ? $this->database->nodeStorage->load($module_firmware_node_id) : NULL;
    $available_version = $module_firmware_node ? $module_firmware_node->get('field_version')->value : NULL;

    $data = Json::decode($request->getContent());
    if ($data && isset($data['process'])) {
      if (!!Xss::filter($data['process'])) {
        if ($current_version != $available_version) {
          $module_node->set('field_firmware', $module_firmware_node_id);
          try {
            $module_node->save();
            $status = SM_STATUS_PROCESSING;
          }
          catch (EntityStorageException $e) {
          }
        }
      }
      if ($status) {
        $this->database->connection->update('sm_status')->fields(['fw_update' => $status])->condition('nid', $module_node->id())->execute();
      }
    }
    $response = [];
    $response['status'] = $status;
    $response['currentVersion'] = $current_version;
    $response['availableVersion'] = $available_version;
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) module auto control settings.
   */
  public function autoControl(Request $request) {
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2110,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $data = Json::decode($request->getContent());
    if ($data && $data['conditions']) {
      $module_node->set('field_conditions_enabled', !!Xss::filter($data['conditions']['enabled']));
      $module_node->set('field_conditions_threshold', (float) Xss::filter($data['conditions']['threshold']));
      $module_node->set('field_conditions_json', JSON::encode($data['conditions']['data']));
      try {
        $module_node->save();
      }
      catch (EntityStorageException $e) {
        return new JsonResponse([
          'errors' => [
            [
              'id' => 2111,
              'status' => 500,
              'code' => 'SERVER_ERROR',
              'title' => 'Server error',
              'detail' => 'Database server error.',
            ],
          ],
        ], 500);
      }
    }
    $response['conditions']['enabled'] = !!($module_node->get('field_conditions_enabled')->value);
    $response['conditions']['threshold'] = (float) $module_node->get('field_conditions_threshold')->value;
    $response['conditions']['data'] = Json::decode($module_node->get('field_conditions_json')->value);
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) module schedule control settings.
   */
  public function scheduleControl(Request $request) {
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2120,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $data = Json::decode($request->getContent());
    if ($data && $data['schedule']) {
      $module_node->set('field_schedule_enabled', !!Xss::filter($data['schedule']['enabled']));
      $module_node->set('field_schedule_json', JSON::encode($data['schedule']['data']));
      try {
        $module_node->save();
      }
      catch (EntityStorageException $e) {
        return new JsonResponse([
          'errors' => [
            [
              'id' => 2121,
              'status' => 500,
              'code' => 'SERVER_ERROR',
              'title' => 'Server error',
              'detail' => 'Database server error.',
            ],
          ],
        ], 500);
      }
    }
    $response['schedule']['enabled'] = !!($module_node->get('field_schedule_enabled')->value);
    $response['schedule']['data'] = Json::decode($module_node->get('field_schedule_json')->value);
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) module notification settings.
   */
  public function notificationSettings(Request $request) {
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2130,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $data = Json::decode($request->getContent());
    if ($data && $data['notifications']) {
      $module_node->set('field_notifications_enabled', !!Xss::filter($data['notifications']['enabled']));
      $module_node->set('field_notifications_message', Xss::filter($data['notifications']['message']));
      $module_node->set('field_notifications_json', JSON::encode($data['notifications']['data']));
      try {
        $module_node->save();
      }
      catch (EntityStorageException $e) {
        return new JsonResponse([
          'errors' => [
            [
              'id' => 2131,
              'status' => 500,
              'code' => 'SERVER_ERROR',
              'title' => 'Server error',
              'detail' => 'Database server error.',
            ],
          ],
        ], 500);
      }
    }
    $response['notifications']['enabled'] = !!($module_node->get('field_notifications_enabled')->value);
    $response['notifications']['message'] = $module_node->get('field_notifications_message')->value;
    $response['notifications']['data'] = Json::decode($module_node->get('field_notifications_json')->value);
    return new JsonResponse($response);
  }

  /**
   * Returns (and updates) module auto control.
   */
  public function notificationTest(Request $request) {
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2140,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $device_uuid = Xss::filter($request->headers->get('X-Device-Id'));
    if (!$device_uuid) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2141,
            'status' => 401,
            'code' => 'WRONG_HEADER_PARAMETERS',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request header parameters.',
          ],
        ],
      ], 401);
    }
    $app_fcm_token = $this->appData->getFcmToken($device_uuid);
    if (!$app_fcm_token) {
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2142,
            'status' => 400,
            'code' => 'FCM_TOKEN_ERROR',
            'title' => 'FCM token error',
            'detail' => 'Wrong FCM token.',
          ],
        ],
      ], 400);
    }
    $result = $this->pushNotification->sendMessage($app_fcm_token, 'Test', 'Pokud se zobrazila tato zpráva, odesílání zpráv na toto zařízení funguje správně.');
    if ($result) {
      return new JsonResponse(['status' => 'OK']);
    }
    return new JsonResponse(['status' => 'FAIL']);
  }

  /**
   * Returns module access log.
   */
  public function accessLog(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2150,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $offset_datetime = Xss::filter($request->query->get('offsetDatetime'));
    if (!$offset_datetime) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2151,
            'status' => 401,
            'code' => 'WRONG_QUERY_PARAMETERS',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request query parameters.',
          ],
        ],
      ], 401);
    }
    $offset_datetime = new DrupalDateTime($offset_datetime);
    $response = [];
    foreach ($this->moduleNodeData->getAccessLog($module_node, $offset_datetime) as $item) {
      $response[] = [
        'isOnline' => $item->isOnline,
        'startDatetime' => $item->startDatetime ? $item->startDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'endDatetime' => $item->endDatetime ? $item->endDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'ssid' => $item->ssid,
      ];
    }
    return new JsonResponse($response);
  }

  /**
   * Returns module output log.
   */
  public function outputLog(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2160,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $offset_datetime = Xss::filter($request->query->get('offsetDatetime'));
    if (!$offset_datetime) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2161,
            'status' => 401,
            'code' => 'WRONG_QUERY_PARAMETERS',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request query parameters.',
          ],
        ],
      ], 401);
    }
    $offset_datetime = new DrupalDateTime($offset_datetime);
    $response = [];
    foreach ($this->moduleNodeData->getRelayStateLog($module_node, $offset_datetime) as $item) {
      $response[] = [
        'isOnline' => $item->isOnline,
        'startDatetime' => $item->startDatetime ? $item->startDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'endDatetime' => $item->endDatetime ? $item->endDatetime->format(SM_DATETIME_API_FORMAT) : NULL,
        'state' => $item->state,
      ];
    }
    return new JsonResponse($response);
  }

  /**
   * Returns module input log.
   */
  public function inputLog(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByControlToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2170,
            'status' => 401,
            'code' => 'CONTROL_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter CONTROL_TOKEN.',
          ],
        ],
      ], 401);
    }
    $offset_datetime = Xss::filter($request->query->get('offsetDatetime'));
    if (!$offset_datetime) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 2171,
            'status' => 401,
            'code' => 'WRONG_QUERY_PARAMETERS',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request query parameters.',
          ],
        ],
      ], 401);
    }
    $offset_datetime = new DrupalDateTime($offset_datetime);
    $response = [];
    foreach ($this->moduleNodeData->getTemperatureLog($module_node, $offset_datetime) as $item) {
      $response[] = [
        'temperatures' => $item->temperatures,
        'datetime' => $item->datetime ? $item->datetime->format(SM_DATETIME_API_FORMAT) : NULL,
      ];
    }
    return new JsonResponse($response);
  }

  /**
   * Returns module node by device token.
   */
  private function getModuleNodeByControlToken(Request $request) {
    $control_token = Xss::filter($request->attributes->get('control_token'));
    if (!$control_token) {
      return NULL;
    }
    $query = $this->database->nodeStorage->getQuery();
    $query->condition('type', 'module');
    $query->condition('field_control_token', $control_token);
    $query->groupBy('nid');
    $result = $query->execute();
    $module_nid = $result ? $result[array_key_first($result)] : NULL;
    return $module_nid ? $this->database->nodeStorage->load($module_nid) : NULL;
  }

  /**
   * Returns module base data.
   */
  private function getModuleBaseData(NodeInterface $module_node, DrupalDateTime $current_datetime, ModuleLastValues $last_values) {
    $response = [];
    $response['deviceId'] = $module_node->get('field_device_id')->value;
    $response['conditionsEnabled'] = !!($module_node->get('field_conditions_enabled')->value);
    $response['scheduleEnabled'] = !!($module_node->get('field_schedule_enabled')->value);
    $response['notificationsEnabled'] = !!($module_node->get('field_notifications_enabled')->value);

    $settings_p_values = $module_node->get('field_settings')->getValue();
    if (isset($settings_p_values[0]['target_id'])) {
      if ($settings_p = $this->database->paragraphStorage->load($settings_p_values[0]['target_id'])) {
        $response['buttonMode'] = !!($settings_p->get('field_button_mode')->value);
      }
    }

    $new_values_p_values = $module_node->get('field_new_values')->getValue();
    if (isset($new_values_p_values[0]['target_id'])) {
      if ($new_values_p = $this->database->paragraphStorage->load($new_values_p_values[0]['target_id'])) {
        $response['newRelayState'] = !!($new_values_p->get('field_relay')->value);
      }
    }

    $response['isOnline'] = $this->moduleNodeData->isOnline($last_values->accessDatetime, $current_datetime);
    $response['relayState'] = $last_values->relay;
    $response['lastAccessTime'] = $last_values->accessDatetime ? $last_values->accessDatetime->format(SM_DATETIME_API_FORMAT) : NULL;
    $response['temperatures'] = $last_values->temperatures;

    $statuses = $this->moduleNodeData->getStatuses($module_node);
    $response['status'] = $statuses->pinChange;
    return $response;
  }

  /**
   * Update module app combinations.
   */
  private function updateModuleAppCombinations(string $device_uuid, array $new_module_ids) {
    if (!$device_uuid) {
      return;
    }
    $query = $this->database->connection->select('sm_app', 't');
    $query->condition('t.uuid', $device_uuid);
    $query->addField('t', 'id');
    $result = $query->execute()->fetchObject();
    $app_id = $result ? $result->id : NULL;
    if (!$app_id) {
      return;
    }
    $query = $this->database->connection->select('sm_module_app', 't');
    $query->condition('t.app_id', $app_id);
    $query->addField('t', 'module_nid');
    $result = $query->execute();
    $old_module_ids = [];
    foreach ($result as $item) {
      $old_module_ids[] = $item->module_nid;
    }

    $modules_to_delete = array_diff($old_module_ids, $new_module_ids);
    if ($modules_to_delete) {
      $this->database->connection->delete('sm_module_app')->condition('app_id', $app_id)->condition('module_nid', $modules_to_delete, 'IN')->execute();
    }

    $modules_to_add = array_diff($new_module_ids, $old_module_ids);
    if ($modules_to_add) {
      $query = $this->database->connection->insert('sm_module_app')->fields(['module_nid', 'app_id']);
      foreach ($modules_to_add as $module_nid) {
        $query->values([
          'module_nid' => $module_nid,
          'app_id' => $app_id,
        ]);
      }
      $query->execute();
    }
  }

  /**
   * Sets string value to entity field.
   */
  private function setStringValueToEntityField(EntityInterface $entity, string $field_key, array $data, string $key, bool $force = FALSE) {
    if (isset($data[$key])) {
      $value = Xss::filter($data[$key]);
      $entity->set($field_key, $value);
    }
    elseif ($force) {
      $entity->set($field_key, NULL);
    }
  }

  /**
   * Sets boolean value to entity field.
   */
  private function setBooleanValueToEntityField(EntityInterface $entity, string $field_key, array $data, string $key) {
    if (isset($data[$key])) {
      $value = !!Xss::filter($data[$key]);
      $entity->set($field_key, $value);
    }
  }

  /**
   * Sets integer value to entity field.
   */
  private function setIntegerValueToEntityField(EntityInterface $entity, string $field_key, array $data, string $key) {
    if (isset($data[$key])) {
      $value = (int) Xss::filter($data[$key]);
      $entity->set($field_key, $value);
    }
  }

  /**
   * Sets pin value to entity field.
   */
  private function setPinValueToEntityField(EntityInterface $entity, string $field_key, array $data, string $key) {
    $value = self::getPinValue($entity->get($field_key)->value);
    if ($value !== SM_PIN_NONE && isset($data[$key])) {
      $entity->set($field_key, !!Xss::filter($data[$key]) ? $value : $value - SM_PIN_DISABLED_DIFF);
    }
  }

  /**
   * Returns pin enabled.
   */
  private function getPinEnabled(int $value) {
    if ($value === SM_PIN_NONE) {
      return NULL;
    }
    return !!($value > SM_PIN_NONE);
  }

  /**
   * Returns pin value.
   */
  private function getPinValue(int $value) {
    if ($value === SM_PIN_NONE) {
      return NULL;
    }
    return $value > SM_PIN_NONE ? $value : $value + 256;
  }

}
