<?php

namespace Drupal\sm\Controller;

use Drupal\ban_ip\Service\BanManager;
use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\node\NodeInterface;
use Drupal\sm\Object\ModuleLastValues;
use Drupal\sm\Service\Data\ModuleNodeData;
use Drupal\sm\Service\Database;
use Drupal\sm\Service\PushNotification;
use Exception;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Class ModuleController.
 */
class ModuleController extends ControllerBase {

  /**
   * Database service.
   *
   * @var \Drupal\sm\Service\Database
   */
  protected Database $database;

  /**
   * Module node data service.
   *
   * @var \Drupal\sm\Service\Data\ModuleNodeData
   */
  protected ModuleNodeData $moduleNodeData;

  /**
   * Ban manager.
   *
   * @var \Drupal\ban_ip\Service\BanManager
   */
  protected BanManager $banManager;

  /**
   * Push notification service.
   *
   * @var \Drupal\sm\Service\PushNotification
   */
  protected PushNotification $pushNotification;

  /**
   * Constructor.
   */
  public function __construct(Database $database, ModuleNodeData $module_node_data, BanManager $ban_manager, PushNotification $push_notification) {
    $this->database = $database;
    $this->moduleNodeData = $module_node_data;
    $this->banManager = $ban_manager;
    $this->pushNotification = $push_notification;
  }

  /**
   * Static create function.
   */
  public static function create(ContainerInterface $container) {
    /** @var \Drupal\sm\Service\Database $database */
    $database = $container->get('sm.database');
    /** @var \Drupal\sm\Service\Data\ModuleNodeData $module_node_data */
    $module_node_data = $container->get('sm.module_node_data');
    /** @var \Drupal\ban_ip\Service\BanManager $ban_manager */
    $ban_manager = $container->get('ban_ip.ban_manager');
    /** @var \Drupal\sm\Service\PushNotification $push_notification */
    $push_notification = $container->get('sm.push_notification');
    return new static($database, $module_node_data, $ban_manager, $push_notification);
  }

  /**
   * Returns module data.
   */
  public function getModuleData(Request $request) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = self::getModuleNodeByDeviceToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 1000,
            'status' => 401,
            'code' => 'DEVICE_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter DEVICE_TOKEN.',
          ],
        ],
      ], 401);
    }
    $params = self::getRequestParameters($request);
    if (!is_array($params)) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 1001,
            'status' => 401,
            'code' => 'WRONG_PARAMETERS',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameters.',
          ],
        ],
      ], 401);
    }

    $response = [];
    $current_datetime = new DrupalDateTime('NOW');
    $last_values = $this->moduleNodeData->getLastValues($module_node);
    $current_relay_state = isset($params['relay']) ? !!(Xss::filter($params['relay'])) : NULL;
    $current_ssid = isset($params['ssid']) ? Xss::filter($params['ssid']) : NULL;
    $current_rssi = isset($params['rssi']) ? Xss::filter($params['rssi']) : NULL;
    $current_firmware_version = isset($params['fwVer']) ? Xss::filter($params['fwVer']) : NULL;
    $current_temperatures = isset($params['temps']) ? $params['temps'] : [];
    self::processAutoControl($module_node, $response, $current_datetime, $last_values, $current_relay_state, $current_firmware_version, $current_temperatures);
    self::processAll($module_node, $response, $params, $current_firmware_version);
    self::processDelay($response, $current_datetime, $last_values);
    self::updateLastValues($module_node, $current_datetime, $current_relay_state, $current_ssid, $current_rssi, $current_firmware_version, $current_temperatures);
    self::logAccess($module_node, $current_datetime, $current_ssid, $current_firmware_version);
    self::logRelayState($module_node, $current_datetime, $current_relay_state);
    self::logTemperatures($module_node, $current_datetime, $current_temperatures);
    Cache::invalidateTags($module_node->getCacheTags());
    return new JsonResponse($response);
  }

  /**
   * Returns request parameters.
   */
  private function getRequestParameters(Request $request) {
    $content = $request->getContent();
    return $content ? Json::decode($content) : [];
  }

  /**
   * Returns module node by device token.
   */
  private function getModuleNodeByDeviceToken(Request $request) {
    $device_token = Xss::filter($request->attributes->get('device_token'));
    if (!$device_token) {
      return NULL;
    }
    $query = $this->database->nodeStorage->getQuery();
    $query->condition('type', 'module');
    $query->condition('field_device_token', $device_token);
    $query->groupBy('nid');
    $result = $query->execute();
    $module_nid = $result ? $result[array_key_first($result)] : NULL;
    return $module_nid ? $this->database->nodeStorage->load($module_nid) : NULL;
  }

  /**
   * Updates last module values.
   */
  private function updateLastValues(NodeInterface $module_node, DrupalDateTime $current_datetime, $current_relay_state, $current_ssid, $current_rssi, $current_firmware_version, array $current_temperatures) {
    $fields = [];
    $fields['acc_time'] = $current_datetime->getTimestamp();
    $fields['relay'] = $current_relay_state ? 1 : 0;
    if ($current_ssid && $current_rssi) {
      $fields['ssid'] = $current_ssid;
      $fields['rssi'] = $current_rssi;
    }
    if ($current_firmware_version) {
      $fields['rst_time'] = $current_datetime->getTimestamp();
      $fields['fw_ver'] = $current_firmware_version;
    }
    $fields['temps'] = NULL;
    if ($current_temperatures && is_array($current_temperatures)) {
      $tmp_temperatures = [];
      foreach ($current_temperatures as $temperature) {
        if (is_numeric($temperature) && $temperature >= SM_TEMPERATURE_SENSOR_MIN_VALUE && $temperature <= SM_TEMPERATURE_SENSOR_MAX_VALUE) {
          $tmp_temperatures[] = $temperature;
        }
      }
      $fields['temps'] = serialize($tmp_temperatures);
    }
    $this->database->connection->update('sm_last_values')->fields($fields)->condition('nid', $module_node->id())->execute();
  }

  /**
   * Process auto control.
   */
  private function processAutoControl(NodeInterface $module_node, array &$response, DrupalDateTime $current_datetime, ModuleLastValues $last_values, $current_relay_state, $current_firmware_version, array $current_temperatures) {
    if ($module_node->get('field_notifications_enabled')->value) {
      if ($notifications_json = $module_node->get('field_notifications_json')->value) {
        $values = Json::decode($notifications_json);
        $old_value = self::processCondition($values, $last_values->temperatures, 0);
        $new_value = self::processCondition($values, $current_temperatures, 0);
        if (($current_firmware_version && $new_value) || (!$old_value && $new_value)) {
          $fcm_tokens = $this->moduleNodeData->getModuleAppFcmTokens($module_node);
          foreach ($fcm_tokens as $fcm_token) {
            $title = 'Upozornění';
            $message = $module_node->get('field_notifications_message')->value;
            $this->pushNotification->sendMessage($fcm_token, $title, $message);
          }
        }
      }
    }
    if ($module_node->get('field_schedule_enabled')->value) {
      if ($conditions_json = $module_node->get('field_schedule_json')->value) {
        $values = Json::decode($conditions_json);
        $old_value = self::processSchedule($values, $last_values->accessDatetime);
        $new_value = self::processSchedule($values, $current_datetime);
        if ($current_firmware_version) {
          $this->moduleNodeData->setNewValues($module_node, !!$new_value);
        }
        else {
          if (!$old_value && $new_value) {
            $this->moduleNodeData->setNewValues($module_node, TRUE);
          }
          elseif ($old_value && !$new_value) {
            $this->moduleNodeData->setNewValues($module_node, FALSE);
          }
        }
      }
    }
    if ($module_node->get('field_conditions_enabled')->value) {
      if ($conditions_json = $module_node->get('field_conditions_json')->value) {
        $threshold = $module_node->get('field_conditions_threshold')->value;
        $threshold = $current_relay_state && is_numeric($threshold) && $threshold > 0 ? $threshold : 0;
        $values = Json::decode($conditions_json);
        $old_value = self::processCondition($values, $last_values->temperatures, $threshold);
        $new_value = self::processCondition($values, $current_temperatures, $threshold);
        if ($current_firmware_version) {
          $this->moduleNodeData->setNewValues($module_node, !!$new_value);
        }
        else {
          if (!$old_value && $new_value) {
            $this->moduleNodeData->setNewValues($module_node, TRUE);
          }
          elseif ($old_value && !$new_value) {
            $this->moduleNodeData->setNewValues($module_node, FALSE);
          }
        }
      }
    }
  }

  /**
   * Process all.
   */
  private function processAll(NodeInterface $module_node, array &$response, array $params, $current_firmware_version) {
    $state_fields = $this->moduleNodeData->getStatuses($module_node);
    $fields = [];
    self::processPinChange($module_node, $params, $state_fields->pinChange, $fields, $response);
    self::processFirmwareUpdate($module_node, $state_fields->fwUpdate, $fields, $response, $current_firmware_version);
    self::processRestart($module_node, $state_fields->restart, $fields, $response, $current_firmware_version);
    self::processConfigurationSave($module_node, $params, $state_fields->cfgSave, $fields, $response);
    self::processSettingsSave($module_node, $params, $state_fields->stgSave, $fields, $response);
    self::processNetworkSave($module_node, $params, $state_fields->netSave, $fields, $response);
    self::processNetworkScan($module_node, $params, $state_fields->netScan, $fields, $response);
    self::processConnect($module_node, $params, $state_fields->connect, $fields, $response);
    if ($fields) {
      $this->database->connection->update('sm_status')->fields($fields)->condition('nid', $module_node->id())->execute();
    }
  }

  /**
   * Process delay.
   */
  private function processDelay(array &$response, DrupalDateTime $current_datetime, ModuleLastValues $last_values) {
    if ($response) {
      $response['delay'] = SM_LOOP_DELAY_MIN;
    }
    elseif ($this->moduleNodeData->isControlOnline($last_values->controlDatetime, $current_datetime)) {
      $response['delay'] = SM_LOOP_DELAY_CONTROL;
    }
    else {
      $response['delay'] = SM_LOOP_DELAY_DEFAULT;
    }
  }

  /**
   * Changes value of module pins.
   */
  private function processPinChange(NodeInterface $module_node, array $params, int $process, array &$fields, array &$response) {
    if ($process == SM_STATUS_PROCESSING || $process == SM_STATUS_CONFIRMATION) {
      $new_values_p_values = $module_node->get('field_new_values')->getValue();
      if (isset($new_values_p_values[0]['target_id'])) {
        if ($new_values_p = $this->database->paragraphStorage->load($new_values_p_values[0]['target_id'])) {
          $new_relay_state = !!($new_values_p->get('field_relay')->value);
          $new_led_1_value = (int) $new_values_p->get('field_led_1')->value;
          $new_led_2_value = (int) $new_values_p->get('field_led_2')->value;
          $new_led_3_value = (int) $new_values_p->get('field_led_3')->value;

          $configuration_p_values = $module_node->get('field_configuration')->getValue();
          if (isset($configuration_p_values[0]['target_id'])) {
            if ($configuration_p = $this->database->paragraphStorage->load($configuration_p_values[0]['target_id'])) {
              $pin_relay = (int) $configuration_p->get('field_pin_relay')->value;
              $pin_led_1 = (int) $configuration_p->get('field_pin_led_1')->value;
              $pin_led_2 = (int) $configuration_p->get('field_pin_led_2')->value;
              $pin_led_3 = (int) $configuration_p->get('field_pin_led_3')->value;
              if ($process == SM_STATUS_CONFIRMATION) {
                $current_relay_state = isset($params['relay']) ? !!(Xss::filter($params['relay'])) : NULL;
                $current_led_1_value = isset($params['led1']) ? Xss::filter($params['led1']) : NULL;
                $current_led_2_value = isset($params['led2']) ? Xss::filter($params['led2']) : NULL;
                $current_led_3_value = isset($params['led3']) ? Xss::filter($params['led3']) : NULL;
                if (($pin_relay == SM_PIN_NONE || $current_relay_state == $new_relay_state)
                  && ($pin_led_1 == SM_PIN_NONE || $current_led_1_value == $new_led_1_value)
                  && ($pin_led_2 == SM_PIN_NONE || $current_led_2_value == $new_led_2_value)
                  && ($pin_led_3 == SM_PIN_NONE || $current_led_3_value == $new_led_3_value)) {
                  $fields['pin_change'] = SM_STATUS_OK;
                  return TRUE;
                }
                $process = SM_STATUS_PROCESSING;
              }
              if ($process == SM_STATUS_PROCESSING) {
                $fields['pin_change'] = SM_STATUS_CONFIRMATION;
                if ($pin_relay > SM_PIN_NONE) {
                  $response['relay'] = $new_relay_state;
                }
                if ($pin_led_1 > SM_PIN_NONE) {
                  $response['led1'] = $new_led_1_value;
                }
                if ($pin_led_2 > SM_PIN_NONE) {
                  $response['led2'] = $new_led_2_value;
                }
                if ($pin_led_3 > SM_PIN_NONE) {
                  $response['led3'] = $new_led_3_value;
                }
                return TRUE;
              }
            }
          }
        }
      }
      $fields['pin_change'] = SM_STATUS_ERROR;
    }
    return FALSE;
  }

  /**
   * Updates module firmware.
   */
  private function processFirmwareUpdate(NodeInterface $module_node, int $process, array &$fields, array &$response, $current_firmware_version) {
    if ($process == SM_STATUS_PROCESSING || $process == SM_STATUS_CONFIRMATION) {
      /** @var \Drupal\node\NodeInterface $firmware_node */
      if ($firmware_node = $module_node->get('field_firmware')->entity) {
        if ($new_firmware_version = $firmware_node->get('field_version')->value) {
          if ($process == SM_STATUS_PROCESSING) {
            if ($device_token = $module_node->get('field_device_token')->value) {
              $fields['fw_update'] = SM_STATUS_CONFIRMATION;
              $response['updateUrl'] = '/api/v1/' . $device_token . '/firmware';
              return TRUE;
            }
          }
          if ($process == SM_STATUS_CONFIRMATION) {
            if ($current_firmware_version == $new_firmware_version) {
              $fields['fw_update'] = SM_STATUS_OK;
              return TRUE;
            }
          }
        }
      }
      $fields['fw_update'] = SM_STATUS_ERROR;
    }
    return FALSE;
  }

  /**
   * Restarts module.
   */
  private function processRestart(EntityInterface $module_node, int $process, array &$fields, array &$response, $current_firmware_version) {
    if ($process == SM_STATUS_PROCESSING) {
      $fields['restart'] = SM_STATUS_CONFIRMATION;
      $response['restart'] = TRUE;
      return TRUE;
    }
    if ($process == SM_STATUS_CONFIRMATION) {
      if ($current_firmware_version) {
        $fields['restart'] = SM_STATUS_OK;
        return TRUE;
      }
      $fields['restart'] = SM_STATUS_ERROR;
    }
    return FALSE;
  }

  /**
   * Saves configuration to module.
   */
  private function processConfigurationSave(NodeInterface $module_node, array $params, int $process, array &$fields, array &$response) {
    if ($process == SM_STATUS_PROCESSING || $process == SM_STATUS_CONFIRMATION) {
      $configuration_p_values = $module_node->get('field_configuration')->getValue();
      if (isset($configuration_p_values[0]['target_id'])) {
        if ($configuration_p = $this->database->paragraphStorage->load($configuration_p_values[0]['target_id'])) {
          $new_server_password = $configuration_p->get('field_server_password')->value;
          $new_pin_button = (int) max($configuration_p->get('field_pin_button')->value, SM_PIN_NONE);
          $new_pin_relay = (int) max($configuration_p->get('field_pin_relay')->value, SM_PIN_NONE);
          $new_pin_led = (int) max($configuration_p->get('field_pin_led')->value, SM_PIN_NONE);
          $new_pin_ext_button = (int) max($configuration_p->get('field_pin_ext_button')->value, SM_PIN_NONE);
          $new_pin_ext_led = (int) max($configuration_p->get('field_pin_ext_led')->value, SM_PIN_NONE);
          $new_pin_led_1 = (int) max($configuration_p->get('field_pin_led_1')->value, SM_PIN_NONE);
          $new_pin_led_2 = (int) max($configuration_p->get('field_pin_led_2')->value, SM_PIN_NONE);
          $new_pin_led_3 = (int) max($configuration_p->get('field_pin_led_3')->value, SM_PIN_NONE);
          $new_pin_temp_sensors = (int) max($configuration_p->get('field_pin_temp_sensors')->value, SM_PIN_NONE);
          $new_temp_sensor_addresses = [];
          if ($new_pin_temp_sensors > SM_PIN_NONE) {
            $new_temp_sensor_values = $configuration_p->get('field_temp_sensors')->getValue();
            foreach ($new_temp_sensor_values as $new_temp_sensor_value) {
              if ($new_temp_sensor_value && $new_temp_sensor_value['target_id']) {
                $new_temp_sensor_node = $this->database->nodeStorage->load($new_temp_sensor_value['target_id']);
                $new_temp_sensor_address = $new_temp_sensor_node->get('field_device_address')->value;
                $new_temp_sensor_address_array = [];
                for ($i = 0; $i < 8; $i++) {
                  $new_temp_sensor_address_array[$i] = hexdec(substr($new_temp_sensor_address, $i * 2, 2));
                }
                $new_temp_sensor_addresses[] = $new_temp_sensor_address_array;
              }
            }
          }
          if ($process == SM_STATUS_PROCESSING) {
            $fields['cfg_save'] = SM_STATUS_CONFIRMATION;
            $response['serverPw'] = $new_server_password;
            $response['pinButton'] = $new_pin_button;
            $response['pinRelay'] = $new_pin_relay;
            $response['pinLed'] = $new_pin_led;
            $response['pinExtButton'] = $new_pin_ext_button;
            $response['pinExtLed'] = $new_pin_ext_led;
            $response['pinLed1'] = $new_pin_led_1;
            $response['pinLed2'] = $new_pin_led_2;
            $response['pinLed3'] = $new_pin_led_3;
            $response['pinTemp'] = $new_pin_temp_sensors;
            $response['tempAdrs'] = $new_temp_sensor_addresses;
            return TRUE;
          }
          if ($process == SM_STATUS_CONFIRMATION) {
            $server_password = isset($params['serverPw']) ? Xss::filter($params['serverPw']) : NULL;
            $pin_button = isset($params['pinButton']) ? Xss::filter($params['pinButton']) : SM_PIN_NONE;
            $pin_relay = isset($params['pinRelay']) ? Xss::filter($params['pinRelay']) : SM_PIN_NONE;
            $pin_led = isset($params['pinLed']) ? Xss::filter($params['pinLed']) : SM_PIN_NONE;
            $pin_ext_button = isset($params['pinExtButton']) ? Xss::filter($params['pinExtButton']) : SM_PIN_NONE;
            $pin_ext_led = isset($params['pinExtLed']) ? Xss::filter($params['pinExtLed']) : SM_PIN_NONE;
            $pin_led_1 = isset($params['pinLed1']) ? Xss::filter($params['pinLed1']) : SM_PIN_NONE;
            $pin_led_2 = isset($params['pinLed2']) ? Xss::filter($params['pinLed2']) : SM_PIN_NONE;
            $pin_led_3 = isset($params['pinLed3']) ? Xss::filter($params['pinLed3']) : SM_PIN_NONE;
            $pin_temp_sensors = isset($params['pinTemp']) ? Xss::filter($params['pinTemp']) : SM_PIN_NONE;
            $temp_sensor_addresses = isset($params['tempAdrs']) ? $params['tempAdrs'] : [];
            $is_temp_sensor_addresses_ok = FALSE;
            if (!$new_temp_sensor_addresses && !$temp_sensor_addresses) {
              $is_temp_sensor_addresses_ok = TRUE;
            }
            elseif ($new_temp_sensor_addresses && $temp_sensor_addresses && is_array($temp_sensor_addresses)) {
              $is_temp_sensor_addresses_ok = TRUE;
              foreach ($new_temp_sensor_addresses as $i => $new_temp_sensor_address_array) {
                $temp_sensor_address_array = isset($temp_sensor_addresses[$i]) ? $temp_sensor_addresses[$i] : [];
                if (!is_array($new_temp_sensor_address_array) || !is_array($temp_sensor_address_array)) {
                  $is_temp_sensor_addresses_ok = FALSE;
                  break;
                }
                if (count(array_diff_assoc($new_temp_sensor_address_array, $temp_sensor_address_array))) {
                  $is_temp_sensor_addresses_ok = FALSE;
                  break;
                }
              }
            }
            if ($server_password == $new_server_password && $pin_button == $new_pin_button && $pin_relay == $new_pin_relay
              && $pin_led == $new_pin_led && $pin_ext_button == $new_pin_ext_button && $pin_ext_led == $new_pin_ext_led
              && $pin_led_1 == $new_pin_led_1 && $pin_led_2 == $new_pin_led_2 && $pin_led_3 == $new_pin_led_3
              && $pin_temp_sensors == $new_pin_temp_sensors && $is_temp_sensor_addresses_ok) {
              $fields['cfg_save'] = SM_STATUS_OK;
              return TRUE;
            }
          }
        }
      }
      $fields['cfg_save'] = SM_STATUS_ERROR;
    }
    return FALSE;
  }

  /**
   * Saves settings to module.
   */
  private function processSettingsSave(NodeInterface $module_node, array $params, int $process, array &$fields, array &$response) {
    if ($process == SM_STATUS_PROCESSING || $process == SM_STATUS_CONFIRMATION) {
      $settings_p_values = $module_node->get('field_settings')->getValue();
      if (isset($settings_p_values[0]['target_id'])) {
        if ($settings_p = $this->database->paragraphStorage->load($settings_p_values[0]['target_id'])) {
          $new_button_mode = !!($settings_p->get('field_button_mode')->value);
          $new_default_state = !!($settings_p->get('field_default_state')->value);
          $new_state_delay = (int) $settings_p->get('field_state_delay')->value;
          $new_loop_delay = (int) $settings_p->get('field_loop_delay')->value;
          $new_default_led_1 = (int) $settings_p->get('field_default_led_1')->value;
          $new_default_led_2 = (int) $settings_p->get('field_default_led_2')->value;
          $new_default_led_3 = (int) $settings_p->get('field_default_led_3')->value;
          if ($process == SM_STATUS_PROCESSING) {
            $fields['stg_save'] = SM_STATUS_CONFIRMATION;
            $response['buttonMode'] = $new_button_mode;
            $response['defaultState'] = $new_default_state;
            $response['stateDelay'] = $new_state_delay;
            $response['loopDelay'] = $new_loop_delay;
            $response['defaultLed1'] = $new_default_led_1;
            $response['defaultLed2'] = $new_default_led_2;
            $response['defaultLed3'] = $new_default_led_3;
            return TRUE;
          }
          if ($process == SM_STATUS_CONFIRMATION) {
            $button_mode = isset($params['buttonMode']) ? !!(Xss::filter($params['buttonMode'])) : NULL;
            $default_state = isset($params['defaultState']) ? !!(Xss::filter($params['defaultState'])) : NULL;
            $state_delay = isset($params['stateDelay']) ? Xss::filter($params['stateDelay']) : NULL;
            $loop_delay = isset($params['loopDelay']) ? Xss::filter($params['loopDelay']) : NULL;
            $default_led_1 = isset($params['defaultLed1']) ? Xss::filter($params['defaultLed1']) : NULL;
            $default_led_2 = isset($params['defaultLed2']) ? Xss::filter($params['defaultLed2']) : NULL;
            $default_led_3 = isset($params['defaultLed3']) ? Xss::filter($params['defaultLed3']) : NULL;
            if ($button_mode == $new_button_mode && $default_state == $new_default_state && $state_delay == $new_state_delay && $loop_delay == $new_loop_delay
              && $default_led_1 == $new_default_led_1 && $default_led_2 == $new_default_led_2 && $default_led_3 == $new_default_led_3) {
              $fields['stg_save'] = SM_STATUS_OK;
              return TRUE;
            }
          }
        }
      }
      $fields['stg_save'] = SM_STATUS_ERROR;
    }
    return FALSE;
  }

  /**
   * Saves networks to module.
   */
  private function processNetworkSave(NodeInterface $module_node, array $params, int $process, array &$fields, array &$response) {
    if ($process == SM_STATUS_PROCESSING) {
      $fields['net_save'] = SM_STATUS_CONFIRMATION;
      $network_save_p_values = $module_node->get('field_network_save')->getValue();
      foreach ($network_save_p_values as $i => $network_save_p_value) {
        if ($network_save_p = $this->database->paragraphStorage->load($network_save_p_value['target_id'])) {
          $new_ssid = $network_save_p->get('field_ssid')->value;
          $new_password = $network_save_p->get('field_password')->value;
          if ($new_ssid) {
            $response['saveNets'][$i]['ssid'] = $new_ssid;
            $response['saveNets'][$i]['pw'] = $new_password;
          }
        }
      }
      return TRUE;
    }
    if ($process == SM_STATUS_CONFIRMATION) {
      $saved_networks = isset($params['savedNets']) ? $params['savedNets'] : NULL;
      $network_save_p_values = $module_node->get('field_network_save')->getValue();
      foreach ($network_save_p_values as $i => $network_save_p_value) {
        if ($network_save_p = $this->database->paragraphStorage->load($network_save_p_value['target_id'])) {
          $new_ssid = $network_save_p->get('field_ssid')->value;
          $new_password = $network_save_p->get('field_password')->value;
          if ($new_ssid) {
            $saved_ssid = isset($saved_networks[$i]['ssid']) ? Xss::filter($saved_networks[$i]['ssid']) : NULL;
            $saved_password = isset($saved_networks[$i]['pw']) ? Xss::filter($saved_networks[$i]['pw']) : NULL;
            if ($saved_ssid != $new_ssid || $saved_password != $new_password) {
              $fields['net_save'] = SM_STATUS_ERROR;
              return FALSE;
            }
          }
        }
      }
      $fields['net_save'] = SM_STATUS_OK;
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Scans available networks near module.
   */
  private function processNetworkScan(NodeInterface $module_node, array $params, int $process, array &$fields, array &$response) {
    if ($process == SM_STATUS_PROCESSING) {
      $fields['net_scan'] = SM_STATUS_CONFIRMATION;
      $response['scanNets'] = TRUE;
      return TRUE;
    }
    if ($process == SM_STATUS_CONFIRMATION) {
      $network_scan_p_values = $module_node->get('field_network_scan')->getValue();
      $available_networks = isset($params['availNets']) ? $params['availNets'] : NULL;
      for ($i = 0; $i < SM_MAX_NETWORK_CNT; $i++) {
        $tmp_ssid = isset($available_networks[$i]['ssid']) ? Xss::filter($available_networks[$i]['ssid']) : NULL;
        $tmp_rssi = isset($available_networks[$i]['rssi']) ? Xss::filter($available_networks[$i]['rssi']) : NULL;
        $tmp_bssid = isset($available_networks[$i]['bssid']) ? Xss::filter($available_networks[$i]['bssid']) : NULL;
        $tmp_channel = isset($available_networks[$i]['channel']) ? Xss::filter($available_networks[$i]['channel']) : NULL;
        $tmp_encryption_type = isset($available_networks[$i]['encType']) ? Xss::filter($available_networks[$i]['encType']) : NULL;
        if (isset($network_scan_p_values[$i]['target_id']) && $network_scan_p = $this->database->paragraphStorage->load($network_scan_p_values[$i]['target_id'])) {
          $network_scan_p->set('field_ssid', $tmp_ssid);
          $network_scan_p->set('field_rssi', $tmp_rssi);
          $network_scan_p->set('field_bssid', $tmp_bssid);
          $network_scan_p->set('field_channel', $tmp_channel);
          $network_scan_p->set('field_encryption_type', $tmp_encryption_type);
          try {
            $network_scan_p->save();
          }
          catch (EntityStorageException $e) {
          }
        }
      }
      $fields['net_scan'] = SM_STATUS_OK;
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Connects module to selected network.
   */
  private function processConnect(NodeInterface $module_node, array $params, int $process, array &$fields, array &$response) {
    if ($process == SM_STATUS_PROCESSING || $process == SM_STATUS_CONFIRMATION) {
      if ($new_connection_ssid = $module_node->get('field_connect_ssid')->value) {
        $new_connection_password = $module_node->get('field_connect_password')->value;
        if ($process == SM_STATUS_PROCESSING) {
          $fields['connect'] = SM_STATUS_CONFIRMATION;
          $response['connSsid'] = $new_connection_ssid;
          $response['connPw'] = $new_connection_password;
          return TRUE;
        }
        $connection_ssid = isset($params['ssid']) ? Xss::filter($params['ssid']) : NULL;
        if ($connection_ssid == $new_connection_ssid) {
          $fields['connect'] = SM_STATUS_OK;
          return TRUE;
        }
        $fields['connect'] = SM_STATUS_ERROR;
      }
    }
    return FALSE;
  }

  /**
   * Logs module access.
   */
  private function logAccess(NodeInterface $module_node, DrupalDateTime $current_datetime, $current_ssid, $current_firmware_version) {
    $query = $this->database->connection->select('sm_log_access', 't');
    $query->condition('t.nid', $module_node->id());
    $query->addField('t', 'start_time');
    $query->addField('t', 'end_time');
    $query->addField('t', 'ssid');
    $query->addField('t', 'restart_cnt');
    $query->orderBy('t.start_time', 'DESC');
    $result = $query->execute();
    $last_fields = $result->fetchAssoc();
    $fields = [];
    $fields['end_time'] = $current_datetime->getTimestamp();
    if ($last_fields && $last_fields['start_time'] && $last_fields['end_time'] && $last_fields['ssid']) {
      $diff_timestamp = $current_datetime->getTimestamp() - $last_fields['end_time'];
      if ($diff_timestamp <= SM_MAX_ACCESS_DELAY_TIME && $current_ssid == $last_fields['ssid']) {
        if ($current_firmware_version) {
          $fields['restart_cnt'] = $last_fields['restart_cnt'] + 1;
        }
        $query = $this->database->connection->update('sm_log_access');
        $query->fields($fields);
        $query->condition('nid', $module_node->id());
        $query->condition('start_time', $last_fields['start_time']);
        $query->execute();
        return;
      }
    }
    $fields['nid'] = $module_node->id();
    $fields['start_time'] = $current_datetime->getTimestamp();
    $fields['ssid'] = $current_ssid;
    $fields['restart_cnt'] = $current_firmware_version ? 1 : 0;
    try {
      $this->database->connection->insert('sm_log_access')->fields($fields)->execute();
    }
    catch (Exception $e) {
    }
  }

  /**
   * Logs module relay state.
   */
  private function logRelayState(NodeInterface $module_node, DrupalDateTime $current_datetime, $current_relay_state) {
    $button_mode = FALSE;
    $settings_p_values = $module_node->get('field_settings')->getValue();
    if (isset($settings_p_values[0]['target_id'])) {
      if ($settings_p = $this->database->paragraphStorage->load($settings_p_values[0]['target_id'])) {
        $button_mode = !!($settings_p->get('field_button_mode')->value);
      }
    }
    $query = $this->database->connection->select('sm_log_relay_state', 't');
    $query->condition('t.nid', $module_node->id());
    $query->addField('t', 'start_time');
    $query->addField('t', 'end_time');
    $query->addField('t', 'state');
    $query->orderBy('t.start_time', 'DESC');
    $result = $query->execute();
    $last_fields = $result->fetchAssoc();
    $fields = [];
    $fields['end_time'] = $current_datetime->getTimestamp();
    if ($last_fields && $last_fields['start_time'] && $last_fields['end_time']) {
      $diff_timestamp = $current_datetime->getTimestamp() - $last_fields['end_time'];
      if ($diff_timestamp <= SM_MAX_ACCESS_DELAY_TIME && $current_relay_state == $last_fields['state']) {
        $query = $this->database->connection->update('sm_log_relay_state');
        $query->fields($fields);
        $query->condition('nid', $module_node->id());
        $query->condition('start_time', $last_fields['start_time']);
        $query->execute();
        return;
      }
    }
    $fields['nid'] = $module_node->id();
    $fields['start_time'] = $current_datetime->getTimestamp();
    $fields['state'] = $current_relay_state ? ($button_mode ? 2 : 1) : 0;
    try {
      $this->database->connection->insert('sm_log_relay_state')->fields($fields)->execute();
    }
    catch (Exception $e) {
    }
  }

  /**
   * Logs module temperatures.
   */
  private function logTemperatures(NodeInterface $module_node, DrupalDateTime $current_datetime, array $current_temperatures) {
    if ($current_temperatures && is_array($current_temperatures)) {
      $query = $this->database->connection->select('sm_log_temperatures', 't');
      $query->condition('t.nid', $module_node->id());
      $query->addField('t', 'time');
      $query->orderBy('t.time', 'DESC');
      $result = $query->execute();
      $last_fields = $result->fetchAssoc();
      $last_timestamp = $last_fields && $last_fields['time'] ? $last_fields['time'] : 0;
      $current_timestamp = $current_datetime->getTimestamp();
      if ($current_timestamp - $last_timestamp >= SM_LOG_TEMPERATURE_DELAY_TIME) {
        $fields = [];
        $fields['nid'] = $module_node->id();
        $fields['time'] = $current_timestamp;
        $tmp_temperatures = [];
        foreach ($current_temperatures as $temperature) {
          if (is_numeric($temperature) && $temperature >= SM_TEMPERATURE_SENSOR_MIN_VALUE && $temperature <= SM_TEMPERATURE_SENSOR_MAX_VALUE) {
            $tmp_temperatures[] = $temperature;
          }
        }
        if ($tmp_temperatures) {
          $fields['temperatures'] = serialize($tmp_temperatures);
          try {
            $this->database->connection->insert('sm_log_temperatures')->fields($fields)->execute();
          }
          catch (Exception $e) {
          }
        }
      }
    }
  }

  /**
   * Returns module firmware file.
   */
  public function getFirmwareFile(Request $request) {
    $module_node = self::getModuleNodeByDeviceToken($request);
    if (!$module_node) {
      $this->banManager->banIp($request->getClientIp());
      return new JsonResponse([
        'errors' => [
          [
            'id' => 1010,
            'status' => 401,
            'code' => 'DEVICE_TOKEN_ERROR',
            'title' => 'Unauthorized',
            'detail' => 'Wrong request parameter DEVICE_TOKEN.',
          ],
        ],
      ], 401);
    }
    $query = $this->database->connection->select('sm_status', 't');
    $query->condition('t.nid', $module_node->id());
    $query->addField('t', 'fw_update');
    $result = $query->execute();
    $process = $result->fetchAssoc()['fw_update'];
    if ($process == SM_STATUS_CONFIRMATION) {
      if ($firmware_node = $module_node->get('field_firmware')->entity) {
        /** @var \Drupal\node\NodeInterface $firmware_node */
        $firmware_version = $firmware_node->get('field_version')->value;
        $firmware_file = $firmware_node->get('field_firmware_file')->entity;
        if ($firmware_version && $firmware_file) {
          /** @var \Drupal\file\FileInterface $firmware_file */
          $firmware_file_name = 'plix-sm-fw-' . $firmware_version;
          $firmware_file_uri = $firmware_file->getFileUri();
          $firmware_file_mime = $firmware_file->getMimeType();
          $headers = [
            'Content-Type' => $firmware_file_mime,
            'Content-Disposition' => 'attachment;filename="' . $firmware_file_name . '.bin"',
          ];
          return new BinaryFileResponse($firmware_file_uri, 200, $headers, TRUE);
        }
      }
    }

    return new JsonResponse([
      'errors' => [
        [
          'id' => 1011,
          'status' => 403,
          'code' => 'FIRMWARE_DOWNLOAD_FORBIDDEN',
          'title' => 'Forbidden',
          'detail' => 'This module is not currently allowed to download firmware.',
        ],
      ],
    ], 403);
  }

  /**
   * Process condition.
   */
  private function processCondition($values, array $temperatures, float $threshold) {
    if (is_array($values)) {
      foreach ($values as $operator => $value) {
        switch ($operator) {
          case 'AND':
            foreach ($value as $item) {
              if (!self::processCondition($item, $temperatures, $threshold)) {
                return FALSE;
              }
            }
            return TRUE;

          case 'OR':
            foreach ($value as $item) {
              if (self::processCondition($item, $temperatures, $threshold)) {
                return TRUE;
              }
            }
            return FALSE;
        }
        if (!isset($value[0]) || !isset($value[1])) {
          return NULL;
        }
        $a = self::processCondition($value[0], $temperatures, $threshold);
        $b = self::processCondition($value[1], $temperatures, $threshold);
        switch ($operator) {
          case 'EQ':
            return $a == $b;

          case 'GT':
            return $a - $b > -$threshold;

          case 'LT':
            return $a - $b < $threshold;

          case 'GTE':
            return $a - $b >= -$threshold;

          case 'LTE':
            return $a - $b <= $threshold;

          case 'SUM':
            return $a + $b;

          case 'DIFF':
            return $a - $b;
        }
      }
    }
    elseif (is_numeric($values)) {
      return $values;
    }
    elseif ($values) {
      if (substr($values, 0, 1) == 'T') {
        $key = explode('T', $values);
        if (is_array($key) && isset($key[1])) {
          $key = $key[1];
          if (isset($temperatures[$key])) {
            return $temperatures[$key];
          }
        }
      }
    }
    return NULL;
  }

  /**
   * Process schedule.
   */
  private function processSchedule(array $values, DrupalDateTime $datetime) {
    foreach ($values as $value) {
      if (isset($value['days']) && is_array($value['days']) && isset($value['time']) && isset($value['dur']) && is_numeric($value['dur'])) {
        if (in_array($datetime->format('N'), $value['days']) && preg_match('/^(?:2[0-3]|[01][0-9]):[0-5][0-9]$/', $value['time']) && $value['dur'] > 0) {
          $start_datetime = DrupalDateTime::createFromFormat('Y-m-d\TH:i:s', $datetime->format('Y-m-d') . 'T' . $value['time'] . ':00');
          $end_datetime = DrupalDateTime::createFromTimestamp($start_datetime->getTimestamp() + $value['dur']);
          $timestamp = $datetime->getTimestamp();
          if ($start_datetime->getTimestamp() <= $timestamp && $timestamp <= $end_datetime->getTimestamp()) {
            return TRUE;
          }
        }
      }
    }
    return FALSE;
  }

}
