<?php

namespace Drupal\sm\Form;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\sm\Service\Data\AppData;
use Drupal\sm\Service\Database;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class AdminAppsForm.
 */
class AdminAppsForm extends FormBase {

  /**
   * Database service.
   *
   * @var \Drupal\sm\Service\Database
   */
  protected Database $database;

  /**
   * Module node data service.
   *
   * @var \Drupal\sm\Service\Data\AppData
   */
  protected AppData $appData;

  /**
   * Constructor.
   */
  public function __construct(Database $database, AppData $app_data) {
    $this->database = $database;
    $this->appData = $app_data;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /** @var \Drupal\sm\Service\Database $database */
    $database = $container->get('sm.database');
    /** @var \Drupal\sm\Service\Data\AppData $app_data */
    $app_data = $container->get('sm.app_data');
    return new static($database, $app_data);
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'admin_modules_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $header = [
      ['data' => 'UUID', 'field' => 't.uuid'],
      ['data' => 'Platforma'],
      ['data' => 'Verze', 'field' => 't.version'],
      [
        'data' => 'Poslední spojení',
        'field' => 't.access_time',
        'sort' => 'desc',
      ],
      ['data' => 'FCM token'],
      ['data' => 'Moduly'],
    ];

    $query = $this->database->connection->select('sm_app', 't');
    $query->leftJoin('sm_module_app', 'u', 'u.app_id = t.id');
    $query->addField('t', 'id');
    $query->addField('t', 'uuid');
    $query->addField('t', 'platform');
    $query->addField('t', 'version');
    $query->addField('t', 'fcm_token');
    $query->addField('t', 'access_time');
    $query->addField('u', 'module_nid');
    $query = $query->extend('Drupal\Core\Database\Query\TableSortExtender')->orderByHeader($header);
    $query = $query->extend('Drupal\Core\Database\Query\PagerSelectExtender')->limit(100);
    $results = $query->execute();

    $data = [];
    foreach ($results as $result) {
      $id = $result->id;
      if (!isset($data[$id])) {
        $data[$id] = [
          'uuid' => $result->uuid,
          'platform' => $result->platform,
          'version' => $result->version,
          'access_time' => $result->access_time,
          'fcm_token' => $result->fcm_token,
          'module_ids' => [],
        ];
      }
      if ($result->module_nid) {
        $data[$id]['module_ids'][] = $result->module_nid;
      }
    }

    $rows = [];
    foreach ($data as $id => $item) {
      $modules = '';
      $module_nodes = $item['module_ids'] ? $this->database->nodeStorage->loadMultiple($item['module_ids']) : [];
      foreach ($module_nodes as $key => $module_node) {
        $url = Url::fromRoute('entity.node.canonical', ['node' => $module_node->id()]);
        $link = Link::fromTextAndUrl($module_node->label(), $url);
        $modules .= ($key === array_key_first($module_nodes) ? '' : ' | ') . $link->toString();
      }
      $rows[] = [
        [
          'class' => ['code'],
          'data' => ['#markup' => $item['uuid']],
        ],
        ['data' => ['#markup' => $this->appData->getPlatformName($item['platform'])]],
        ['data' => ['#markup' => $item['version']]],
        [
          'data' => [
            '#markup' => $item['access_time'] ? DrupalDateTime::createFromTimestamp($item['access_time'])->format(SM_DATETIME_DEFAULT_FORMAT) : NULL,
          ],
        ],
        [
          'class' => ['code', 'fcm-token'],
          'data' => ['#markup' => $item['fcm_token']],
        ],
        ['data' => ['#markup' => $modules]],
      ];
    }

    $form['table'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $rows,
      '#empty' => 'Momentálně neexistují žádné aplikace.',
    ];
    $form['pager'] = ['#type' => 'pager'];
    $form['#attached']['library'][] = 'sm/admin';
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
  }

}
