<?php

namespace Drupal\sm\Service\Data;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\node\NodeInterface;
use Drupal\sm\Object\ModuleLastValues;
use Drupal\sm\Object\ModuleStatuses;
use Drupal\sm\Service\Database;

/**
 * Class ModuleNodeData.
 */
class ModuleNodeData {

  /**
   * Database service.
   *
   * @var \Drupal\sm\Service\Database
   */
  protected Database $database;

  /**
   * Constructor.
   */
  public function __construct(Database $database) {
    $this->database = $database;
  }

  /**
   * Returns module data.
   */
  public function getModule(NodeInterface $module_node) {
    $current_datetime = new DrupalDateTime('NOW');
    $last_values = self::getLastValues($module_node);
    return (object) [
      'isOnline' => self::isOnline($last_values->accessDatetime, $current_datetime),
      'lastValues' => $last_values,
      'status' => self::getStatuses($module_node),
      'accessLog' => self::getAccessLog($module_node, $current_datetime),
      'relayStateLog' => self::getRelayStateLog($module_node, $current_datetime),
      'temperatureLog' => self::getTemperatureLog($module_node, $current_datetime),
    ];
  }

  /**
   * Returns module application FCM tokens.
   */
  public function getModuleAppFcmTokens(NodeInterface $module_node) {
    $query = $this->database->connection->select('sm_module_app', 't');
    $query->condition('t.module_nid', $module_node->id());
    $query->leftJoin('sm_app', 'u', 'u.id = t.app_id');
    $query->addField('u', 'fcm_token');
    $results = $query->execute();
    $data = [];
    foreach ($results as $result) {
      $data[] = $result->fcm_token;
    }
    return $data;
  }

  /**
   * Sets module new values.
   */
  public function setNewValues(NodeInterface $module_node, bool $relay_state, int $led_1_value = 0, int $led_2_value = 0, int $led_3_value = 0) {
    $new_values_p_values = $module_node->get('field_new_values')->getValue();
    if (isset($new_values_p_values[0]['target_id'])) {
      $new_values_p = $this->database->paragraphStorage->load($new_values_p_values[0]['target_id']);
      $new_values_p->set('field_relay', $relay_state);
      $new_values_p->set('field_led_1', $led_1_value);
      $new_values_p->set('field_led_2', $led_2_value);
      $new_values_p->set('field_led_3', $led_3_value);
      try {
        $new_values_p->save();
      }
      catch (EntityStorageException $e) {
        return FALSE;
      }
    }
    $this->database->connection->update('sm_status')->fields(['pin_change' => SM_STATUS_PROCESSING])->condition('nid', $module_node->id())->execute();
    return TRUE;
  }

  /**
   * Returns last values of passed module.
   */
  public function getLastValues(NodeInterface $module_node) {
    $query = $this->database->connection->select('sm_last_values', 't');
    $query->condition('t.nid', $module_node->id());
    $query->addField('t', 'acc_time');
    $query->addField('t', 'relay');
    $query->addField('t', 'ssid');
    $query->addField('t', 'rssi');
    $query->addField('t', 'fw_ver');
    $query->addField('t', 'rst_time');
    $query->addField('t', 'ctrl_time');
    $query->addField('t', 'led_1');
    $query->addField('t', 'led_2');
    $query->addField('t', 'led_3');
    $query->addField('t', 'temps');
    $result = $query->execute()->fetchObject();
    $last_values = new ModuleLastValues();
    $last_values->accessDatetime = $result->acc_time ? DrupalDateTime::createFromTimestamp($result->acc_time) : NULL;
    $last_values->relay = $result->relay === NULL ? NULL : !!($result->relay);
    $last_values->ssid = $result->ssid;
    $last_values->rssi = $result->rssi === NULL ? NULL : (int) $result->rssi;
    $last_values->firmwareVersion = $result->fw_ver;
    $last_values->restartDatetime = $result->rst_time ? DrupalDateTime::createFromTimestamp($result->rst_time) : NULL;
    $last_values->controlDatetime = $result->ctrl_time ? DrupalDateTime::createFromTimestamp($result->ctrl_time) : NULL;
    $last_values->led1 = $result->led_1 === NULL ? NULL : (int) $result->led_1;
    $last_values->led2 = $result->led_2 === NULL ? NULL : (int) $result->led_2;
    $last_values->led3 = $result->led_3 === NULL ? NULL : (int) $result->led_3;
    $last_values->temperatures = $result->temps ? unserialize($result->temps) : [];
    return $last_values;
  }

  /**
   * Returns status of passed module.
   */
  public function getStatuses(NodeInterface $module_node) {
    $query = $this->database->connection->select('sm_status', 't');
    $query->condition('t.nid', $module_node->id());
    $query->addField('t', 'pin_change');
    $query->addField('t', 'fw_update');
    $query->addField('t', 'restart');
    $query->addField('t', 'cfg_save');
    $query->addField('t', 'stg_save');
    $query->addField('t', 'net_save');
    $query->addField('t', 'net_scan');
    $query->addField('t', 'connect');
    $result = $query->execute()->fetchObject();
    $statuses = new ModuleStatuses();
    $statuses->pinChange = (int) $result->pin_change;
    $statuses->fwUpdate = (int) $result->fw_update;
    $statuses->restart = (int) $result->restart;
    $statuses->cfgSave = (int) $result->cfg_save;
    $statuses->stgSave = (int) $result->stg_save;
    $statuses->netSave = (int) $result->net_save;
    $statuses->netScan = (int) $result->net_scan;
    $statuses->connect = (int) $result->connect;
    return $statuses;
  }

  /**
   * Returns access log data of passed module.
   */
  public function getAccessLog(NodeInterface $module_node, DrupalDateTime $offset_datetime, $online_items_limit = 100) {
    $query = $this->database->connection->select('sm_log_access', 't');
    $query->condition('t.nid', $module_node->id());
    if ($offset_datetime) {
      $query->condition('t.start_time', $offset_datetime->getTimestamp(), '<=');
    }
    $query->addField('t', 'start_time');
    $query->addField('t', 'end_time');
    $query->addField('t', 'ssid');
    $query->addField('t', 'restart_cnt');
    $query->orderBy('start_time', 'DESC');
    $query->range(0, $online_items_limit);
    $results = $query->execute();
    $log = [];
    $last_start_datetime = NULL;
    foreach ($results as $result) {
      $start_datetime = $result->start_time ? DrupalDateTime::createFromTimestamp($result->start_time) : NULL;
      $end_datetime = $result->end_time ? DrupalDateTime::createFromTimestamp($result->end_time) : NULL;
      if ($start_datetime && $end_datetime) {
        $last_start_datetime = $last_start_datetime ? $last_start_datetime : $offset_datetime;
        if ($last_start_datetime->getTimestamp() - $end_datetime->getTimestamp() > SM_MAX_ACCESS_DELAY_TIME) {
          $log[] = (object) [
            'isOnline' => FALSE,
            'startDatetime' => $end_datetime,
            'endDatetime' => $last_start_datetime,
            'ssid' => NULL,
            'restartCnt' => NULL,
          ];
        }
        $log[] = (object) [
          'isOnline' => TRUE,
          'startDatetime' => $start_datetime,
          'endDatetime' => $end_datetime,
          'ssid' => $result->ssid,
          'restartCnt' => (int) $result->restart_cnt,
        ];
        $last_start_datetime = $start_datetime;
      }
    }
    return $log;
  }

  /**
   * Returns relay state log data of passed module.
   */
  public function getRelayStateLog(NodeInterface $module_node, DrupalDateTime $offset_datetime, $online_items_limit = 100) {
    $query = $this->database->connection->select('sm_log_relay_state', 't');
    $query->condition('t.nid', $module_node->id());
    if ($offset_datetime) {
      $query->condition('t.start_time', $offset_datetime->getTimestamp(), '<=');
    }
    $query->addField('t', 'start_time');
    $query->addField('t', 'end_time');
    $query->addField('t', 'state');
    $query->orderBy('start_time', 'DESC');
    $query->range(0, $online_items_limit);
    $results = $query->execute();
    $log = [];
    $last_start_datetime = NULL;
    foreach ($results as $result) {
      $start_datetime = $result->start_time ? DrupalDateTime::createFromTimestamp($result->start_time) : NULL;
      $end_datetime = $result->end_time ? DrupalDateTime::createFromTimestamp($result->end_time) : NULL;
      if ($start_datetime && $end_datetime) {
        $last_start_datetime = $last_start_datetime ? $last_start_datetime : $offset_datetime;
        if ($last_start_datetime->getTimestamp() - $end_datetime->getTimestamp() > SM_MAX_ACCESS_DELAY_TIME) {
          $log[] = (object) [
            'isOnline' => FALSE,
            'startDatetime' => $end_datetime,
            'endDatetime' => $last_start_datetime,
            'state' => NULL,
          ];
        }
        $log[] = (object) [
          'isOnline' => TRUE,
          'startDatetime' => $start_datetime,
          'endDatetime' => $end_datetime,
          'state' => $result->state === NULL ? NULL : (int) $result->state,
        ];
        $last_start_datetime = $start_datetime;
      }
    }
    return $log;
  }

  /**
   * Returns temperature log data of passed module.
   */
  public function getTemperatureLog(NodeInterface $module_node, DrupalDateTime $offset_datetime, $items_limit = 100) {
    $query = $this->database->connection->select('sm_log_temperatures', 't');
    $query->condition('t.nid', $module_node->id());
    if ($offset_datetime) {
      $query->condition('t.time', $offset_datetime->getTimestamp(), '<=');
    }
    $query->addField('t', 'time');
    $query->addField('t', 'temperatures');
    $query->orderBy('time', 'DESC');
    $query->range(0, $items_limit);
    $results = $query->execute();
    $log = [];
    foreach ($results as $result) {
      $datetime = $result->time ? DrupalDateTime::createFromTimestamp($result->time) : NULL;
      $temperatures = unserialize($result->temperatures);
      if ($datetime && $temperatures && is_array($temperatures)) {
        $log[] = (object) [
          'temperatures' => $temperatures,
          'datetime' => $datetime,
        ];
      }
    }
    return $log;
  }

  /**
   * Returns TRUE if passed module is online.
   */
  public function isOnline(?DrupalDateTime $access_datetime, ?DrupalDateTime $current_datetime) {
    if ($access_datetime && $current_datetime) {
      return $current_datetime->getTimestamp() - $access_datetime->getTimestamp() < SM_MIN_ACCESS_DELAY_TIME;
    }
    return NULL;
  }

  /**
   * Returns TRUE if control of passed module is online.
   */
  public function isControlOnline(?DrupalDateTime $last_control_datetime, ?DrupalDateTime $current_datetime) {
    if ($last_control_datetime && $current_datetime) {
      return $current_datetime->getTimestamp() - $last_control_datetime->getTimestamp() < SM_MAX_CONTROL_DELAY_TIME;
    }
    return NULL;
  }

  /**
   * Returns RSSI level.
   */
  public function getRssiLevel(?int $value) {
    if ($value >= -80) {
      return 3;
    }
    elseif ($value >= -90) {
      return 2;
    }
    elseif ($value >= -100) {
      return 1;
    }
    return NULL;
  }

  /**
   * Returns RSSI percentages.
   */
  public function getRssiPercentage(?int $value) {
    if ($value <= 0) {
      return min(max(2 * ($value + 100), 0), 100);
    }
    return NULL;
  }

  /**
   * Returns encryption type.
   */
  public function getEncryptionType(?int $value) {
    switch ($value) {
      case 0:
        return 'NONE';

      case 1:
        return 'WEP';

      case 2:
        return 'WPA';

      case 3:
        return 'WPA2';

      case 4:
        return 'AUTO';
    }
    return NULL;
  }

  /**
   * Returns HTML availability.
   */
  public function printIsOnline(?DrupalDateTime $access_datetime, ?DrupalDateTime $current_datetime) {
    if ($access_datetime && $current_datetime) {
      $is_online = self::isOnline($access_datetime, $current_datetime);
      return $is_online ? '<span class="bold color-success">ONLINE</span>' : '<span class="bold color-danger">OFFLINE</span>';
    }
    return NULL;
  }

  /**
   * Returns HTML last access.
   */
  public function printLastAccess(?DrupalDateTime $access_datetime, ?DrupalDateTime $current_datetime) {
    if ($access_datetime && $current_datetime) {
      return 'před ' . self::printDatetimeInterval($access_datetime, $current_datetime, 2) . ' <span class="note">(' . $access_datetime->format(SM_DATETIME_DEFAULT_FORMAT) . ')</span>';
    }
    return NULL;
  }

  /**
   * Returns HTML relay value.
   */
  public function printRelayValue(?bool $relay_value) {
    if ($relay_value === TRUE || $relay_value === FALSE) {
      return $relay_value ? '<span class="bold color-warning">ZAPNUTO</span>' : '<span class="bold color-success">VYPNUTO</span>';
    }
    return 'NULL';
  }

  /**
   * Returns HTML RSSI.
   */
  public function printRssi(?int $value) {
    $result = $value . ' dBm (' . self::getRssiPercentage($value) . ' %)';
    $level = self::getRssiLevel($value);
    if ($level == 3) {
      return '<span class="bold color-success">' . $result . '</span>';
    }
    elseif ($level == 2) {
      return '<span class="bold color-warning">' . $result . '</span>';
    }
    elseif ($level == 1) {
      return '<span class="bold color-danger">' . $result . '</span>';
    }
    return NULL;
  }

  /**
   * Returns HTML LED value.
   */
  public function printLedValue(?int $value) {
    if ($value >= 0) {
      return $value . '/1023 (' . ceil($value / 1023 * 100) . ' %)';
    }
    return NULL;
  }

  /**
   * Returns HTML status.
   */
  public function printStatus(?int $value) {
    switch ($value) {
      case 0:
        return '<span class="color-success">OK</span>';

      case 1:
        return '<span class="color-warning">Zpracovává se</span>';

      case 2:
        return '<span class="color-warning">Ověřuje se</span>';

      case 3:
        return '<span class="color-danger">Došlo k chybě!</span>';
    }
    return NULL;
  }

  /**
   * Returns HTML temperatures.
   */
  public function printTemperatures(array $temperatures) {
    if ($temperatures) {
      $temperatures_html = '';
      foreach ($temperatures as $key => $temperature) {
        $temperatures_html .= ($key === array_key_first($temperatures) ? '' : ' | ') . round($temperature) . ' °C';
      }
      return $temperatures_html;
    }
    return NULL;
  }

  /**
   * Returns HTML encryption type.
   */
  public function printEncryptionType(?int $value) {
    switch ($value) {
      case 0:
        return '<i>žádné</i>';

      case 1:
        return 'WEP';

      case 2:
        return 'WPA';

      case 3:
        return 'WPA2';

      case 4:
        return '<i>auto</i>';
    }
    return NULL;
  }

  /**
   * Returns HTML datetime interval.
   */
  public function printDatetimeInterval(?DrupalDateTime $start_datetime, ?DrupalDateTime $end_datetime, int $units_cnt = 4) {
    if ($start_datetime && $end_datetime) {
      $seconds = $end_datetime->getTimestamp() - $start_datetime->getTimestamp();
      $result = '';
      $d = floor($seconds / 86400);
      $result .= $d ? $d . '&nbsp;d ' : '';
      if ($units_cnt == 1 && $d) {
        return trim($result);
      }
      $seconds %= 86400;
      $h = floor($seconds / 3600);
      $result .= $d || $h ? $h . '&nbsp;h ' : '';
      if (($units_cnt == 1 && $h) || ($units_cnt == 2 && $d)) {
        return trim($result);
      }
      $seconds %= 3600;
      $m = floor($seconds / 60);
      $result .= $d || $h || $m ? $m . '&nbsp;m ' : '';
      if (($units_cnt == 1 && $m) || ($units_cnt == 2 && $h) || ($units_cnt == 3 && $d)) {
        return trim($result);
      }
      $seconds %= 60;
      $s = floor($seconds);
      $result .= $s . '&nbsp;s';
      return $result;
    }
    return NULL;
  }

}
