<?php

namespace Drupal\sm\Service;

use DateTimeZone;
use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Exception;
use Mpdf\QrCode\Output\Png;
use Mpdf\QrCode\QrCode;

/**
 * Class Database.
 */
class Database {

  /**
   * Node storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  public EntityStorageInterface $nodeStorage;

  /**
   * Paragraph storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  public EntityStorageInterface $paragraphStorage;

  /**
   * Term storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  public EntityStorageInterface $termStorage;

  /**
   * Database connection service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  public Connection $connection;

  /**
   * File system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * Constructor.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, Connection $connection, FileSystemInterface $file_system) {
    $this->nodeStorage = $entity_type_manager->getStorage('node');
    $this->paragraphStorage = $entity_type_manager->getStorage('paragraph');
    $this->termStorage = $entity_type_manager->getStorage('taxonomy_term');
    $this->connection = $connection;
    $this->fileSystem = $file_system;
  }

  /**
   * Generates new module with unique ID and tokens.
   */
  public function generateModule() {
    $device_id = $this->getUniqueDeviceId();
    $device_token = $this->getUniqueDeviceToken();
    $control_token = $this->getUniqueControlToken();
    $status_token = $this->getUniqueStatusToken();
    /** @var \Drupal\node\NodeInterface $node */
    $node = $this->nodeStorage->create([
      'type' => SM_BUNDLE_MODULE,
      'title' => 'Module ' . $device_id,
      'field_device_id' => $device_id,
      'field_device_token' => $device_token,
      'field_control_token' => $control_token,
      'field_status_token' => $status_token,
    ]);
    try {
      $node->save();

      $paragraph_values = [];
      $paragraph = $this->paragraphStorage->create([
        'type' => SM_PARAGRAPH_MODULE_NEW_VALUES,
        'field_relay' => FALSE,
        'field_firmware' => NULL,
        'field_connect_ssid' => '',
        'field_connect_password' => '',
        'field_led_1' => 0,
        'field_led_2' => 0,
        'field_led_3' => 0,
      ]);
      $paragraph->save();
      $paragraph_values[] = ['target_id' => $paragraph->id(), 'target_revision_id' => $paragraph->getRevisionId()];
      $node->set('field_new_values', $paragraph_values);

      $paragraph_values = [];
      $paragraph = $this->paragraphStorage->create([
        'type' => SM_PARAGRAPH_MODULE_CONFIGURATION,
        'field_server_password' => $this->generateRandomNumber(8, TRUE),
        'field_pin_button' => 0,
        'field_pin_relay' => 12,
        'field_pin_led' => 13,
        'field_pin_ext_button' => SM_PIN_NONE,
        'field_pin_led_1' => SM_PIN_NONE,
        'field_pin_led_2' => SM_PIN_NONE,
        'field_pin_led_3' => SM_PIN_NONE,
        'field_pin_temp_sensors' => SM_PIN_NONE,
        'field_temp_sensors' => NULL,
      ]);
      $paragraph->save();
      $paragraph_values[] = ['target_id' => $paragraph->id(), 'target_revision_id' => $paragraph->getRevisionId()];
      $node->set('field_configuration', $paragraph_values);

      $paragraph_values = [];
      $paragraph = $this->paragraphStorage->create([
        'type' => SM_PARAGRAPH_MODULE_SETTINGS,
        'field_button_mode' => FALSE,
        'field_default_state' => FALSE,
        'field_state_delay' => 60000,
        'field_loop_delay' => 10000,
        'field_default_led_1' => 102,
        'field_default_led_2' => 102,
        'field_default_led_3' => 102,
      ]);
      $paragraph->save();
      $paragraph_values[] = ['target_id' => $paragraph->id(), 'target_revision_id' => $paragraph->getRevisionId()];
      $node->set('field_settings', $paragraph_values);

      $paragraph_values = [];
      for ($i = 0; $i < SM_MAX_NETWORK_CNT; $i++) {
        $paragraph = $this->paragraphStorage->create([
          'type' => SM_PARAGRAPH_MODULE_NETWORK_SAVE,
          'field_ssid' => '',
          'field_password' => '',
        ]);
        $paragraph->save();
        $paragraph_values[] = ['target_id' => $paragraph->id(), 'target_revision_id' => $paragraph->getRevisionId()];
      }
      $node->set('field_network_save', $paragraph_values);

      $paragraph_values = [];
      for ($i = 0; $i < SM_MAX_NETWORK_CNT; $i++) {
        $paragraph = $this->paragraphStorage->create([
          'type' => SM_PARAGRAPH_MODULE_NETWORK_SCAN,
          'field_ssid' => '',
          'field_rssi' => NULL,
          'field_bssid' => '',
          'field_channel' => NULL,
          'field_encryption_type' => NULL,
        ]);
        $paragraph->save();
        $paragraph_values[] = ['target_id' => $paragraph->id(), 'target_revision_id' => $paragraph->getRevisionId()];
      }
      $node->set('field_network_scan', $paragraph_values);

      try {
        $this->connection->insert('sm_status')->fields(['nid' => $node->id()])->execute();
        $this->connection->insert('sm_last_values')->fields(['nid' => $node->id()])->execute();
      }
      catch (Exception $e) {
      }

      $qr_code = new QrCode('PLIX_SM_' . $control_token);
      $output = new Png();
      $private_directory = 'private://module/qr';
      $this->fileSystem->prepareDirectory($private_directory, FileSystemInterface::CREATE_DIRECTORY);
      $file_location = $private_directory . '/' . $device_id . '.png';
      $output_data = $output->output($qr_code, 400, [255, 255, 255], [0, 0, 0]);
      $file = file_save_data($output_data, $file_location, FileSystemInterface::EXISTS_REPLACE);
      $node->set('field_qr_code_file', $file);

      $node->save();
      return $node;
    }
    catch (EntityStorageException $e) {
      return NULL;
    }
  }

  /**
   * Generates new temperature sensor with unique ID.
   */
  public function generateTemperatureSensor() {
    $device_id = $this->getUniqueDeviceId();
    /** @var \Drupal\node\NodeInterface $node */
    $node = $this->nodeStorage->create([
      'type' => SM_BUNDLE_TEMPERATURE_SENSOR,
      'title' => 'Temperature sensor ' . $device_id,
      'field_device_id' => $device_id,
    ]);
    try {
      $node->save();
      return $node;
    }
    catch (EntityStorageException $e) {
      return NULL;
    }
  }

  /**
   * Clears old logs.
   */
  public function clearLogs() {
    $current_datetime = new DrupalDateTime('NOW');
    $threshold_timestamp = $current_datetime->getTimestamp() - SM_MAX_DATABASE_LOG_AGE;
    $this->connection->delete('sm_log_access')->condition('end_time', $threshold_timestamp, '<')->execute();
    $this->connection->delete('sm_log_relay_state')->condition('end_time', $threshold_timestamp, '<')->execute();
    $this->connection->delete('sm_log_temperatures')->condition('time', $threshold_timestamp, '<')->execute();

    $query = $this->nodeStorage->getQuery();
    $query->condition('type', 'module');
    $module_node_ids = $query->execute();

    foreach ($module_node_ids as $module_node_id) {
      $query = $this->connection->select('sm_log_access', 't');
      $query->condition('t.nid', $module_node_id);
      $query->addField('t', 'end_time');
      $query->orderBy('t.start_time', 'desc');
      $query->range(SM_MAX_DATABASE_LOG_ITEMS_PER_MODULE - 1, 1);
      $result = $query->execute()->fetchObject();
      if ($result && $result->end_time) {
        $this->connection->delete('sm_log_access')->condition('nid', $module_node_id)->condition('end_time', $result->end_time, '<')->execute();
      }

      $query = $this->connection->select('sm_log_relay_state', 't');
      $query->condition('t.nid', $module_node_id);
      $query->addField('t', 'end_time');
      $query->orderBy('t.start_time', 'desc');
      $query->range(SM_MAX_DATABASE_LOG_ITEMS_PER_MODULE - 1, 1);
      $result = $query->execute()->fetchObject();
      if ($result && $result->end_time) {
        $this->connection->delete('sm_log_relay_state')->condition('nid', $module_node_id)->condition('end_time', $result->end_time, '<')->execute();
      }

      $query = $this->connection->select('sm_log_temperatures', 't');
      $query->condition('t.nid', $module_node_id);
      $query->addField('t', 'time');
      $query->orderBy('t.time', 'desc');
      $query->range(SM_MAX_DATABASE_LOG_ITEMS_PER_MODULE - 1, 1);
      $result = $query->execute()->fetchObject();
      if ($result && $result->time) {
        $this->connection->delete('sm_log_temperatures')->condition('nid', $module_node_id)->condition('end_time', $result->time, '<')->execute();
      }
    }
  }

  /**
   * Returns target ID from the passed field.
   */
  public function getTargetIdFromField($field) {
    if (is_array($field) && isset($field[0]) && isset($field[0]['target_id'])) {
      return $field[0]['target_id'];
    }
    else {
      if (isset($field) && isset($field['target_id'])) {
        return $field['target_id'];
      }
    }
    return NULL;
  }

  /**
   * Returns value from the passed field.
   */
  public function getValueFromField(array $field) {
    if (isset($field) && isset($field[0]) && isset($field[0]['value'])) {
      return $field[0]['value'];
    }
    else {
      if (isset($field) && isset($field['value'])) {
        return $field['value'];
      }
    }
    return NULL;
  }

  /**
   * Returns unique (non-existent) device ID.
   */
  private function getUniqueDeviceId() {
    while (TRUE) {
      $device_id = $this->generateRandomString(8, TRUE);
      $query = $this->nodeStorage->getQuery();
      $query->condition('field_device_id', $device_id);
      if (!$query->execute()) {
        return $device_id;
      }
    }
  }

  /**
   * Returns unique (non-existent) device token.
   */
  private function getUniqueDeviceToken() {
    while (TRUE) {
      $device_token = $this->generateRandomString(48);
      $query = $this->nodeStorage->getQuery();
      $query->condition('field_device_token', $device_token);
      if (!$query->execute()) {
        return $device_token;
      }
    }
  }

  /**
   * Returns unique (non-existent) control token.
   */
  private function getUniqueControlToken() {
    while (TRUE) {
      $control_token = $this->generateRandomString(48);
      $query = $this->nodeStorage->getQuery();
      $query->condition('field_control_token', $control_token);
      if (!$query->execute()) {
        return $control_token;
      }
    }
  }

  /**
   * Returns unique (non-existent) status token.
   */
  private function getUniqueStatusToken() {
    while (TRUE) {
      $status_token = $this->generateRandomString(16);
      $query = $this->nodeStorage->getQuery();
      $query->condition('field_status_token', $status_token);
      if (!$query->execute()) {
        return $status_token;
      }
    }
  }

  /**
   * Returns random string.
   */
  private function generateRandomString($length, $easy_to_read = FALSE) {
    $char_set = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    if ($easy_to_read) {
      $char_set = '23456789abcdefghjkmnpqrstuvwxyzABCDEFGHJKMNPQRSTUVWXYZ';
    }
    $char_cnt = strlen($char_set);
    $random_string = '';
    for ($i = 0; $i < $length; $i++) {
      try {
        $random_string .= $char_set[random_int(0, $char_cnt - 1)];
      }
      catch (Exception $e) {
        return NULL;
      }
    }
    return $random_string;
  }

  /**
   * Returns random number.
   */
  private function generateRandomNumber($length, $easy_to_read = FALSE) {
    $char_set = '0123456789';
    if ($easy_to_read) {
      $char_set = '23456789';
    }
    $char_cnt = strlen($char_set);
    $random_string = '';
    for ($i = 0; $i < $length; $i++) {
      try {
        $random_string .= $char_set[random_int(0, $char_cnt - 1)];
      }
      catch (Exception $e) {
        return NULL;
      }
    }
    return $random_string;
  }

  /**
   * Returns current date formatted for database query.
   */
  private function getCurrentDate() {
    $current_user_timezone = new DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setTime(0, 0, 0);
    $datetime->setTimezone(new DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

  /**
   * Returns current datetime formatted for database query.
   */
  private function getCurrentDatetime() {
    $current_user_timezone = new DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setTimezone(new DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

  /**
   * Returns date formatted for database query.
   */
  private function getDate(int $year, int $month, int $day) {
    $current_user_timezone = new \DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setDate($year, $month, $day);
    $datetime->setTime(0, 0, 0);
    $datetime->setTimezone(new \DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

  /**
   * Returns datetime formatted for database query.
   */
  private function getDatetime(int $year, int $month, int $day, int $hour, int $minute, int $second) {
    $current_user_timezone = new \DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setDate($year, $month, $day);
    $datetime->setTime($hour, $minute, $second);
    $datetime->setTimezone(new \DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

}
