<?php

namespace Drupal\sm\Service\Validator;

use Drupal\Core\Form\FormStateInterface;
use Drupal\node\NodeInterface;
use Drupal\sm\Service\Database;

/**
 * Class ModuleNodeValidator.
 */
class ModuleNodeValidator {

  /**
   * Database service.
   *
   * @var \Drupal\sm\Service\Database
   */
  protected Database $database;

  /**
   * Constructor.
   */
  public function __construct(Database $database) {
    $this->database = $database;
  }

  /**
   * Validates all.
   */
  public function validateAll(FormStateInterface $form_state) {
    /** @var \Drupal\node\NodeInterface $module_node */
    $module_node = $form_state->getFormObject()->getEntity();
    if (!$module_node->id()) {
      return;
    }
    if ($form_state->getValue('checkbox_no_process')) {
      return;
    }
    $fields = [];
    if ($form_state->getValue('checkbox_restart_module')) {
      $fields['restart'] = SM_STATUS_PROCESSING;
    }
    if ($form_state->getValue('checkbox_scan_networks')) {
      $fields['net_scan'] = SM_STATUS_PROCESSING;
    }
    $firmware_nid = $this->database->getTargetIdFromField($form_state->getValue('field_firmware'));
    if ($firmware_nid && $firmware_nid != $module_node->get('field_firmware')->target_id) {
      $fields['fw_update'] = SM_STATUS_PROCESSING;
    }
    $connection_ssid = $this->database->getValueFromField($form_state->getValue('field_connect_ssid'));
    if ($connection_ssid) {
      $connection_password = $this->database->getValueFromField($form_state->getValue('field_connect_password'));
      if ($connection_ssid != $module_node->get('field_connect_ssid')->value || $connection_password != $module_node->get('field_connect_password')->value) {
        $fields['connect'] = SM_STATUS_PROCESSING;
      }
    }
    if ($this->fieldChanged($form_state, $module_node, 'field_new_values')) {
      $fields['pin_change'] = SM_STATUS_PROCESSING;
    }
    if ($this->fieldChanged($form_state, $module_node, 'field_configuration')) {
      $fields['cfg_save'] = SM_STATUS_PROCESSING;
    }
    if ($this->fieldChanged($form_state, $module_node, 'field_settings')) {
      $fields['stg_save'] = SM_STATUS_PROCESSING;
    }
    if ($this->fieldChanged($form_state, $module_node, 'field_network_save')) {
      $fields['net_save'] = SM_STATUS_PROCESSING;
    }
    if ($fields) {
      $this->database->connection->update('sm_status')->fields($fields)->condition('nid', $module_node->id())->execute();
    }
  }

  /**
   * Returns TRUE if the passed field has changed.
   */
  private function fieldChanged(FormStateInterface $form_state, NodeInterface $module_node, $root_field_key) {
    $configuration_pf_values = $form_state->getValue($root_field_key);
    $configuration_p_values = $module_node->get($root_field_key)->getValue();
    foreach ($configuration_pf_values as $configuration_pf_key => $configuration_pf_value) {
      if (isset($configuration_pf_value['subform']) && isset($configuration_p_values[$configuration_pf_key]['target_id'])) {
        if ($configuration_p = $this->database->paragraphStorage->load($configuration_p_values[$configuration_pf_key]['target_id'])) {
          if ($configuration_ps_value = $configuration_pf_value['subform']) {
            foreach ($configuration_ps_value as $field_key => $field_value) {
              if (is_array($configuration_ps_value[$field_key])) {
                $field_value_old = $configuration_p->get($field_key)->getValue();
                if (isset($field_value['value'])) {
                  // Check changes in boolean fields.
                  if ($field_value['value'] != $field_value_old[0]['value']) {
                    return TRUE;
                  }
                }
                elseif (isset($field_value[0]['value'])) {
                  // Check changes in value fields.
                  foreach ($field_value as $field_s_key => $field_s_value) {
                    if (is_array($field_s_value)) {
                      $value = isset($field_s_value['value']) ? $field_s_value['value'] : NULL;
                      $value_old = isset($field_value_old[$field_s_key]['value']) ? $field_value_old[$field_s_key]['value'] : NULL;
                      if ($value != $value_old) {
                        return TRUE;
                      }
                    }
                  }
                }
                elseif (isset($field_value[0]['target_id'])) {
                  // Check changes in reference fields.
                  foreach ($field_value as $field_s_key => $field_s_value) {
                    if (is_array($field_s_value)) {
                      $target_id = isset($field_s_value['target_id']) ? $field_s_value['target_id'] : NULL;
                      $target_id_old = isset($field_value_old[$field_s_key]['target_id']) ? $field_value_old[$field_s_key]['target_id'] : NULL;
                      if ($target_id != $target_id_old) {
                        return TRUE;
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    return FALSE;
  }

}
