<?php

namespace Drupal\tarabka\Plugin\Field\FieldWidget;

use Drupal;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\paragraphs\Plugin\Field\FieldWidget\InlineParagraphsWidget;
use Drupal\tarabka\Service\Database;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'tarabka_entity_reference_paragraphs' widget.
 *
 * @FieldWidget(
 *   id = "tarabka_entity_reference_paragraphs",
 *   label = @Translation("Tarabka Paragraphs Classic"),
 *   description = @Translation("Extended paragraphs inline form widget primarily for the Tarabka module."),
 *   field_types = {
 *     "entity_reference_revisions"
 *   }
 * )
 */
class TarabkaInlineParagraphsWidget extends InlineParagraphsWidget {

  /**
   * Entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Database service.
   *
   * @var \Drupal\tarabka\Service\Database
   */
  protected Database $database;

  /**
   * Constructor.
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, array $third_party_settings, EntityTypeManagerInterface $entity_type_manager, Database $database) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings);
    $this->entityTypeManager = $entity_type_manager;
    $this->database = $database;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager */
    $entity_type_manager = $container->get('entity_type.manager');
    /** @var \Drupal\tarabka\Service\Database $database */
    $database = $container->get('tarabka.database');
    return new static($plugin_id, $plugin_definition, $configuration['field_definition'], $configuration['settings'], $configuration['third_party_settings'], $entity_type_manager, $database);
  }

  /**
   * Shows minimum number of fields.
   *
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function formMultipleElements(FieldItemListInterface $items, array &$form, FormStateInterface $form_state) {
    $field_name = $this->fieldDefinition->getName();
    $this->fieldParents = $form['#parents'];
    $field_state = static::getWidgetState($this->fieldParents, $field_name, $form_state);

    $max = $field_state['items_count'];

    $min_count = $this->getSetting('tarabka_min_count');

    // Special custom min count for a specific page.
    $route_name = Drupal::routeMatch()->getRouteName();
    if ($field_name == 'field_hunting_permits' && $route_name == 'node.add') {
      $min_count = $this->database->countHunterNodes(NULL, TRUE, NULL);
    }

    $min_count_always_generated = $this->getSetting('tarabka_min_count_always_generated');
    if ($min_count_always_generated && !isset($field_state['tarabka_min_count_always_generated_applied'])) {
      $min_count = $max + $min_count;
      $field_state['tarabka_min_count_always_generated_applied'] = TRUE;
    }

    if ($max < $min_count) {
      $default_type = $this->getDefaultParagraphTypeMachineName();

      // Checking if default_type is not none and if is allowed.
      if ($default_type) {

        // Place the default paragraphs.
        $target_type = $this->getFieldSetting('target_type');
        while ($max < $min_count) {
          /** @var \Drupal\paragraphs\ParagraphInterface $paragraphs_entity */
          $paragraphs_entity = $this->entityTypeManager->getStorage($target_type)->create([
            'type' => $default_type,
          ]);
          $paragraphs_entity->setParentEntity($items->getEntity(), $field_name);
          $field_state['selected_bundle'] = $default_type;
          $display = EntityFormDisplay::collectRenderDisplay($paragraphs_entity, $this->getSetting('form_display_mode'));
          $field_state['paragraphs'][$max] = [
            'entity' => $paragraphs_entity,
            'display' => $display,
            'mode' => 'edit',
            'original_delta' => $max + 1,
          ];
          $max++;
        }
        $field_state['items_count'] = $max;
      }
      static::setWidgetState($this->fieldParents, $field_name, $form_state, $field_state);
    }
    return parent::formMultipleElements($items, $form, $form_state);
  }

  /**
   * Sets default value for minimum number of fields.
   *
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    $settings = parent::defaultSettings();

    $settings['tarabka_min_count'] = 1;

    $settings['tarabka_min_count_always_generated'] = 0;

    return $settings;
  }

  /**
   * Populate settings form to input minimum number of fields value.
   *
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $elements = parent::settingsForm($form, $form_state);

    $elements['tarabka_min_count'] = [
      '#type' => 'number',
      '#title' => $this->t('Minimum number of fields'),
      '#default_value' => $this->getSetting('tarabka_min_count'),
      '#required' => TRUE,
      '#min' => 1,
      '#description' => $this->t('Set minimum number of fields.'),
    ];

    $elements['tarabka_min_count_always_generated'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Always generate minimum number of fields?'),
      '#default_value' => $this->getSetting('tarabka_min_count_always_generated'),
      '#description' => $this->t('Check if you want to always generate minimum number of fields as new empty fields.'),
    ];

    return $elements;
  }

  /**
   * Shows settings summary for fields at form Manage Form Display.
   *
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();

    $min_count = $this->getSetting('tarabka_min_count');
    if (is_numeric($min_count)) {
      $summary[] = $this->t('Minimum number of fields : @mincount', ['@mincount' => $min_count]);
    }

    $min_count_always_generated = $this->getSetting('tarabka_min_count_always_generated');
    if ($min_count_always_generated == TRUE) {
      $summary[] = $this->t('Always generate minimum number of fields as new empty fields.');
    }
    else {
      $summary[] = $this->t('Only generate the minimum number of fields for the first time.');
    }

    return $summary;
  }

}
