<?php

namespace Drupal\tarabka\Service;

use DateTimeZone;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;

/**
 * Class Base.
 */
class Base {

  /**
   * Formats address.
   */
  public function formatAddress($street, $city, $postal_code, $html = FALSE) {
    $address = '';
    if ($city) {
      if ($street) {
        $address .= $street . ($html ? ',<br>' : ', ');
      }
      if ($postal_code) {
        $address .= $this->formatPostalCode($postal_code) . ' ';
      }
      $address .= $city;
    }
    return $address;
  }

  /**
   * Returns link to address on Google Maps.
   */
  public function getAddressGoogleMapsLink($street, $city, $postal_code) {
    return 'https://maps.google.com/?q=' . urlencode($street . ', ' . $postal_code . ' ' . $city);
  }

  /**
   * Returns link to address on Mapy.cz.
   */
  public function getAddressMapyCzLink($street, $city, $postal_code) {
    return 'https://mapy.cz/?q=' . urlencode($street . ', ' . $postal_code . ' ' . $city);
  }

  /**
   * Formats postal code.
   */
  public function formatPostalCode($postal_code) {
    if (strlen($postal_code) == 5) {
      return substr($postal_code, 0, 3) . ' ' . substr($postal_code, 3, 2);
    }
    return $postal_code;
  }

  /**
   * Formats phone number.
   */
  public function formatPhoneNumber($phone_number, $with_calling_code = FALSE, $html = FALSE) {
    if ($phone_number && strlen($phone_number) == 13 && substr($phone_number, 0, 4) == '+420') {
      $phone_number_text = substr($phone_number, 4, 3) . ' ' . substr($phone_number, 7, 3) . ' ' . substr($phone_number, 10, 3);
      if ($with_calling_code) {
        $phone_number_text = substr($phone_number, 0, 4) . ' ' . $phone_number_text;
      }
      if ($html) {
        return '<a href="tel:' . $phone_number . '">' . $phone_number_text . '</a>';
      }
      return $phone_number_text;
    }
    return $phone_number;
  }

  /**
   * Formats email.
   */
  public function formatEmail($email, $html = FALSE) {
    if ($email && $html) {
      return '<a href="mailto:' . $email . '">' . $email . '</a>';
    }
    return $email;
  }

  /**
   * Returns age.
   */
  public function getAge($date_of_birth) {
    if ($date_of_birth instanceof DrupalDateTime) {
      $current_datetime = new DrupalDateTime('NOW');
      $datetime_diff = $current_datetime->diff($date_of_birth);
      return $datetime_diff->y;
    }
    return NULL;
  }

  /**
   * Formats datetime.
   */
  public function formatDatetime($datetime, $format, $with_age = FALSE) {
    if ($datetime instanceof DrupalDateTime) {
      if ($with_age) {
        return $datetime->format($format) . ' (' . $this->getAge($datetime) . ' let)';
      }
      return $datetime->format($format);
    }
    return NULL;
  }

  /**
   * Validates e-mail.
   */
  public function validateEmail($email) {
    return filter_var($email, FILTER_VALIDATE_EMAIL);
  }

  /**
   * Validates past date.
   */
  public function validatePastDate($datetime) {
    if ($datetime instanceof DrupalDateTime) {
      $current_datetime = (new DrupalDateTime())->setTimezone(new DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
      $datetime = new DrupalDateTime($datetime);
      $datetime_diff = $current_datetime->diff($datetime);
      return !!($datetime_diff->invert);
    }
    return NULL;
  }

  /**
   * Validates datetime range.
   */
  public function validateDatetimeRange(DrupalDateTime $start_date, DrupalDateTime $end_date) {
    return $start_date->getTimestamp() <= $end_date->getTimestamp();
  }

  /**
   * Returns string without all whitespaces.
   */
  public function removeWhitespaces($str) {
    return trim(preg_replace('/\s/', '', $str));
  }

  /**
   * Returns string without multiple whitespaces.
   */
  public function removeMultipleWhitespaces($str) {
    return trim(preg_replace('/\s\s+/', ' ', $str));
  }

}
