<?php

namespace Drupal\tarabka\Service\Base;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Render\Markup;
use Drupal\node\NodeInterface;
use Drupal\tarabka\Service\Database;

/**
 * Class HuntingPermitSeriesNode.
 */
class HuntingPermitSeriesNode {

  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected MessengerInterface $messenger;

  /**
   * Database service.
   *
   * @var \Drupal\tarabka\Service\Database
   */
  protected Database $database;

  /**
   * Constructor.
   */
  public function __construct(MessengerInterface $messenger, Database $database) {
    $this->messenger = $messenger;
    $this->database = $database;
  }

  /**
   * Runs before saving the hunting permit series node.
   */
  public function beforeSave(NodeInterface $hunting_permit_series_node) {
    if (!$hunting_permit_series_node) {
      return;
    }
    $this->regenerateTitle($hunting_permit_series_node);
    $this->processHuntingPermits($hunting_permit_series_node);
  }

  /**
   * Runs after saving the hunting permit series node.
   */
  public function afterSave(NodeInterface $hunting_permit_series_node) {
    if (!$hunting_permit_series_node) {
      return;
    }
    $cache_tags = ['node:' . TARABKA_HUNTING_PERMIT_SERIES_LIST_PAGE_ID];
    Cache::invalidateTags($cache_tags);
  }

  /**
   * Returns hunting permit series node page title.
   */
  public function getPageTitle(&$variables, $route_name, NodeInterface $hunting_permit_series_node = NULL) {
    if ($hunting_permit_series_node) {
      switch ($route_name) {
        case 'entity.node.edit_form':
          $variables['title'] = Markup::create('Upravit sérii povolenek <em>' . $hunting_permit_series_node->label() . '</em>');
          return;

        case 'entity.node.delete_form':
          $variables['title'] = Markup::create('Opravdu smazat sérii povolenek <em>' . $hunting_permit_series_node->label() . '</em> ?');
          return;
      }
    }
    else {
      switch ($route_name) {
        case 'node.add':
          $variables['title'] = 'Vytvořit novou sérii povolenek';
          return;
      }
    }
  }

  /**
   * Regenerates hunting permit series node title.
   */
  private function regenerateTitle(NodeInterface $hunting_permit_series_node) {
    $id = $hunting_permit_series_node->get('field_id')->value;
    $start_date = $hunting_permit_series_node->get('field_start_date')->date;
    if ($start_date instanceof DrupalDateTime) {
      $hunting_permit_series_node->setTitle('Série ' . $id . ' | ' . $start_date->format('Y'));
    }
  }

  /**
   * Deletes empty hunting permits.
   */
  private function processHuntingPermits(NodeInterface $hunting_permit_series_node) {
    $hunting_permit_p_values_new = [];
    $hunting_permit_p_values = $hunting_permit_series_node->get('field_hunting_permits')->getValue();
    foreach ($hunting_permit_p_values as $hunting_permit_p_value) {
      $hunting_permit_p = $this->database->paragraphStorage->load($hunting_permit_p_value['target_id']);
      if ($hunter_node = $hunting_permit_p->get('field_hunter')->entity) {
        if ($wild_animal_p_values = $hunting_permit_p->get('field_wild_animals')->getValue()) {
          foreach ($wild_animal_p_values as $wild_animal_p_value) {
            $wild_animal_p = $this->database->paragraphStorage->load($wild_animal_p_value['target_id']);
            $kind = $wild_animal_p->get('field_kind')->value;
            $quantity = $wild_animal_p->get('field_quantity')->value;
            $sex = $wild_animal_p->get('field_sex')->value;
            $age_class = $wild_animal_p->get('field_age_class')->value;
            if ($wild_animal = $this->database->createWildAnimal($kind, $quantity, $sex, $age_class)) {
              $message = Markup::create('Do databáze byla automaticky přidána nová zvěř <a href="/zver/' . $wild_animal->id() . '">' . $wild_animal->label() . '</a>');
              $this->messenger->addStatus($message);
            }
          }
        }
        $hunting_permit_p_values_new[] = $hunting_permit_p;
      }
    }
    usort($hunting_permit_p_values_new, function ($a, $b) {
      $a_hunter_id = $a->get('field_hunter')->entity->get('field_id')->value;
      $b_hunter_id = $b->get('field_hunter')->entity->get('field_id')->value;
      if ($a_hunter_id == $b_hunter_id) {
        return 0;
      }
      return $a_hunter_id < $b_hunter_id ? -1 : 1;
    });
    $hunting_permit_series_node->set('field_hunting_permits', $hunting_permit_p_values_new);
  }

}
