<?php

namespace Drupal\tarabka\Service\Data;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Render\RendererInterface;
use Drupal\node\NodeInterface;
use Drupal\tarabka\Service\Base;
use Drupal\tarabka\Service\Database;
use Drupal\tarabka\Service\Pdf;

/**
 * Class HunterNodeData.
 */
class HunterNodeData {

  /**
   * Renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * Base service.
   *
   * @var \Drupal\tarabka\Service\Base
   */
  protected Base $base;

  /**
   * Database service.
   *
   * @var \Drupal\tarabka\Service\Database
   */
  protected Database $database;

  /**
   * Base service.
   *
   * @var \Drupal\tarabka\Service\Pdf
   */
  protected Pdf $pdf;

  /**
   * Constructor.
   */
  public function __construct(RendererInterface $renderer, Base $base, Database $database, Pdf $pdf) {
    $this->renderer = $renderer;
    $this->base = $base;
    $this->database = $database;
    $this->pdf = $pdf;
  }

  /**
   * Returns hunter data.
   */
  public function getHunter(NodeInterface $hunter_node, $detailed = FALSE) {
    $node_id = $hunter_node->id();
    $id = $hunter_node->get('field_id')->value;
    $first_name = $hunter_node->get('field_first_name')->value;
    $last_name = $hunter_node->get('field_last_name')->value;
    $is_member = $hunter_node->get('field_is_member')->value;
    $street = $hunter_node->get('field_street')->value;
    $city = $hunter_node->get('field_city')->value;
    $postal_code = $hunter_node->get('field_postal_code')->value;
    $email = $hunter_node->get('field_email')->value;
    $phone_number = $hunter_node->get('field_phone_number')->value;
    $date_of_birth = $hunter_node->get('field_date_of_birth')->date;
    $hunting_license_number = $hunter_node->get('field_hunting_license_number')->value;
    $gun_license_expiry_date = $hunter_node->get('field_gun_license_expiry_date')->date;
    $hunting_permits = [];

    if ($detailed) {
      $current_datetime = new DrupalDateTime();
      $hunting_permit_series_nodes = $this->database->getHuntingPermitSeriesNodesByHunterNode($hunter_node);
      foreach ($hunting_permit_series_nodes as $hunting_permit_series_node) {
        $hunting_permit_p_values = $hunting_permit_series_node->get('field_hunting_permits')->getValue();
        foreach ($hunting_permit_p_values as $hunting_permit_p_value) {
          $hunting_permit_p = $this->database->paragraphStorage->load($hunting_permit_p_value['target_id']);
          if ($hunter_node = $hunting_permit_p->get('field_hunter')->entity) {
            if ($hunter_node->id() == $node_id) {
              $start_date = $hunting_permit_series_node->get('field_start_date')->date;
              $end_date = $hunting_permit_series_node->get('field_end_date')->date;
              $wild_animals = [];
              if ($wild_animal_p_values = $hunting_permit_p->get('field_wild_animals')->getValue()) {
                foreach ($wild_animal_p_values as $wild_animal_p_value) {
                  $wild_animal_p = $this->database->paragraphStorage->load($wild_animal_p_value['target_id']);
                  $kind = $wild_animal_p->get('field_kind')->value;
                  $quantity = $wild_animal_p->get('field_quantity')->value;
                  $sex = $wild_animal_p->get('field_sex')->value;
                  $age_class = $wild_animal_p->get('field_age_class')->value;
                  $wild_animals[] = [
                    'kind' => $kind,
                    'quantity' => $quantity,
                    'sex' => $sex,
                    'age_class' => $age_class,
                  ];
                }
              }
              $hunting_permits[] = [
                'label' => $hunting_permit_series_node->getTitle(),
                'date_html' => $start_date->format('j. n. Y') . ' &ndash; ' . $end_date->format('j. n. Y'),
                'is_valid' => $this->base->validateDatetimeRange($start_date, $current_datetime) && $this->base->validateDatetimeRange($current_datetime, $end_date),
                'wild_animals' => $wild_animals,
              ];
            }
          }
        }
      }
    }

    return [
      'id' => $id,
      'first_name' => $first_name,
      'last_name' => $last_name,
      'name' => $first_name . ' ' . $last_name,
      'is_member' => $is_member ? 'ano' : 'ne',
      'address_html' => $this->base->formatAddress($street, $city, $postal_code, TRUE),
      'address_google_maps_link_html' => $this->base->getAddressGoogleMapsLink($street, $city, $postal_code),
      'address_mapy_cz_link_html' => $this->base->getAddressMapyCzLink($street, $city, $postal_code),
      'date_of_birth' => $this->base->formatDatetime($date_of_birth, 'j. n. Y', FALSE),
      'date_of_birth_with_age' => $this->base->formatDatetime($date_of_birth, 'j. n. Y', TRUE),
      'email' => $this->base->formatEmail($email, FALSE),
      'email_html' => $this->base->formatEmail($email, TRUE),
      'phone_number' => $this->base->formatPhoneNumber($phone_number, FALSE, FALSE),
      'phone_number_with_calling_code_html' => $this->base->formatPhoneNumber($phone_number, TRUE, TRUE),
      'hunting_license_number' => $hunting_license_number,
      'gun_license_expiry_date' => $this->base->formatDatetime($gun_license_expiry_date, 'j. n. Y'),
      'gun_license_expiry_date_highlighted' => $this->base->validatePastDate($gun_license_expiry_date),
      'hunting_permits' => $hunting_permits,
      'view_url' => '/myslivci/' . $id,
      'edit_url' => '/node/' . $node_id . '/edit',
      'delete_url' => '/node/' . $node_id . '/delete',
    ];
  }

  /**
   * Returns hunter list data.
   */
  public function getHunterList($for_pdf_export = FALSE) {
    $results = [];
    $is_member = $for_pdf_export ? TRUE : NULL;
    $hunter_nodes = $this->database->getHunterNodes($is_member);
    foreach ($hunter_nodes as $hunter_node) {
      $results[] = $this->getHunter($hunter_node);
    }
    return $results;
  }

  /**
   * Exports hunter list to PDF file.
   */
  public function exportHunterListToPdf() {
    $current_datetime = new DrupalDateTime();
    $file_name = 'seznam-myslivcu-' . $current_datetime->format('Y-m-d') . '.pdf';

    $results = $this->getHunterList(TRUE);
    $elements = [
      '#theme' => 'hunter_list_pdf',
      '#custom' => ['results' => $results, 'current_date' => $current_datetime->format('j. n. Y')],
    ];
    $html_markup = $this->renderer->renderPlain($elements);

    $this->pdf->generate('Seznam členů mysliveckého spolku Tarabka Nasavrky', $file_name, $html_markup, 'hunter-list', ['orientation' => 'L']);
  }

}
