<?php

namespace Drupal\tarabka\Service\Data;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Render\RendererInterface;
use Drupal\node\NodeInterface;
use Drupal\paragraphs\ParagraphInterface;
use Drupal\tarabka\Service\Base;
use Drupal\tarabka\Service\Database;
use Drupal\tarabka\Service\Pdf;

/**
 * Class HuntingPermitSeriesNodeData.
 */
class HuntingPermitSeriesNodeData {

  /**
   * Renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * Config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected ConfigFactoryInterface $config;

  /**
   * Base service.
   *
   * @var \Drupal\tarabka\Service\Base
   */
  protected Base $base;

  /**
   * Database service.
   *
   * @var \Drupal\tarabka\Service\Database
   */
  protected Database $database;

  /**
   * Base service.
   *
   * @var \Drupal\tarabka\Service\Pdf
   */
  protected Pdf $pdf;

  /**
   * Constructor.
   */
  public function __construct(RendererInterface $renderer, ConfigFactoryInterface $config, Base $base, Database $database, Pdf $pdf) {
    $this->renderer = $renderer;
    $this->config = $config;
    $this->base = $base;
    $this->database = $database;
    $this->pdf = $pdf;
  }

  /**
   * Returns hunting permit data.
   */
  public function getHuntingPermit(ParagraphInterface $hunting_permit_p) {
    $hunter_node = $hunting_permit_p->get('field_hunter')->entity;
    if (!$hunter_node) {
      return [];
    }

    $id = $hunter_node->get('field_id')->value;
    $first_name = $hunter_node->get('field_first_name')->value;
    $last_name = $hunter_node->get('field_last_name')->value;
    $street = $hunter_node->get('field_street')->value;
    $city = $hunter_node->get('field_city')->value;
    $postal_code = $hunter_node->get('field_postal_code')->value;
    $hunting_license_number = $hunter_node->get('field_hunting_license_number')->value;
    $hunter = [
      'id' => $id,
      'name' => $first_name . ' ' . $last_name,
      'address_html' => $this->base->formatAddress($street, $city, $postal_code, TRUE),
      'hunting_license_number' => $hunting_license_number,
    ];

    $wild_animals = [];
    $wild_animal_p_values = $hunting_permit_p->get('field_wild_animals')->getValue();
    foreach ($wild_animal_p_values as $wild_animal_p_value) {
      $wild_animal_p = $this->database->paragraphStorage->load($wild_animal_p_value['target_id']);
      $wild_animals[] = [
        'kind' => $wild_animal_p->get('field_kind')->value,
        'quantity' => $wild_animal_p->get('field_quantity')->value,
        'sex' => $wild_animal_p->get('field_sex')->value,
        'age_class' => $wild_animal_p->get('field_age_class')->value,
      ];
    }

    return [
      'pid' => $hunting_permit_p->id(),
      'hunter' => $hunter,
      'wild_animals' => $wild_animals,
    ];
  }

  /**
   * Returns hunting permit series data.
   */
  public function getHuntingPermitSeries(NodeInterface $hunting_permit_series_node) {
    $node_id = $hunting_permit_series_node->id();
    $label = $hunting_permit_series_node->getTitle();
    $id = $hunting_permit_series_node->get('field_id')->value;

    $current_datetime = new DrupalDateTime();
    $start_date = $hunting_permit_series_node->get('field_start_date')->date;
    $end_date = $hunting_permit_series_node->get('field_end_date')->date;
    $start_date_formatted = $start_date->format('j. n. Y');
    $end_date_formatted = $end_date->format('j. n. Y');
    $year = $start_date->format('Y');

    $hunting_permits = [];
    $hunting_permit_p_values = $hunting_permit_series_node->get('field_hunting_permits')->getValue();
    foreach ($hunting_permit_p_values as $hunting_permit_p_value) {
      /** @var \Drupal\paragraphs\ParagraphInterface $hunting_permit_p */
      if ($hunting_permit_p = $this->database->paragraphStorage->load($hunting_permit_p_value['target_id'])) {
        if ($hunting_permit = $this->getHuntingPermit($hunting_permit_p)) {
          $hunting_permits[] = $hunting_permit;
        }
      }
    }

    return [
      'nid' => $node_id,
      'id' => $id,
      'label' => $label,
      'start_date' => $start_date_formatted,
      'end_date' => $end_date_formatted,
      'date_html' => $start_date_formatted . ' &ndash; ' . $end_date_formatted,
      'is_valid' => $this->base->validateDatetimeRange($start_date, $current_datetime) && $this->base->validateDatetimeRange($current_datetime, $end_date),
      'hunting_permits' => $hunting_permits,
      'view_url' => '/povolenky/' . $year . '/' . $id,
      'edit_url' => '/node/' . $node_id . '/edit',
      'delete_url' => '/node/' . $node_id . '/delete',
    ];
  }

  /**
   * Returns hunting permit series list data.
   */
  public function getHuntingPermitSeriesList() {
    $results = [];
    $hunting_permit_series_nodes = $this->database->getHuntingPermitSeriesNodes();
    foreach ($hunting_permit_series_nodes as $hunting_permit_series_node) {
      $results[] = $this->getHuntingPermitSeries($hunting_permit_series_node);
    }
    return $results;
  }

  /**
   * Exports hunting permits to PDF file.
   */
  public function exportHuntingPermitsToPdf($ids = NULL) {
    $current_datetime = new DrupalDateTime();
    $file_name = 'povolenky-' . $current_datetime->format('Y-m-d') . '.pdf';

    $elements = [
      '#theme' => 'hunting_permits_pdf',
      '#custom' => ['results' => $this->getHuntingPermitListByIds($ids)],
    ];
    $html_markup = $this->renderer->renderPlain($elements);

    $this->pdf->generate('Povolenky k lovu', $file_name, $html_markup, 'hunting-permits');
  }

  /**
   * Exports hunting permit list to PDF file.
   */
  public function exportHuntingPermitListToPdf($ids = NULL) {
    $current_datetime = new DrupalDateTime();
    $file_name = 'seznam-povolenek-' . $current_datetime->format('Y-m-d') . '.pdf';

    $elements = [
      '#theme' => 'hunting_permit_list_pdf',
      '#custom' => ['results' => $this->getHuntingPermitListByIds($ids)],
    ];
    $html_markup = $this->renderer->renderPlain($elements);

    $this->pdf->generate('Seznam vydaných povolenek k lovu', $file_name, $html_markup, 'hunting-permit-list');
  }

  /**
   * Returns hunting permit ID.
   */
  private function getHuntingPermitId(DrupalDateTime $start_date, int $series_id, int $hunter_id) {
    if ($year = $start_date->format('y')) {
      return $start_date->format('y') . '/' . $series_id . '/' . $hunter_id;
    }
    return NULL;
  }

  /**
   * Returns hunting permit list data by provided IDs.
   */
  private function getHuntingPermitListByIds($ids) {
    $results = [];
    foreach ($ids as $id) {
      /** @var \Drupal\paragraphs\ParagraphInterface $hunting_permit_p */
      if ($hunting_permit_p = $this->database->paragraphStorage->load($id)) {
        if ($p_id = $hunting_permit_p->id()) {
          $hunting_permit_data = $this->getHuntingPermit($hunting_permit_p);
          if (isset($hunting_permit_data['hunter']['id'])) {
            $hunter_id = $hunting_permit_data['hunter']['id'];

            $hunting_permit_series_node = $this->database->getHuntingPermitSeriesByHuntingPermitPid($p_id);
            $series_id = $hunting_permit_series_node->get('field_id')->value;
            $start_date = $hunting_permit_series_node->get('field_start_date')->date;
            $end_date = $hunting_permit_series_node->get('field_end_date')->date;

            $settings = $this->config->get('tarabka.settings');
            $hunt_name = $settings->get('hunt.name');
            $hunt_user_name = $settings->get('hunt.user.name');
            $hunt_user_residence = $settings->get('hunt.user.residence');
            $print_signatures = $settings->get('hunting_permit.print_signatures');

            $results[] = array_merge($hunting_permit_data, [
              'id' => $this->getHuntingPermitId($start_date, $series_id, $hunter_id),
              'start_date' => $start_date->format('j. n. Y'),
              'end_date' => $end_date->format('j. n. Y'),
              'hunt' => [
                'name' => $hunt_name,
                'user' => [
                  'name' => $hunt_user_name,
                  'residence' => $hunt_user_residence,
                ],
              ],
              'signature_image_1_html' => $print_signatures ? '<img src="' . DRUPAL_ROOT . '/../private/tarabka/lk.png" height="8mm" width="auto" alt="Podpis">' : NULL,
              'signature_image_2_html' => $print_signatures ? '<img src="' . DRUPAL_ROOT . '/../private/tarabka/pk.png" height="10mm" width="auto" alt="Podpis">' : NULL,
            ]);
          }
        }
      }
    }
    return $results;
  }

}
