<?php

namespace Drupal\tarabka\Service;

use DateTimeZone;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\node\NodeInterface;

/**
 * Class Database.
 */
class Database {

  /**
   * Node storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  public EntityStorageInterface $nodeStorage;

  /**
   * Paragraph storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  public EntityStorageInterface $paragraphStorage;

  /**
   * Term storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  public EntityStorageInterface $termStorage;

  /**
   * Constructor.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    $this->nodeStorage = $entity_type_manager->getStorage('node');
    $this->paragraphStorage = $entity_type_manager->getStorage('paragraph');
    $this->termStorage = $entity_type_manager->getStorage('taxonomy_term');
  }

  /**
   * Returns hunting permit series nodes.
   */
  public function getHuntingPermitSeriesNodes() {
    $query = $this->nodeStorage->getQuery();
    $query->condition('type', TARABKA_HUNTING_PERMIT_SERIES_BUNDLE);
    $query->sort('field_start_date', 'DESC');
    $query->sort('field_id', 'DESC');
    $node_ids = $query->execute();
    return $node_ids ? $this->nodeStorage->loadMultiple($node_ids) : [];
  }

  /**
   * Returns number of stored hunting permit series nodes.
   */
  public function countHuntingPermitSeriesNodes(int $id = NULL, int $year = NULL, int $not_node_id = NULL) {
    $query = $this->nodeStorage->getQuery();
    $query->condition('type', TARABKA_HUNTING_PERMIT_SERIES_BUNDLE);
    if (isset($id)) {
      $query->condition('field_id', $id);
    }
    if (isset($year)) {
      $query->condition('field_start_date', $this->getDate($year, 1, 1), '>=');
      $query->condition('field_start_date', $this->getDate($year + 1, 1, 1), '<');
    }
    if (isset($not_node_id)) {
      $query->condition('nid', $not_node_id, '!=');
    }
    return count($query->execute());
  }

  /**
   * Returns hunting permit series node next id.
   */
  public function getHuntingPermitSeriesNodeNextId() {
    $year = (new DrupalDateTime())->format('Y');
    $query = $this->nodeStorage->getQuery();
    $query->condition('type', TARABKA_HUNTING_PERMIT_SERIES_BUNDLE);
    $query->condition('field_start_date', $this->getDate($year, 1, 1), '>=');
    $query->sort('field_id', 'DESC');
    $node_ids = $query->execute();
    foreach ($node_ids as $node_id) {
      if ($node_id) {
        if ($node = $this->nodeStorage->load($node_id)) {
          if ($id = $node->get('field_id')->value) {
            return $id + 1;
          }
        }
      }
    }
    return 1;
  }

  /**
   * Returns hunter nodes.
   */
  public function getHunterNodes(bool $is_member = NULL) {
    $query = $this->nodeStorage->getQuery();
    $query->condition('type', TARABKA_HUNTER_BUNDLE);
    if (isset($is_member)) {
      $query->condition('field_is_member', TRUE);
    }
    $query->sort('field_last_name');
    $query->sort('field_first_name');
    $query->sort('field_id');
    $node_ids = $query->execute();
    return $node_ids ? $this->nodeStorage->loadMultiple($node_ids) : [];
  }

  /**
   * Returns number of stored hunter nodes.
   */
  public function countHunterNodes(int $id = NULL, bool $is_member = NULL, int $not_node_id = NULL) {
    $query = $this->nodeStorage->getQuery();
    $query->condition('type', TARABKA_HUNTER_BUNDLE);
    if (isset($id)) {
      $query->condition('field_id', $id);
    }
    if (isset($is_member)) {
      $query->condition('field_is_member', $is_member);
    }
    if (isset($not_node_id)) {
      $query->condition('nid', $not_node_id, '!=');
    }
    return count($query->execute());
  }

  /**
   * Returns wild animal terms.
   */
  public function getWildAnimalTerms() {
    $query = $this->termStorage->getQuery();
    $query->condition('vid', TARABKA_WILD_ANIMAL_BUNDLE);
    $query->sort('field_kind');
    $query->sort('field_sex');
    $query->sort('field_age_class');
    $query->sort('field_quantity');
    $term_ids = $query->execute();
    return $term_ids ? $this->termStorage->loadMultiple($term_ids) : [];
  }

  /**
   * Returns number of stored wild animal terms.
   */
  public function countWildAnimalTerms(string $kind, string $quantity, string $sex, string $age_class, int $not_term_id = NULL) {
    $query = $this->termStorage->getQuery();
    $query->condition('vid', TARABKA_WILD_ANIMAL_BUNDLE);
    $query->condition('field_kind', $kind);
    $query->condition('field_quantity', $quantity);
    $query->condition('field_sex', $sex);
    $query->condition('field_age_class', $age_class);
    if (isset($not_term_id)) {
      $query->condition('tid', $not_term_id, '!=');
    }
    return count($query->execute());
  }

  /**
   * Returns hunting permit series node by provided PID.
   */
  public function getHuntingPermitSeriesByHuntingPermitPid($p_id) {
    $query = $this->nodeStorage->getQuery();
    $query->condition('type', TARABKA_HUNTING_PERMIT_SERIES_BUNDLE);
    $query->condition('status', TRUE);
    $query->condition('field_hunting_permits.entity.id', $p_id);
    $node_ids = $query->execute();
    foreach ($node_ids as $node_id) {
      if ($node_id) {
        if ($node = $this->nodeStorage->load($node_id)) {
          return $node;
        }
      }
    }
    return NULL;
  }

  /**
   * Returns hunting permit series nodes by provided hunter node.
   */
  public function getHuntingPermitSeriesNodesByHunterNode(NodeInterface $hunter_node) {
    $query = $this->nodeStorage->getQuery();
    $query->condition('type', TARABKA_HUNTING_PERMIT_SERIES_BUNDLE);
    $query->condition('field_hunting_permits.entity.field_hunter', $hunter_node->id());
    $query->sort('field_start_date', 'DESC');
    $query->sort('field_id', 'DESC');
    $node_ids = $query->execute();
    return $node_ids ? $this->nodeStorage->loadMultiple($node_ids) : [];
  }

  /**
   * Creates new wild animal term (if not exist yet).
   */
  public function createWildAnimal($kind, $quantity, $sex, $age_class) {
    if ($kind && !$this->countWildAnimalTerms($kind, $quantity, $sex, $age_class)) {
      $wild_animal = $this->termStorage->create([
        'name' => 'TMP',
        'vid' => TARABKA_WILD_ANIMAL_BUNDLE,
      ]);
      $wild_animal->set('field_kind', $kind);
      $wild_animal->set('field_quantity', $quantity);
      $wild_animal->set('field_sex', $sex);
      $wild_animal->set('field_age_class', $age_class);
      try {
        $wild_animal->save();
        return $wild_animal;
      }
      catch (EntityStorageException $e) {
        return FALSE;
      }
    }
    return NULL;
  }

  /**
   * Returns current date formatted for database query.
   */
  private function getCurrentDate() {
    $current_user_timezone = new DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setTime(0, 0, 0);
    $datetime->setTimezone(new DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

  /**
   * Returns current datetime formatted for database query.
   */
  private function getCurrentDatetime() {
    $current_user_timezone = new DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setTimezone(new DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

  /**
   * Returns date formatted for database query.
   */
  private function getDate(int $year, int $month, int $day) {
    $current_user_timezone = new \DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setDate($year, $month, $day);
    $datetime->setTime(0, 0, 0);
    $datetime->setTimezone(new \DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

  /**
   * Returns datetime formatted for database query.
   */
  private function getDatetime(int $year, int $month, int $day, int $hour, int $minute, int $second) {
    $current_user_timezone = new \DateTimeZone(date_default_timezone_get());
    $datetime = new DrupalDateTime('NOW', $current_user_timezone);
    $datetime->setDate($year, $month, $day);
    $datetime->setTime($hour, $minute, $second);
    $datetime->setTimezone(new \DateTimeZone(DateTimeItemInterface::STORAGE_TIMEZONE));
    return $datetime->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

}
