<?php

namespace Drupal\tarabka\Service\Validator;

use Drupal\Core\Form\FormStateInterface;
use Drupal\tarabka\Service\Base;

/**
 * Class DefaultValidator.
 */
class DefaultValidator {

  /**
   * Base service.
   *
   * @var \Drupal\tarabka\Service\Base
   */
  public Base $base;

  /**
   * Constructor.
   */
  public function __construct(Base $base) {
    $this->base = $base;
  }

  /**
   * Returns target ID from the passed field array.
   */
  public function getTargetIdFromField($field) {
    if (is_array($field)) {
      if (isset($field[0]['target_id'])) {
        return $field[0]['target_id'];
      }
      elseif (isset($field['target_id'])) {
        return $field['target_id'];
      }
    }
    return [];
  }

  /**
   * Returns value from the passed field array.
   */
  public function getValueFromField($field) {
    if (is_array($field)) {
      if (isset($field[0]['value'])) {
        return $field[0]['value'];
      }
      elseif (isset($field['value'])) {
        return $field['value'];
      }
    }
    return [];
  }

  /**
   * Returns subform from the passed paragraph array.
   */
  public function getSubformFromParagraph($paragraph) {
    if (is_array($paragraph) && isset($paragraph['subform'])) {
      return $paragraph['subform'];
    }
    return [];
  }

  /**
   * Validates past date.
   */
  public function validatePastDate(FormStateInterface $form_state, $field_key, $error_message) {
    if ($value = $this->getValueFromField($form_state->getValue($field_key))) {
      if (!$this->base->validatePastDate($value)) {
        $form_state->setErrorByName($field_key, $error_message);
      }
      return $value;
    }
    return NULL;
  }

  /**
   * Validates postal code field.
   */
  public function validatePostalCode(FormStateInterface $form_state, $field_key) {
    $old_value = $this->getValueFromField($form_state->getValue($field_key));
    $new_value = intval($this->base->removeWhitespaces($old_value));
    if ($new_value && !(is_int($new_value) && $new_value >= 10000 && $new_value <= 99999)) {
      $form_state->setErrorByName($field_key, 'PSČ není ve správném formátu.');
      return NULL;
    }
    $form_state->setValue($field_key, [['value' => $new_value]]);
    return $new_value;
  }

  /**
   * Validates e-mail field.
   */
  public function validateEmail(FormStateInterface $form_state, $field_key) {
    $old_value = $this->getValueFromField($form_state->getValue($field_key));
    $new_value = $this->base->removeWhitespaces($old_value);
    if ($new_value && !$this->base->validateEmail($new_value)) {
      $form_state->setErrorByName($field_key, 'E-mail není ve správném formátu.');
      return NULL;
    }
    $form_state->setValue($field_key, [['value' => $new_value]]);
    return $new_value;
  }

  /**
   * Validates telephone number field.
   */
  public function validatePhoneNumber(FormStateInterface $form_state, $field_key) {
    $old_value = $this->getValueFromField($form_state->getValue($field_key));
    $new_value = $this->base->removeWhitespaces($old_value);
    if ($new_value && strlen($new_value) == 9) {
      $new_value = '+420' . $new_value;
    }
    $form_state->setValue($field_key, [['value' => $new_value]]);
    return $new_value;
  }

  /**
   * Returns and save new value of provided field without all whitespaces.
   */
  public function removeWhitespaces(FormStateInterface $form_state, $field_key) {
    $old_value = $this->getValueFromField($form_state->getValue($field_key));
    $new_value = $this->base->removeWhitespaces($old_value);
    $form_state->setValue($field_key, [['value' => $new_value]]);
    return $new_value;
  }

  /**
   * Returns and save new value of provided field without multiple whitespaces.
   */
  public function removeMultipleWhitespaces(FormStateInterface $form_state, $field_key) {
    $old_value = $this->getValueFromField($form_state->getValue($field_key));
    $new_value = $this->base->removeMultipleWhitespaces($old_value);
    $form_state->setValue($field_key, [['value' => $new_value]]);
    return $new_value;
  }

}
