<?php

namespace Drupal\tarabka\Service\Validator;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Form\FormStateInterface;
use Drupal\node\NodeInterface;
use Drupal\tarabka\Service\Database;

/**
 * Class HuntingPermitSeriesNodeValidator.
 */
class HuntingPermitSeriesNodeValidator {

  /**
   * Database service.
   *
   * @var \Drupal\tarabka\Service\Database
   */
  protected Database $database;

  /**
   * Default validator service.
   *
   * @var \Drupal\tarabka\Service\Validator\DefaultValidator
   */
  protected DefaultValidator $defaultValidator;

  /**
   * Constructor.
   */
  public function __construct(Database $database, DefaultValidator $default_validator) {
    $this->database = $database;
    $this->defaultValidator = $default_validator;
  }

  /**
   * Validates all hunting permit series node fields.
   */
  public function validateAll(FormStateInterface $form_state) {
    /** @var \Drupal\node\NodeInterface $hunting_permit_series_node */
    $hunting_permit_series_node = $form_state->getFormObject()->getEntity();
    $this->validateUniqueness($form_state, $hunting_permit_series_node);
    $this->validateDates($form_state);
    $this->validateHuntingPermits($form_state);
  }

  /**
   * Validates uniqueness of hunting permit series node.
   */
  private function validateUniqueness(FormStateInterface $form_state, NodeInterface $hunting_permit_series_node) {
    $id = (int) $this->defaultValidator->getValueFromField($form_state->getValue('field_id'));
    if (!$id) {
      $form_state->setErrorByName('field_id', 'ID není ve správném formátu.');
      return;
    }

    /** @var \Drupal\Core\Datetime\DrupalDateTime $start_date */
    $start_date = $this->defaultValidator->getValueFromField($form_state->getValue('field_start_date'));
    $year = (int) $start_date->format('Y');
    if (!$year) {
      $form_state->setErrorByName('field_start_date', 'Datum vystavení není ve správném formátu.');
      return;
    }

    if ($this->database->countHuntingPermitSeriesNodes($id, $year, $hunting_permit_series_node->id())) {
      $form_state->setErrorByName('field_id', 'ID není unikátní (aktuálně reprezentuje jinou sérii ve stejném roce).');
    }
  }

  /**
   * Validates date fields.
   */
  private function validateDates(FormStateInterface $form_state) {
    $start_date = $this->defaultValidator->getValueFromField($form_state->getValue('field_start_date'));
    $end_date = $this->defaultValidator->getValueFromField($form_state->getValue('field_end_date'));
    if ($start_date instanceof DrupalDateTime && $end_date instanceof DrupalDateTime && !$this->defaultValidator->base->validateDatetimeRange($start_date, $end_date)) {
      $form_state->setErrorByName('field_start_date', 'Datum platnosti musí být pozdější než datum vystavení.');
      $form_state->setErrorByName('field_end_date', 'Datum platnosti musí být pozdější než datum vystavení.');
    }
  }

  /**
   * Validates and removes empty hunting permits.
   */
  private function validateHuntingPermits(FormStateInterface $form_state) {
    $hunting_permit_hunter_ids = [];
    $hunting_permit_p_values_new = [];
    $hunting_permit_p_values = $form_state->getValue('field_hunting_permits');
    foreach ($hunting_permit_p_values as $i => $hunting_permit_p_value) {
      if ($hunting_permit_ps_value = $this->defaultValidator->getSubformFromParagraph($hunting_permit_p_value)) {
        if (isset($hunting_permit_ps_value['field_hunter'])) {
          if ($hunter_id = $this->defaultValidator->getTargetIdFromField($hunting_permit_ps_value['field_hunter'])) {
            if ($hunter_node = $this->database->nodeStorage->load($hunter_id)) {
              $wild_animal_values = [];
              $wild_animal_p_values_new = [];
              foreach ($hunting_permit_ps_value['field_wild_animals'] as $j => $wild_animal_p_value) {
                if ($wild_animal_ps_value = $this->defaultValidator->getSubformFromParagraph($wild_animal_p_value)) {
                  if (isset($wild_animal_ps_value['field_kind'])) {
                    $kind = $this->defaultValidator->base->removeMultipleWhitespaces($this->defaultValidator->getValueFromField($wild_animal_ps_value['field_kind']));
                    $quantity = $this->defaultValidator->base->removeMultipleWhitespaces($this->defaultValidator->getValueFromField($wild_animal_ps_value['field_quantity']));
                    $sex = $this->defaultValidator->base->removeMultipleWhitespaces($this->defaultValidator->getValueFromField($wild_animal_ps_value['field_sex']));
                    $age_class = $this->defaultValidator->base->removeMultipleWhitespaces($this->defaultValidator->getValueFromField($wild_animal_ps_value['field_age_class']));
                    if ($kind) {
                      foreach ($wild_animal_values as $wild_animal_value) {
                        if ($wild_animal_value['kind'] == $kind && $wild_animal_value['quantity'] == $quantity && $wild_animal_value['sex'] == $sex && $wild_animal_value['age_class'] == $age_class) {
                          $form_state->setErrorByName('field_hunting_permits][' . $i . '][subform][field_wild_animals][' . $j, 'Povolenka nesmí obsahovat vícekrát totožnou zvěř.');
                        }
                      }
                      $wild_animal_values[] = [
                        'kind' => $kind,
                        'quantity' => $quantity,
                        'sex' => $sex,
                        'age_class' => $age_class,
                      ];
                      $wild_animal_p_value['subform']['field_kind'][0]['value'] = $kind;
                      $wild_animal_p_value['subform']['field_quantity'][0]['value'] = $quantity;
                      $wild_animal_p_value['subform']['field_sex'][0]['value'] = $sex;
                      $wild_animal_p_value['subform']['field_age_class'][0]['value'] = $age_class;
                      $wild_animal_p_values_new[] = $wild_animal_p_value;
                    }
                  }
                }
              }
              if ($wild_animal_p_values_new) {
                if (in_array($hunter_id, $hunting_permit_hunter_ids)) {
                  $form_state->setErrorByName('field_hunting_permits][' . $i . '][subform][field_hunter', 'Série nesmí obsahovat vícekrát povolenku pro totožného myslivce.');
                }
                $hunting_permit_hunter_ids[] = $hunter_id;
                $hunting_permit_p_value['subform']['field_wild_animals'] = $wild_animal_p_values_new;
                $hunting_permit_p_values_new[] = $hunting_permit_p_value;
              }
            }
          }
        }
      }
    }
    $form_state->setValue('field_hunting_permits', $hunting_permit_p_values_new);
  }

}
